/*
Example of how to use template variable type "ctl::any".
In this example, class "AnyMapCI" is designed based on stadnard map. It
uses string as key and "any" type of variable as value of the map: "std::map<string, any>",
so that "AnyMapCI" can map any types of variables or class. In this example, class "studentBase" and 
"informaticStudent" are mapped. Since "AnyMapCI" is used as remote component(the main program invoke 
the "AnyMapCI" through "ctlanymap.ci"), we also need to serialize the class to send it to the remote component..
*/
#include <iostream>
#include <string>
#include <map>
#include <ctl.h>
#include <ctl/ci/anymap.ci>  // interface to the remote component "AnyMapCI".


/*
class studentBase contains basic information of student, name, age and id, and the methods of setting 
and getting the variables.
*/
class studentBase
{
public:
	studentBase(void){}
	virtual ~studentBase(){}
	
	void setName(std::string name_){name = name_;}
	void setAge(int age_){age = age_;}
	void setId(int id_){id = id_;}
	std::string getName(void){return name;}
	int getAge(void){return age;}
	int getId(void){return id;}

	// Since we need to map the class in a remote component, we should decide how to pass the 
	// variables to the remote component. In ctl, we have to serialize the variables through 
	// the tupel. You can simply think this as a protocol, just tell ctl the order of the variables
	// to be passed. ctl has already had serialization of all basic types: int, float, double, char,..., 
	// standard library types, array and so on. So, what we should do is just give ctl the order of all 
	// basic variables of the class. 
	// e.g. CTL_Type(ClassName, tupel, (variable1,...variableN), N)
	CTL_Type(studentBase, tupel, (name, age, id), 3);
protected:
	std::string name;
	int age;
	int id;
};

/*
Derived class informaticStudent based on class studentBase contains additional variable "languages".
*/
class informaticStudent : public studentBase
{
public:
	informaticStudent(void){}

	void setLanguages(std::vector<std::string> languages_)
	{
		int size = (int)languages_.size();
		languages.resize(size);
		for(int i =0;i<size;i++)
			languages[i] = languages_[i];
	}
	std::vector<std::string> getLanguages(void){return languages;}
	
	// There are two possible methods to tell ctl the order of serialization.
	// Method1: CTL_Type(DerivedClassName, tupel, (*(BaseClassName*)this, DerivedClassVariable1,...,DerivedClassVariableN), 1+N)
	// Method2: CTL_Type(DerivedClassName, tepel, (BaseClassVariable1,...,BaseClassVariableN, DerivedClassVariable1,...,DerivedClassVariableM),N+M)
	CTL_Type(informaticStudent, tupel,( *(studentBase*)this, languages), 2); // method1
	//CTL_Type(informaticStudent, tupel, (name, age, id, languages), 4); // method2
protected:
	std::vector<std::string> languages;
};


int main()
{
	// Invoke remote component "AnyMapCI", we need to tell the location of the remote component. 
	ctl::location ANYMAPCILocation("anymap.so -l thread -f log.txt");
//	ctl::location ANYMAPCILocation("anymap.exe -l tcp -f log.txt");

	// Link the location
	ctl::link p(ANYMAPCILocation);
	// Declare the remote component and link it to the location.
	AnyMapCI AnyMap(p);
	

	// Set student A
	studentBase studentA;
	studentA.setName("Joey");
	studentA.setAge(26);
	studentA.setId(10000);

	// Set student B
	informaticStudent studentB;
	studentBase &studentBRef=studentB;
	studentB.setName("Ross");
	studentB.setAge(28);
	studentB.setId(20000);
	std::vector<std::string> languages;
	languages.resize(2);
	languages[0]="C/C++";
	languages[1]="Java";
	studentB.setLanguages(languages);
	
	// Set map
	
	std::string key, value;
	key = "student";
	value = "Basic Student Information:";
try {
	AnyMap.set(key,value);
}
catch(const ctl::exception &e)
{ printf(" cought %s\n", e.what() ); }
catch(...)
{ printf(" cought \n"); }
	key = "infostudent";
	value = "Informatic Student Information:";
	AnyMap.set(key,value);
	// Put student A and student B to the map. See that because we use ctl::any, the 
	// template variable type, we can put any types of variables or classes into it.
	AnyMap.set("student1", studentA);
	ctl::connect<informaticStudent>();
	AnyMap.set("student2Ref", studentBRef);

	// Get map
	
	// Declare "ctl::any" variable res, and get the variables or classes we put in the map above in the res.
	ctl::any res;
	res <<  AnyMap.get("student");
	// Convert the res to the original variable type by using ctl 
	// pointer casting: ctl::pointer_cast<OriginalType>(ctl::any)
	std::string *pd1 = ctl::pointer_cast<std::string>(res);
	if(pd1!=0) // check whether the pointer exist
	{
		// print out the result.
		std::cout<< *pd1 <<std::endl;
	}
	
	// Get studnet A out of map
	res << AnyMap.get("student1");
	studentBase *ps1 = ctl::pointer_cast<studentBase>(res);
	if(ps1!=0)
	{
		std::cout<< "Name: "<<ps1->getName() <<std::endl;
		std::cout<< "Age: "<<ps1->getAge() <<std::endl;
		std::cout<< "Id: "<<ps1->getId() <<std::endl;
	}

	res << AnyMap.get("infostudent");
	const std::string *pd2 = ctl::pointer_cast<std::string>(res);
	if(pd2!=0)
	{
		std::cout<< *pd2 <<std::endl;
	}
	
	// Get student B out of map

	res << AnyMap.get("student2Ref");
	studentBase *ps3 = ctl::pointer_cast<studentBase>(res);
	if(ps3!=0)
	{
		informaticStudent *ps4 = dynamic_cast<informaticStudent*>(ps3);
		if(!ps4)
			 std::cout<< "cast failed\n";
		std::cout<< "Name: "<<ps4->getName() <<std::endl;
		std::cout<< "Age: "<<ps4->getAge() <<std::endl;
		std::cout<< "Id: "<<ps4->getId() <<std::endl;
		std::cout<< "Languages: "<<ps4->getLanguages()[0] <<std::endl;
		std::cout<< "Languages: "<<ps4->getLanguages()[1] <<std::endl;
	}
	return 0;
}
