#ifndef _QUASI_NEWTON_H_
#define _QUASI_NEWTON_H_

#include <math.h>
#include <ctl/utl/vector.h>

namespace solver
{
typedef ctl::real8 real;
typedef ctl::vector<real> realVec;
typedef ctl::vector<ctl::int4> intVec;

//! quasiNewtonSolver ScalarSolver solves nonlinear equations by BMGF algorithmus
template<class F> class quasiNewtonSolver
{
	const F &fM;
	realVec x0M;
	
	mutable ctl::vector<realVec> svecM, yvecM;
	mutable realVec syM, stM, tM;
	real epsM;
	int nM;
	
	realVec H(const realVec x,int k) const
	{
		int j;
		if(k<=nM)
			tM = x;
		int m= k>=nM ? k-nM : 0;
		for(j=k-1; j>=m; j--)
		{
			stM[j] =  fM.scalMult(svecM[j],tM);
			tM -=  (stM[j]/syM[j]) * yvecM[j] ;
		}
		for(j=m; j<k; j++)
			tM += ( (stM[j] - fM.scalMult(yvecM[j],tM) )/syM[j] )* svecM[j];
		return tM;
	}
	void addDyads(const realVec &dx,const realVec &dy) const
	{   
		svecM.append(dx);
		yvecM.append(dy);
		syM.append(fM.scalMult(dx,dy));
		stM.append(0.0);	
	}
	real norm(const realVec &x) const
	{ return sqrt(fM.scalMult(x,x)); }
public:
	typedef  realVec result;
	quasiNewtonSolver(const F &f) : fM(f),epsM(1e-4), nM(1000)
	{}
	void setX0(const  realVec &x0)
	{ x0M = x0; }
	void setEps(real eps)
	{ epsM = eps; }
	void setNDyads(int n)
	{ nM = n; }
				
	realVec operator()(const realVec &rc =realVec()) const
	{                                                                        
		realVec x=x0M,y, r=rc;
		if(!r)
		{
			y = -fM(x);
			r.resize(y.size());
		}
		else
		{
			x.resize(r.size());
			y = r - fM(x);
		}
		if(!x)
			x.resize(r.size());
				
		realVec ay=y;
		const int  K=1000;	
		int k;		
		for(k=0; norm(y)>epsM && k<K; k++)
		{	 	
			real f = 1;
			realVec dx = H(y, k);
			while(fM.scalMult(y = r-fM(x+dx),dx)<-epsM)
				dx *= 0.5, f*=0.9;
//			printf("line search lambda = %e\n",f);
			x += dx;
			if(norm(ay-y)<epsM)
				break;
			addDyads(dx, ay-y);
			ay = y;
		}
		if(k==K)
			x.resize(0);
printf(" #iteration = %d\n",k);
		return x;
	}
};

} // solver
#endif
