/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifdef __alpha
#       define MPICH_SKIP_MPICXX
#endif

#include "lammpi.h"
#include <mpi.h>

namespace ctl
{

	CTL_Inline mpi_comm CommunicatorImpl<lam>::mpiCommu(int logId) const
	{
		bool ID0 = (logId==0), me0 = (getLogId()==0);
		if(ID0==me0)
			return MPI_COMM_WORLD;
		return interCommM;
	}
	CTL_Inline void CommunicatorImpl<lam>::attachBuffer(int size)
	{
		mpiBufferM.resize(size);
		MPI_Buffer_attach(mpiBufferM.begin(),size);
	}
	CTL_Inline int CommunicatorImpl<lam>::test(PID &source, int &type,  mpi_comm commu) const
   {
//	   ctl::log(envM, "#1 mpi::test src=%d type = %d\n",*source,*type);
		int flag=0;
		MPI_Status status;
		int MPIsource 	= !source  ? MPI_ANY_SOURCE 	: mpiID(source);
		int MPItag 			= type<0		?  MPI_ANY_TAG			: type;
	
		MPI_Iprobe(MPIsource, MPItag, (MPI_Comm)commu, &flag, &status);
		int size = -1;
		if (flag)
		{
			source = logID(status.MPI_SOURCE,  commu==interCommM);
			type = status.MPI_TAG;
			MPI_Get_count(&status, MPI_CHAR, &size);
		}
  		return size;
	}
	CTL_Inline int CommunicatorImpl<lam>::read(iStream &is, PID &source, int &type, mpi_comm commu) const
   {
		int size = is.size();
		MPI_Status status;
		int MPIsource 	= !source  ? MPI_ANY_SOURCE 	: mpiID(source);
		int MPItag 			= type<0		?  MPI_ANY_TAG			: type;
		if(size<=0)
  		{
			MPI_Probe(MPIsource, MPItag,  (MPI_Comm)commu, &status);
			MPIsource = status.MPI_SOURCE; 
			type = status.MPI_TAG;
 	  		MPI_Get_count(&status, MPI_CHAR, &size);
		}
		CTL_Assert(size>0);
		is.set(size,0);
		MPI_Recv(is.buffer(), size, MPI_CHAR, MPIsource , type,   (MPI_Comm)commu, &status);
  		source = logID(MPIsource,  commu==interCommM);
		ctl::log(envM, "#m >> lam: src=%d size = %d tag = %d\n", int(source), size, type);
		return size;
	}
	CTL_Inline void CommunicatorImpl<lam>::initWithoutParent(const ctl::vector<location> &loc)
	{
		int nProcs = loc.size()+1;
		PID myPID = 0;
		ctl::vector<int> ptid(nProcs);
		int i;
		MPI_Info info;
		MPI_Info_create(&info);
		if(nProcs>1)
		{
			int errcodes[nProcs];
			char start[24], nodes[24], sysCall[81];
			sprintf(start,".lamstart%d",int(sys::getPid()));

			sprintf(nodes,".lamnodes%d",int(sys::getPid()));
			sprintf(sysCall,"lamnodes -i > %s", nodes);
			ctl::log(envM, "#d >$ '%s'\n", sysCall);
			int err=::system(sysCall);
			if(err)
			{	ctl::log(envM, "#d !! call of \"%s\" failed\n", sysCall); return; }
			FILE *nodesF = ctl::fopen(nodes,"r");
			if(!nodesF)
			{	ctl::log(envM, "#d !! can't read lamnodes\n");  return; }
			int logID= -1;
			char machine[81];
			std::map<ctl::string, int> list;
			while(1)
			{
				int read=fscanf(nodesF,"n%d %s\n", &logID, machine);
				if(read<2)
					break;
				char *p=strchr(machine,':');
				if(p)
					*p ='\0';
				ctl::log(envM, "#d ?? read: logID=%d machine = \"%s\"\n",	 logID, machine);

				list[machine] = logID;
//			list[location::shortHostName(machine)] = logID;
			}
			ctl::fclose(nodesF);
			remove(nodes);
			FILE *startF = ctl::fopen(start,"w");
			int p=0;
			for(p=0; p<nProcs-1; p++)
			{
				ctl::string host=location::shortHostName(loc[p].host);
				if(location::isLocalHost(loc[p].host) )
					list[host]=0;
				const std::map<ctl::string, int>::iterator n=list.find(host);
				if(n==list.end())
				{
					ctl::log(envM, "#d !! lam: machine '%s' not known\n", loc[p].host.c_str());
					ctl::log(envM, "#d !! lam: machine '%s' not known\n", host.c_str());
				 	break;
				}
				ctl::string cmd="";
				if(!loc[p].debug.empty())
					cmd = loc[p].debug;
				if(!loc[p].term.empty())		
					cmd = loc[p].term + cmd ;
				if(loc[p].args.empty())
		   		fprintf(startF," n%d %s  %s -l lam\n", n->second, cmd.c_str(), loc[p].pathExec().c_str());
				else
		   		fprintf(startF," n%d %s  %s -l lam %s\n", n->second, cmd.c_str(), loc[p].pathExec().c_str(), loc[p].args.c_str());
	   	}
			ctl::fclose(startF);
	   	if(p+1<nProcs)
	   		return;
			MPI_Info_set(info,(char*)"file", start);
	 		MPI_Comm_spawn(0,0,0, info, 0, MPI_COMM_WORLD,  (MPI_Comm*)&interCommM, errcodes);
			remove(start);	 	
			for(i=0; i<nProcs-1; i++)
			{
				char err[512];
				int len=512;
				if(errcodes[i]!=MPI_SUCCESS)
				{
					MPI_Error_string(errcodes[i],err,&len);
					ctl::log(envM, " #d !! MPI_Error  %s occured\n",err);
					ctl::log(envM, " #d !! Could not spawn \"%s\"\n",loc[i].getSSH().c_str());
					return;
				}
				else
					ctl::log(envM, "#m ?? '%s' spawned on %s\n" ,loc[i].exec.c_str(), loc[i].host.c_str() );
			}
		}
#ifdef CTL_MPIBufferSize
		mpiBufferSizeM = CTL_MPIBufferSize;
#else
		mpiBufferSizeM = 1024*1024;
#endif
		attachBuffer(mpiBufferSizeM);
		set(myPID, nProcs);
	}
   CTL_Inline int CommunicatorImpl<lam>::write_(oStream &os,const PID &dest ,int type) const
   {
		int size=os.size();
   		MPI_Request request;
		MPI_Ibsend( (void*)os.buffer(), size, MPI_CHAR, mpiID(dest), type,  (MPI_Comm)mpiCommu(dest),&request);
		ctl::log(envM, "#m << lam: dest=%d size = %d tag = %d\n", int(dest), size,type);
		return size;
   }
   CTL_Inline int CommunicatorImpl<lam>::test_(PID &source, int *type) const
   {
		int size = -1;
   		if(!!source)
   			return test(source, *type, mpiCommu(source));
		size = test(source, *type, interCommM);
		if(size<0)
			size = test(source, *type, MPI_COMM_WORLD);
   		return size;
	}
	CTL_Inline int CommunicatorImpl<lam>::read_(iStream &is, PID &source, int *type) const
   {
		int size=is.size();
		if(!source)
			while(1)
			{
				if((size=test_(source, type))>=0)
					break;
				usleep(0);
			}
		return read(is, source, *type, mpiCommu(source));
	}

// called by ctlGroup(const ctl::vector<location> &loc, type = lammpi)
   CTL_Inline CommunicatorImpl<lam>::CommunicatorImpl(int n, const char * const  *arg): Communicator(false)
   {
   		CommunicatorImpl<mpi>::initMPI(n, arg);
		int  nProcs;
		ctl::int4  myPID;
		MPI_Comm_size(MPI_COMM_WORLD, &nProcs);
		MPI_Comm_rank(MPI_COMM_WORLD, &myPID);
#ifdef CTL_MPIBufferSize
		mpiBufferSizeM = CTL_MPIBufferSize;
#else
		mpiBufferSizeM = 1024*1024;
#endif

		attachBuffer(mpiBufferSizeM);
		set(myPID+1, nProcs+1);
		MPI_Comm_get_parent( (MPI_Comm*)&interCommM);
		int parentSize=0;
		MPI_Comm_size( (MPI_Comm)interCommM, &parentSize);
//      CTL_Assert(parentSize==1);
	}
	CTL_Inline CommunicatorImpl<lam>::~CommunicatorImpl()
	{
// one last Sync. to let  MPI_Finalize work fine;
//	 	sync(this);
		ctl::log(envM, "#m -- Communicator<lam>\n");
//		int size=mpiBufferM.size();
 //		if(size>0)
 //			MPI_Buffer_detach(mpiBufferM.begin(),&size);
 // 		MPI_Finalize();
 	}

} // ctl

