/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#include "mpi.h"
#include <stdlib.h>

#ifdef __alpha
#       define MPICH_SKIP_MPICXX
#endif

#include <mpi.h>

namespace ctl
{
	CTL_Inline  int CommunicatorImpl<mpi>::getLogIdS(int n , const char* const *arg, const location &loc)
	{
		initMPI(n,arg);
		int id;
		MPI_Comm_rank(MPI_COMM_WORLD, &id);
		return id;
	}
 	CTL_Inline void CommunicatorImpl<mpi>::initMPI(int n, const char* const *arg)
	{
		int done;
    MPI_Initialized(&done);
		ctl::log("#m ++ invoking MPI_Init(%d, <args>) done = %d\n", n, done);
		if(!done)
		{
			MPI_Init(&n,(char***)&arg);
			done = true;
		}
	}

   CTL_Inline int CommunicatorImpl<mpi>::write_(oStream &os,const PID &dest ,int type) const
   {
		int size = os.size();
   		MPI_Request request;
		MPI_Ibsend( (void*)os.buffer(), os.size(), MPI_CHAR, int(dest.port), type, MPI_COMM_WORLD,&request);
		ctl::log(envM,"#m << mpi: dest=%d size=%d tag=%d\n",int(dest.port),size,type);
		return size;
   }
   CTL_Inline int CommunicatorImpl<mpi>::test_(PID &source, int *type) const
   {
//    ctl::log("#m mpi::test src=%d type = %d\n",int(source),*type); 
		int flag=0;
		MPI_Status status;
		int MPIsource 	= !source  ? MPI_ANY_SOURCE 	:  int(source);
		int MPItag 			= *type<0		?  MPI_ANY_TAG			: *type;
	
		MPI_Iprobe(MPIsource, MPItag, MPI_COMM_WORLD, &flag, &status);
		int size = -1;
		if (flag )
		{
			source = status.MPI_SOURCE;
			*type = status.MPI_TAG;
			MPI_Get_count(&status, MPI_CHAR, &size);
		}
  	return size;
	}
   CTL_Inline int CommunicatorImpl<mpi>::read_(iStream &is, PID &source, int *type) const
   {
		MPI_Status status;
		int MPIsource 	= !source  ? MPI_ANY_SOURCE 	: int(source.port);
		int MPItag 			= *type<0		?  MPI_ANY_TAG			: *type;
		int size = is.size();
		if(!size)
			size = -1;
		if(size<0)
  		{
			MPI_Probe (MPIsource, MPItag, MPI_COMM_WORLD, &status);
			source=status.MPI_SOURCE;
  			*type = status.MPI_TAG;
  	  		MPI_Get_count(&status, MPI_CHAR, &size);
		}
		CTL_Assert(size>0);
		is.set(size, 0);
		MPI_Recv(is.buffer(), size, MPI_CHAR, int(source) , *type,  MPI_COMM_WORLD, &status);
	   ctl::log(envM, "#m >> mpi: src=%d size = %d tag = %d\n",int(source),size,*type);

		return size;
	}
   CTL_Inline CommunicatorImpl<mpi>::CommunicatorImpl(int n, const char * const  *arg): Communicator(false), mpiBufferM(0)
   { 
   		initMPI(n, arg);
		if(n==0)
		{ set(0, 1);	return; }

		int  nProcs;
		ctl::int4  myPID;
		MPI_Comm_size(MPI_COMM_WORLD, &nProcs);
		MPI_Comm_rank(MPI_COMM_WORLD, &myPID);
#ifdef CTL_MPIBufferSize
		mpiBufferSizeM = CTL_MPIBufferSize;
#else
		mpiBufferSizeM = 1024 * 1024;
#endif
		mpiBufferM = new char[mpiBufferSizeM];

		MPI_Buffer_attach(mpiBufferM,mpiBufferSizeM);
		set(myPID, nProcs);
	}
	CTL_Inline CommunicatorImpl<mpi>::~CommunicatorImpl()
	{
// one last Sync. to let  MPI_Finalize work fine;
//	 	sync(this);
		ctl::log(envM, "#m -- Communicator<mpi>\n");
//		if(mpiBufferM)
// 			MPI_Buffer_detach(mpiBufferM,&mpiBufferSizeM);
//  	();
//  	delete[] mpiBufferM;
	}

} // ctl

