/******************************************************************************

 This file is part of the Component Template Library (CTL)

 Copyright (C) 2006  Rainer Niekamp
 email:              r.niekamp@tu-bs.de

 The CTL is free software; you can redistribute it and/or modify it under the
 terms of the GNU Lesser General Public License as published by the Free Software
 Foundation; either version 2.1 of the License, or (at your option) any later
 version. We only ask that proper credit be given to the author.

 This library is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 for more details.

 You should have received a copy of the GNU Lesser General Public License along
 with this library; if not, write to the Free Software Foundation,
 Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

 ******************************************************************************/

#ifndef CTL_COMMU_INTERFACE_H_
#define CTL_COMMU_INTERFACE_H_

namespace ctl
{
	struct ioTag
	{
		ctl::int4 tag;
		operator ctl::int4() const
		{
			return tag;
		}
		ioTag(int t = -1) :
			tag(t)
		{
		}
	};

	/**///  class Communicator
	class Communicator: public commuBase
	{
		friend class groupImpl;
		protected:
		PID myPIDM;
		PID logIdM;
		int nProcsM;
		const groupImpl *ownerM;
		const link* ownerLinkM;
		mutable int tagCounterM;
		void setOwner(const groupImpl *owner)
		{
			ownerM = owner;
		}
		void check() const;
		mutable ctl::int8 mCounterM;
		mutable bool checkedM;
		mutable int lockedM;
		private:
		mutable PID lastConnectM;
		protected:
		ctl::vector<char> magicM;
		Environment *envM;
		ctl::string destrMessM;
		void set(const PID & logId, int nProcs, const PID &physId = PID(), const ctl::string &destrMess = ctl::string())
		{
			myPIDM = !physId ? logId : physId;
			logIdM = logId;
			nProcsM = nProcs;
			destrMessM = destrMess;
			//fprintf(stderr, "message = %s\n", destrMess.c_str());
		}
		PID getConnectionNB() const
		{
			PID src;
			int t = DAT;
			test(src, t);
			return lastConnectM;
		}
		public:
		mutable iStream isM;
		virtual iStream &getIStream() const
		{
			return isM;
		}
		virtual bool closed() const
		{
			return nProcs() < 2;
		}
		virtual ctl::socket *getSocket() const
		{
			return 0;
		}
		static PID magic()
		{
			return PID(0x1d920d4f, "1f3e:a28e:2cf0:9378:aa01:744:5d31:710a");
		}
		static PID magicSwp()
		{
			return ctl::byteBuffer::mustSwap() ? PID(0x4f0d921d, "3e1f:8ea2:f02c:7893:1aa:4407:315d:a71") : magic();
		}
		virtual PID getConnection(bool blocking = true) const
		{
			PID src;
			while(!getConnectionNB() || !blocking)
				;
			src = lastConnectM;
			lastConnectM = PID();
			return src;
		}
		void setLastConnect(const PID& pid) const
		{
			lastConnectM = pid;
		}
		PID getLastConnect() const
		{
			return lastConnectM;
		}
		virtual ioTag getTag() const
		{
			tagCounterM++;
			if(tagCounterM == UNDEF - 255) tagCounterM = 1;
			return 255 + tagCounterM;
		}
		const groupImpl* getOwner() const
		{
			return ownerM;
		}
		const link *getOwnerLink() const
		{
			return ownerLinkM;
		}
		virtual bool addJob(void(*)(void*), void *) const
		{
			return false;
		}
		// defined in group.h
		virtual PID getPhysId(const PID &) const;
		virtual int nProcs() const
		{
			return nProcsM;
		}
		const PID &getPhysId() const
		{
			return myPIDM;
		}
		const PID &getLogId() const
		{
			return logIdM;
		}
		const PID &getMyId() const
		{
			return nProcsM == 1 ? myPIDM : logIdM;
		}
		class _lock
		{
			int &locked;
			_lock& operator=(const _lock&);
		public:
			_lock(int &locked): locked(locked)
			{
				Wait w;
				while(locked)
				{ w(true); }
				locked = 1;
			}
			~_lock()
			{ locked = 0; }
		};
		int write(oStream &os, const PID & dest, int tag) const
		{
			_lock L(lockedM);
			check();
			return write_(os, dest, tag);
		}
		int test(PID &source, int &tag) const
		{
			_lock L(lockedM);
			check();
			return test_(source, &tag);
		}
		int read(iStream &is, PID &source, int &tag) const
		{
			_lock L(lockedM);
			check();
			return read_(is, source, &tag);
		}

		virtual int write_(oStream &, const PID & dest, int tag) const=0;
		virtual int test_(PID &source, int *tag) const=0;
		virtual int read_(iStream &, PID &source, int *tag) const=0;
		virtual bool isValid() const
		{
			return true;
		}
		virtual channelT getType() const=0;
		virtual bool isEmpty() const
		{
			return true;
		}
		Communicator(bool hasWriteV) :
			commuBase(false, hasWriteV), myPIDM(-1), nProcsM(0), ownerM(0), ownerLinkM(0), tagCounterM(0), mCounterM(0),
			    checkedM(false), lockedM(0), envM(&env())
		{
			char buf[128];
			ctl::oStream os = ctl::serializeB(magicSwp(), buf, 128);
			// independant magic String
			magicM = ctl::vector<char>(os.size(), os.buffer());
		}
		virtual ~Communicator()
		{
			if(!destrMessM.empty()) ctl::log("#lm %s\n", destrMessM.c_str());
		}
		virtual void clean()
		{
		}
		static int getLogIdS(int, const char* const *, const location&);
		static CommunicatorSP newExternal(const location &loc);
	};

	inline CommunicatorSP getCommunicator();

#ifdef CTL_MSC
	template<int> class CommunicatorImpl;
	/*: public Communicator
	 {
	 public:
	 CommunicatorImpl(int =0, const char * const * =0)
	 {}
	 CommunicatorImpl(const ctl::vector<location> &)
	 {}
	 ~CommunicatorImpl()
	 { }
	 int write(const void *data, int size, const PID & dest, int tag) const
	 { return 0; }
	 int test(PID &source, int *tag) const
	 { return 0; }
	 int read(void *&data, int size, PID &source, int *tag) const
	 { return 0; }
	 int rel(void *data) const
	 { return 0; }

	 };
	 */
#else
	template<int> class CommunicatorImpl;

#endif

	CTL_Inline channelT getChannel(const char *s);
	CTL_Inline CommunicatorSP createCommunicator(const ctl::vector<location> &loc, channelT linkage, bool late = false);
	CTL_Inline CommunicatorSP createCommunicator(int n, const char* const *arg, channelT typ);

	inline bool nWrite(const CommunicatorSP &commu, const oStream &os, PID dest, ioTag type)
	{
		return commu->write((oStream&) os, dest, type) >= 0;
	}
	/*
	 namespace internal
	 {
	 template<class T> bool nWrite(ctl::trueT,const CommunicatorSP &commu,const T &t, const PID &dest, ioTag type =DAT)
	 {
	 oStream os(sizeof(T), &t);
	 return commu->write(os,dest,type)>=0;
	 }

	 template<class T> inline bool nWrite(ctl::falseT,const CommunicatorSP &commu,const T &t, const PID &dest, ioTag type =DAT)
	 {
	 char buf[CTL_defaultBufferSize];
	 return ctl::nWrite(commu,ctl::serializeB(t, (void*)buf,CTL_defaultBufferSize,commu),dest,type);
	 }
	 }
	 */
	template<class T> inline bool nWrite(const CommunicatorSP &commu, const T &t, const PID &dest, ioTag type = DAT)
	{
		char buf[CTL_defaultBufferSize];
		return ctl::nWrite(commu, ctl::serializeB(t, (void*) buf, CTL_defaultBufferSize, commu), dest, type);
		//	enum { pod=ctl::property<ctl::isPOD, T>::value && !ctl::typeTraits<T>::isPointer };
		//	return  internal::nWrite(ctl::int2type<pod> (),commu,t,dest,type);
	}
	/*
	 template<class T> inline bool nWrite(const CommunicatorSP &commu,const ctl::vector<T> &t,const PID &dest, ioTag tag =DAT)
	 {
	 enum { pod=ctl::property<ctl::isPOD, T>::value && !ctl::typeTraits<T>::isPointer };
	 return internal::nWrite(ctl::int2type<pod>(),commu,t,dest,tag);
	 }
	 */
	template<class T> inline bool nRead(const CommunicatorSP &commu, T &t, PID source = -1, ioTag tag = DAT)
	{
		iStream &is = commu->getIStream();
		is.sourceM = source;
		is.hostM = commu->getLogId();
		//	iStream is(0,0,source, PID(commu->getLogId() ) );
		int type = tag;
		int size = commu->read(is, source, type);
		if(size < 0) return false;
		is.commuM = commu;

		ctl::deserialize(is, t);
		is.commuM.clear();
		return true;
	}

} // ctl

#endif

