/******************************************************************************

This file is part of the Component Template Library (CTL)

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
for more details.

You should have received a copy of the GNU Lesser General Public License along
with this library; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_COMMUNICATOR_INL
#define CTL_COMMUNICATOR_INL

#include "ctl/communicator.h"
#include "io.inl"

namespace ctl
{
// Communicator with buffering
	template<class bufferT> class bCommunicator : public Communicator
	{
		public:
			bCommunicator ( bool hasWriteV = false) : Communicator ( hasWriteV ) {}
		protected:
			typedef bufferT ioBuffer;
			typedef  std::set<ioBuffer> setT;
			typedef typename setT::iterator iterator;

			mutable setT messageM;
			bool isEmpty() const
			{ return messageM.empty(); }
			iterator find ( const PID &src, int tag ) const
			{
				if(tag<0)
						return messageM.begin();
				bufferT key ( 0,tag, src,-1 );
				iterator end = messageM.end(), mI=messageM.upper_bound ( key );
				if ( mI==end || mI->tag!=tag )
					return end;
				if ( !src || src==mI->pid )
					return mI;
				return end;
			}
			ioBuffer extract ( const PID &src, int tag ) const
			{
				iterator mI = find ( src, tag );
				if ( mI!=messageM.end() )
				{
					bufferT m=*mI;
					messageM.erase ( mI );
					return m;
				}
				return bufferT();
			}
	};
	struct messageForward
	{
		virtual ~messageForward()
		{}
		virtual void forward() const=0;
	};

} // ctl

#ifdef CTL_LAM

#	ifndef	CTL_MPI
#		define CTL_MPI
#	endif
#endif

#ifdef	CTL_MPI
#	include "ctl/com/mpi.inl"

#endif

#ifdef CTL_LAM
#	ifndef LAM_MPI
#		error ctl::Communicator<lam> needs the lam implementation of mpi
#	endif
#	include "ctl/com/lammpi.inl"
#endif

#ifdef CTL_PVM
#	include "ctl/com/pvm.inl"
#endif

#	include "commu_ext.h"
#ifdef CTL_EXT
#	include "comm_conn.inl"
#endif

#	include "ctl/com/serial.inl"
#	include "ctl/com/tcpip.inl"
#	include "ctl/com/dmn.inl"
#	include "ctl/com/file.inl"
#	include "ctl/com/pipe.inl"


namespace ctl
{

	CTL_Inline channelT getChannel ( const char *s )
	{
		if ( strncmp ( "file", s, 4 ) ==0 )
			return file;
		if ( strncmp ( "serial", s, 6 ) ==0 )
			return serial;
		if ( strncmp ( "mpi", s,3 ) ==0 )
			return mpi;
#ifdef CTL_LAM
		if ( strncmp ( "lam", s,3 ) ==0 )
			return lam;
#endif
#ifdef CTL_PVM
		if ( strncmp ( "pvm", s,3 ) ==0 )
			return pvm;
#endif
		if ( strncmp ( "tcp", s, 3 ) ==0 )
			return tcp;
		if ( strncmp ( "thread", s, 6 ) ==0 )
			return thread;
		if ( strncmp ( "lib", s, 3 ) ==0 )
			return lib;
		if ( strncmp ( "ssh", s, 3 ) ==0 )
			return ssh;
		if ( strncmp ( "pipe", s, 4 ) ==0 )
			return pipe;
		if ( strncmp ( "dmn", s, 3 ) ==0 )
			return dmn;
		if ( strncmp ( "ext", s, 3 ) ==0 )
			return ext;
		if ( strncmp ( "logid", s, 5 ) ==0 )
			return logid;
		if ( strncmp ( "default", s,7 ) ==0 )
			return defaultChannel;
		return undefined;
	}

	CTL_Inline CommunicatorSP createCommunicator ( const ctl::vector<location> &loc, channelT channelT, bool late )
	{
		CommunicatorSP commu;
		if(!!loc && !!loc[0].pipeTail)
			channelT = ext;
		switch ( channelT )
		{
			case serial: /*commu = CommunicatorSP(new CommunicatorImpl<serial>());*/ break;
			case file: commu = CommunicatorSP ( new CommunicatorImpl<file> ( loc[0].pathExec().c_str() ) ); break;
#ifdef CTL_MPI
			case mpi: if ( !loc ) commu = CommunicatorSP ( new CommunicatorImpl<mpi> ( 0,0 ) );	break;
#endif
#ifdef CTL_PVM
			case pvm: commu = CommunicatorSP ( new CommunicatorImpl<pvm> ( loc ) ); break;
#endif
#ifdef CTL_LAM
			case lam: commu = CommunicatorSP ( new CommunicatorImpl<lam> ( loc ) ); break;
#endif
			case pipe: commu = CommunicatorSP ( new CommunicatorImpl<pipe> ( loc[0] ) ); break;
#ifdef CTL_EXT
			case cnct: commu = CommunicatorSP ( new CommunicatorImpl<cnct> ( loc[0] ) ); break;
#endif
		case ssh:case tcp:
			commu = CommunicatorSP ( new CommunicatorImpl<tcp> ( loc, "tcp", late ) ); break;
		default:
			if(!!loc)
				commu = Communicator::newExternal( loc[0] ); break;
		}
		if ( !!commu && !commu->getLogId() )
			commu.destruct();

		return	commu;
	}

	CTL_Inline CommunicatorSP createCommunicator ( int n, const char* const *arg, channelT typ )
	{
		CommunicatorSP commu;
		switch ( typ )
		{
			case serial: commu = CommunicatorSP ( new CommunicatorImpl<serial> ( n,arg ) ); break;

			case file: commu = CommunicatorSP(); break;
#ifdef CTL_MPI
			case mpi:commu = CommunicatorSP ( new CommunicatorImpl<mpi> ( n,arg ) ); break;
#endif
#ifdef CTL_PVM
			case pvm: commu = CommunicatorSP ( new CommunicatorImpl<pvm> ( n,arg ) ); break;
#endif
#ifdef CTL_LAM
			case lam: commu = CommunicatorSP ( new CommunicatorImpl<lam> ( n,arg ) ); break;
#endif
			case pipe: commu = CommunicatorSP ( new CommunicatorImpl<pipe> ( n,arg ) ); break;
			case dmn: commu = CommunicatorSP ( new CommunicatorImpl<dmn> ( n,arg ) ); break;
#ifdef CTL_EXT
			case cnct: commu = CommunicatorSP ( new CommunicatorImpl<cnct> ( n,arg ) ); break;
#endif
			case ssh: case tcp: commu = CommunicatorSP ( new CommunicatorImpl<tcp> ( n,arg ) ); break;
			default: break;
		}
//	ctl::log("#4 ++ Communicator<%s>\n",   location::channel2string(typ) );

		if ( !!commu && !commu->getLogId() )
			commu.destruct();
		return	commu;
	}

	CTL_Inline int Communicator::getLogIdS ( int n , const char* const *arg, const location &loc )
	{
		(void)n;
		(void)arg;
		switch ( loc.linkage )
		{
#ifdef CTL_LAM
			case lam:
#endif
#ifdef CTL_MPI
			case mpi:
				return  CommunicatorImpl<mpi>::getLogIdS ( n, arg, loc );
#endif
#ifdef CTL_PVM
			case pvm:
				return  CommunicatorImpl<pvm>::getLogIdS ( n, arg, loc );
#endif
			case tcp:
			default:
				return loc.logId;
		}
	}

} //  ctl

#endif

