/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef  CTL_ENVIRONMENT_H_
#define  CTL_ENVIRONMENT_H_

#include "io.h"
#include <stdarg.h>

/**/// namespace ctl
namespace ctl
{



template<class Job> struct jobQueue
{
	struct jobItem: public ctl::shared
	{
		Job job;
		jobItem *next;
		jobItem(Job j): job(j), next(0) {}
	};
	uint8 I, R;
	volatile uint8 W;
	jobItem *head, *iter, *tail;
	bool closed;
	bool isValid() const
	{ return !closed; }
 
	jobItem *getItem(Job job)
	{
		if(R-I<2)
			return new jobItem(job);
		jobItem *tmp = head;
		head = tmp->next;
		tmp->job = job; tmp->next = 0;
		I++;
		return tmp;
	}
	jobQueue(): I(1), R(1), W(1), head(0), iter(0), tail(0), closed(false) {}
	~jobQueue()
	{
		while(head)
		{
			jobItem *tmp = head;
			head = tmp->next;
			delete tmp;
		}
	}
	bool empty() const
	{ return W<=R; }
	void put(Job job)
	{
		jobItem *it = getItem(job);
		if(!tail)
			head = tail = it;
		else
			tail = tail->next = it;
		W++;
	}
	Job *view()
	{
		if(W<=R)
			return 0;
		return iter ? &iter->next->job : &head->job;
	}
	Job &get()
	{
		Wait w;
		while(W<=R )
			w(true);
		iter = iter ? iter->next : head;
		R++;
		return iter->job;
	}
};

extern "C"
{ typedef void (*sigHandle)(int); }

namespace themes
{
enum { User = 1, eXcpt = 2, Mess = 4, Fnc = 8, Link = 16, Cnnct = 32, Warn = 64,
				Var = 128, Dbg = 256, All = 511, silent = 4096 };
			
CTL_Inline void change(int &ths, int thema, char op)
{	ths = op=='+' ? ths | thema : ths & (All-thema); }

CTL_Inline const char *getThemes(const char *str, int &ths)
{				 
	if(!str || !*str)
		return 0;
	if(strncmp(str, "all", 3)==0)
	{ ths = All; str += 3; }
	if(strncmp(str, "none", 4)==0)
	{ ths = 0; str += 4; }
	if(str[0]=='s')
	{ ths |= silent; return str+1; }
	if(str[0]=='a') // all
		ths = All;
	else str--;
	char op = '+';
	while(op)
	{
		switch(*++str)
		{
		case 'u': case 'U': // user
			change(ths, User, op);		break;
		case 'x': case 'X': // exceptions
			change(ths, eXcpt, op);	break;
		case 'm': case 'M': // messages
			change(ths, Mess, op);		break;
		case 'f': case 'F': // function/method calls
			change(ths, Fnc, op);		break;
		case 'l': case 'L': // link creation
			change(ths, Link, op);		break;
		case 'c': case 'C': // connected CIs/Functions/Types 
			change(ths, Cnnct, op);		break;
		case 'w': case 'W': // warnings 
			change(ths, Warn, op);		break;
		case 'v': case 'V': // variables/environment 
			change(ths, Var, op);		break;
		case 'd': case 'D': // debug info 
			change(ths, Dbg, op);		break;
		case 'a': case 'A': // debug info 
			ths = All;		break;
		case '-': case '+':
			op = *str;
			break;
		default:
			op = 0;
		}
	}
	return str;	
}

} // themes

class Environment
{

	struct Job
	{
		typedef void(*funcT)(void*);
		funcT func;
		void *data;
		sys::threadIdT parentID;
		Job(funcT f=0, void * d=0, sys::threadIdT id=0): func(f), data(d), parentID(id)
		{}
	};
	friend Environment &env(bool ini, const char*);
	friend Environment* createEnv(Environment *parentEnv, channelT);
	typedef void (*sigHandleT)(int);
	typedef std::pair<int, sigHandleT> signalEntryT;
	ctl::vector<signalEntryT> sigHandlesM;
	static const char *sigName(int sig);
	void setSignals();
	void resetSignals();
	static void sighandle(int sig);
	ctl::lPointer<locationSelectBase> locationSelectorM;

	bool inDestrM;
	typedef RemoteSP symbolT;
	typedef ctl::lexicalAssoc<char, symbolT> symbolTableT;
	symbolTableT symbolTableM, intrinsicSymbolTableM;
	std::pair<ctl::string, symbolT> tmpSymbolM;
	ctl::lexicalAssoc<char, ctl::link> linkMapM;
	ctl::lexicalAssoc<char, ctl::string> locationModifierMapM;
	std::set<RemoteSP> objectListM;

	std::set<const Remote* > referenceListM;

	typedef std::map<ctl::typeId, std::set<RemoteSP> > namedObjMapT;
	namedObjMapT namedObjMapM;

	typedef std::set<const groupImpl*> groupListT;
	groupListT groupListM;

	mutable groupSP mainGroupM, dmnGroupM;
	location locationM;
	ctl::string errorM;
	mutable ctl::string parentStrM;
	mutable link parentM, monitorM;
	ctl::vector<ctl::string> cmdArgM;
//	const char *const *argCLM;
	Environment *parentEnvM, *childEnvM, *rootEnvM, *threadEnvM;
	bool threadM;
	link  clientM;
	const link  *pclientM;
	std::set<sys::LibHandle> libHandlesM;
	int lockM;
public:
	// returning h in root.libHandlesM ?
	bool addLibHandle(sys::LibHandle h)
	{
		if(rootEnvM != this)
			return rootEnvM->addLibHandle(h);
		std::set<sys::LibHandle>::const_iterator f=libHandlesM.find(h);
		if(f==libHandlesM.end())
		{
			Communicator::_lock L(lockM);
			libHandlesM.insert(h);
			return false;
		}
		return true;
	}
private:
	typedef void(*connectorT)();  
	connectorT connectorM;
	ctl::string selectStrM;
	vector<TimeMarker*> tMarker;
	link *tmpLinkM;
	typedef std::map<PID, link> remoteLinkT;
	remoteLinkT remoteLinkM;
	typedef lexicalAssoc<char, ctl::string> variablesT;
	struct variableSpaceT
	{
		mutable variablesT variables;
		void set(const char* key, const char* val) const
		{
			if(!val || !*val || *val==location::comma)
				variables.erase(key);
			else
				variables[key] = val;
		}		
		const char* get(const char* key) const
		{
			const ctl::string *s=variables.find(key);
//ctl::log("this = %p get key (%s) found = %p\n",this,  key, s);
			return s ? s->c_str() : 0;
		}	
		void setLocation(const string& scope, const char* loc=0) const
		{	set(("CTL::"+scope).c_str(), loc);	}
		const char* getLocation(const ctl::string& scope) const
		{	return get(("CTL::"+scope).c_str() ); }
		void setVariable(const char* key, const char* val=0) const
		{	set(key, val);	}
		const char* getVariable(const char* key) const
		{	return get(key); }
		bool operator!() const
		{ return variables.empty(); }
		ctl::string getVariableList() const
		{
			ctl::string list;
			for(unsigned int i=0; i<variables.size(); i++)
			{ 
				list += ctl::string(variables.getKey(i) ) + "=" +  variables.get(i);
				if(i+1<variables.size() )
 					list += (char)location::comma; 
			}
			return list;
		}
		const char *readVariable(const char* line, const Environment* expo) const
		{
			char buf[CTL_StringBufferSize] = "CTL::", *c=buf+5;
			while(isspace(*line)) line++;
			while(*line && *line!=location::comma && c-buf<CTL_StringBufferSize-6)
			*c++ = *line++;	
			*c = 0;	
			while(isspace(*--c) )
				*c = 0;
			c= buf+5;
			char *sep=strchr(c,'=');
			if(!sep)
			{
				if(!(sep = strstr(c,"->") ) )
				{
					if(expo)
					{
						const char* s=0;
						if( (s = expo->getVariable(c) ) )
							expo->exportSpaceM.set(c, s);
						else if( (s = expo->getVariable(buf) ) )
							expo->exportSpaceM.set(buf, s);
					}
					return 0;						
				}
				c = buf; *sep++ = 0;
//fprintf(stderr, "this=%p loc key = '%s' val = '%s'\n",this,  buf, sep);
			}
			*sep = 0;
			for(char *i=c; *i; i++)
				if(isspace(*i) )
					*i = 0;
			
			char *end=sep;
			while(isspace(* --end) );
			while(isspace(* ++sep) );
			set(c, sep);
			if(expo)
				expo->exportSpaceM.set(c, sep);
			return *line == location::comma ? line+1 : 0;
		}

	
	}  envSpaceM, _envSpaceM, exportSpaceM ;
//	mutable variablesT variablesM, variablesExportM;
	bool readVariables(const char *file, ctl::string& err, int lev, bool reportErr=true) const;
	typedef std::pair<void*, void(*)(const void*)> staticObjT ;
	typedef std::list<staticObjT>  staticObjListT;
	staticObjListT staticObjectListM;
public:
	const char *readVariable(const char* line, const char* file, ctl::string& err, int lev) const;
	location addExportedVariables(const location &loC) const
	{
		location loc = loC;
		ctl::string monitor=getVariable("CTL_Monitor"),
								themes=getVariable("CTL_LogThemes"),
								varlist;		
		if(!!monitor)
		{	varlist = "CTL_Monitor="; varlist+=monitor; }
		if(!!themes)
		{
			if(!!monitor)
				varlist += ", ";
			varlist += "CTL_LogThemes=";
			varlist += themes;
		}
		if(!!varlist)
		{ loc.args += " -X ["; loc.args += varlist ; loc.args += "]"; }
		return loc;
	}
	void setConnector(connectorT conn)
	{ connectorM = conn; }
	connectorT getConnector()
	{ return connectorM; }
	void addStatic(staticObjT p, bool toParent=false)
	{ 
			if(toParent)
			{
				Environment *e = parentEnvM ? parentEnvM : this;
				e->staticObjectListM.push_back(p);
			}
			else
				staticObjectListM.push_back(p);
//		ctl::clog<<"++ static this="<<this<<"  "<<p.first<<"\n";
	}
	void destroystaticObjects()
	{
//		ctl::clog<<"destroystaticObjects this="<<this<<"\n";
		for(staticObjListT::iterator p=staticObjectListM.begin(); p!=staticObjectListM.end(); p++)
		{
//			ctl::clog<<p->first<<"\n";
			(*p->second)(p->first);
		}
	}
 	ctl::vector<std::pair<const char*, ctl::string> > getAllVariables()
	{
		const lexicalAssoc<char, ctl::string> &a=envSpaceM.variables, &b=exportSpaceM.variables, &c=_envSpaceM.variables;
		ctl::vector<std::pair<const char*, ctl::string> > v(a.size() + b.size() + c.size() );
		size_t z=0;
		for(size_t i=0; i<a.size(); i++)
			v[z++] = std::pair<const char *, ctl::string>(a.getKey(i), a.get(i) );
		for(size_t i=0; i<b.size(); i++)
			v[z++] = std::pair<const char *, ctl::string>(b.getKey(i), b.get(i) );
		for(size_t i=0; i<c.size(); i++)
			v[z++] = std::pair<const char *, ctl::string>(c.getKey(i), c.get(i) );
		return v;
	}

	const lexicalAssoc<char, ctl::string> &getExportVariables() const
	{ return exportSpaceM.variables; }

	void setVariable(const char* name, const char* val = 0, bool expo = false)
	{	
		envSpaceM.setVariable(name, val);
		if(expo)
			exportSpaceM.setVariable(name, val);
	}

	void setExportVariables(const lexicalAssoc<char, ctl::string> &var)
	{
		ctl::string str = "?? setting environment::vaiables:\n";
		for(unsigned int i= 0; i<var.size(); i++)
		{
			const char *name = var.getKey(i);
			if(!exportSpaceM.variables.find(name) )
			{
				envSpaceM.set(name, var.get(i).c_str() );		
				exportSpaceM.set(name, var.get(i).c_str() );		
			}	
			str += " "; str += ctl::string(name) + " -> " + var.get(i) + "\n"; 
		}
		if(var.size()>0)
			ctl::log(this,"#v %s\n", str.c_str() ); 
	}
	ctl::string getExpVariableList() const
	{ return exportSpaceM.getVariableList(); }
	const char *getVariable(const char *name, bool /*global*/=true) const
	{
		const char* var=0;
		(var=envSpaceM.get(name) ) || (var= _envSpaceM.get(name) );
		if(!!var)
			return var;
		return 0;
//		return global ? sys::getEnv(name) : 0;
	}
	void monitor(const ctl::string &m, bool confirm) const
	{
#ifdef CTL_Monitor
		int storeLog = logThemesM;
		logThemesM |= themes::silent;
		int storeMon = monitorThemesM;
		monitorThemesM |= themes::silent;
		CTL_Assert(!!monitorM);
		if(!!monitorM)
			CTL_Try
			{ ctlCI::message(monitorM, m, confirm); }
			CTL_CatchAll
			{ monitorM = ctl::link(); }
		logThemesM = storeLog;
		monitorThemesM = storeMon;
#else
		(void)m); (void)confirm;
#endif
	}

	void setRemoteLink(const PID &pid, const link &p)
	{ threadEnvM->setRemoteLink_(pid, p); }
	void setRemoteLink_(const PID &pid, const link &p)
	{ 
		CTL_Assert(!p.grpM.holdM);
		log(this,"#m ++ (%d, %s) -> link\n", pid.port, pid.host.c_str() ); 
		remoteLinkM[pid] = p; 
	}
	void rmRemoteLink_(const PID &pid)
	{ 
		remoteLinkT::iterator pI=remoteLinkM.find(pid);
		if(pI != remoteLinkM.end() )
			remoteLinkM.erase(pI); 
		connectedLibsM.erase(pid);
	}
	void rmRemoteLink(const PID &pid)
	{ 
		log(this,"#m -- (%d, %s) -> link\n", pid.port, pid.host.c_str() ); 
		 threadEnvM->rmRemoteLink_(pid); 
	}
	const link &getRemoteLink(const PID &pid)
	{ return threadEnvM->getRemoteLink_(pid);	}
	const link &getRemoteLink_(const PID &pid)
	{
		remoteLinkT::const_iterator pI=remoteLinkM.find(pid);
		if(pI != remoteLinkM.end() )
			return pI->second;
		return remoteLinkM[pid] = ctl::link(dmnGroupM->getCommunicator(), dmnGroupM->getCommunicator()->getPhysId(pid) ); 
	}
	const ctl::string &getSelector() const
	{	return selectStrM; }  	
	bool verboseM;

// ***** io-devices *****
	ctl::device *ioDeviceM;
	int fdinM, fdoutM, fderrM, fdlogM;
	FILE *stdinM, *stdoutM, *stderrM;
	ctl::device *getDevice() const
	{ return ioDeviceM; }
	int getfdin() const
	{ return fdinM; }
	int getfdout() const
	{ return fdoutM; }
	int getfderr() const
	{ return fderrM; }
	int getfdlog() const
	{ return fdlogM; }
	mutable FILE *loggerM;
//	File logFileM;
	factory factoryM;
	sys::threadIdT tidM;
	mutable int logThemesM, monitorThemesM;
	int mpiCountM;
	string pathM;
	typedef std::map<ctl::string, link>  locateMapT;
	locateMapT locateMapM;
	char runModeM;
	string argStringM;

	mutable jobQueue<Job> threadQueueM;
	void addJob(void(*fnc)(void*), void *recv, sys::threadIdT ppid) const
	{ threadQueueM.put(Job(fnc, recv, ppid)); }
// observing only threadQueue
	bool runJob(int &n, bool blocking = true)
	{
		int nn=n;
		bool empty = threadQueueM.empty();
		n = empty ? 0 : 1;
		if(!blocking && empty )
		{ return !threadQueueM.closed; }

		Job job;
		while(!threadQueueM.closed && (job=threadQueueM.get()).func)
		{
			CTL_Assert(threadEnvM==this);
			(*job.func)(job.data);
			if(job.parentID)
				sys::wakeUp(job.parentID);
			if(nn>0 && !--nn)
				return true;
			if(!blocking && threadQueueM.empty())
				return !threadQueueM.closed;
		}
		threadQueueM.closed = true;
		return false;
	}
	void setError(const std::string &err)
	{
		errorM = err;
		if(!!mainGroupM)
			mainGroupM->errorM = errorM;
		if(!!dmnGroupM)
			dmnGroupM->errorM = errorM;
		groupListT::iterator groupI;
		for(groupI = groupListM.begin(); groupI !=groupListM.end(); groupI++)
			(*groupI)->setError(errorM);
	}
	bool setSelector(const char *locator)
	{	
		if(!selectStrM.empty() )
			return selectStrM[0]!='#';

		const char *locStr = locator ? getVariable(locator, true) : 0;
		if(locStr)
		{
			log(this,"#v ?? using locator '%s'\n", locStr);
			selectStrM = locStr;
			return true;
		}
		else
		{
			if(!!locator)
				ctl::log(this,"#w !! locator '%s' not active\n", locator);
			selectStrM = "#";
			return false;
		}
	}  	
	void setClient(const link &p)
	{	clientM = p; clientM.release();	}
	link getClient() const
	{  return  clientM; }	
	void setLocator(const  ctl::lPointer<locationSelectBase> &imp)
	{ locationSelectorM = imp; }
	bool setLocator();
	void setThread()
	{  threadM = true; }
	void setParentStr(const ctl::string str)
	 { parentStrM = str; }

	Environment *getChildEnv()
	{
		Environment *child=threadEnvM;
		while(child->childEnvM)
			child =  child->childEnvM;
		return child;
	}

	void clearChildEnv()
	{ childEnvM = 0; }
	Environment *setChildEnv(Environment *e)
	{ 
		if(e==this)
			return e;
		CTL_Assert(e != this);
		Environment *child = getChildEnv();
// if not already set
		if(e!=child)
			child->childEnvM = e;
		return child; 
	}
	Environment *getThreadEnv() const
	{ return threadEnvM; }
	Environment *getParentEnv() const
	{ return parentEnvM; }
	void setParentEnv(Environment *e)
	{ parentEnvM = e; 	}
	sys::threadIdT getTid() const
	{ return tidM; }
	const ctl::vector<ctl::string> &getCommandLine() const
	{ return  cmdArgM; }

	link getLink(const typeId &scope, const ctl::any &property, const char* =0, bool askParent = true);
	const link* getLocalLink(const typeId &scope);
	typeId getLocation(const typeId &scope)
	{
			typeId found = getLocation(scope, envSpaceM);
			if(!found)
				found = getLocation(scope, _envSpaceM);	
			if(!found)
				found = getLocation(scope, exportSpaceM);	
			return found;
	}
	typeId getLocation(const typeId &scope, const variableSpaceT& space);

	void setLink(const typeId &scope, const link &p);
	void setLocationModifier(const typeId &scope, const ctl::string &s)
	{ locationModifierMapM[scope.c_str()] = " " + s; }
	ctl::string getLocationModifier(const typeId &scope) const;
	groupSP getMainGroup() const
	{ return threadEnvM->mainGroupM; }
	group getMPIGroup() const
	{ 
		groupSP grp = getMainGroup();
		if(!!grp && grp->logId()==0 && (grp->getLinkage()==ctl::lam || grp->getLinkage()==ctl::mpi))
			return grp;
		return group(ctl::vector<ctl::location>(), ctl::mpi);
	}
	void setMainGroup(group *grpP)
	{ if(grpP) mainGroupM = grpP->grpM; }
	groupSP getDmnGroup() const
	{ return dmnGroupM; }
	void setDmnGroup(group *grpP=0)
	{
		groupSP grp = grpP ? grpP->grpM : 0;
		if(!!grp && grp->getLinkage() == dmn)
		{	dmnGroupM = grp; pysIDM = dmnGroupM->getCommunicator()->getPhysId(); }
		else
			getPhysId(false);
	}

	const link &getParent(const PID &logId) const;
	static bool initSocket();

	Environment();
	location init(Environment* parentEnv, int n, const char * const *arg);
	location init();
	void initEnv(const link &p, const char *locate, void (*conn)(), ctl::group *grp);
//	void init2(const variablesT &vars, const groupImpl *grp = 0);
	virtual ~Environment();

	const location &getMyLocation() const
	{ return locationM; }
//	bool exec(const location &dest, const ctl::string &arg, std::pair<FILE*, int>*);
	TimeMarker &newTimeMark(const char *name)
	{ 
		TimeMarker *t=new TimeMarker(name);
		tMarker.push_back(t );
		return *t;
	}
 	void addGroup(const groupImpl *grp)
	{
//log("#d addGroup(%p)\n", grp);
		if(!grp)
			return;
		groupListM.insert(grp);
	}
   void deleteGroup(const groupImpl *grp)
   {
//log("#d deleteGroup(%p,%d)\n", grp, int(inDestrM));
		if(!inDestrM)
		{
			if(grp)
				groupListM.erase(grp);
			else
			{
				mainGroupM.clear();
				dmnGroupM.clear();
			}
		}
	}
  socket *getSocket()
	{
		if(!!dmnGroupM)
 			return dmnGroupM->getCommunicator()->getSocket();	
		if(parentEnvM && !! parentEnvM->dmnGroupM)	 
			return parentEnvM->dmnGroupM->getCommunicator()->getSocket(); 
		return 0;
	}
	ctl::socket *pSocketTmpM;
	const char *pVarsTmpM;
	mutable bool ownedByParentEnvM;
	std::list<std::pair<int, ctl::link> > ownedLinksM;
	int ownedProcessIdM;
	std::map<PID, ctl::link> connectedLibsM;
	const ctl::link *getLink(const PID &pid)
	{
		if(!!pid)
		{
			std::map<PID, ctl::link>::const_iterator linkI = threadEnvM->connectedLibsM.find(pid);
			if(linkI != connectedLibsM.end() )
				return &linkI->second;
		}
		return 0;
	}
// called by dmn::connect
	bool redirectSocket(const PID &/*pid*/, socket &reader, const ctl::string& locS_, ctl::string &error)
	{ 
		ctl::string locS = locS_;
//		ctl::log("#m got socket with '%s' fd=%d port = %d remote = (%d, %s)\n", locS.c_str(), reader.getFD(), reader.getPid().port, reader.getRPid().port, reader.getRPid().host.c_str());	
		bool enableThread=runModeM != '0', forceThread=runModeM == '2';
#ifdef CTL_Dmn0Thread
		enableThread = false;
#endif
		if(!enableThread)
			return false;
#ifdef CTL_Dmn2Thread
		forceThread = true;
#endif
		if(!locS || !locS_[0])
		{
			if(forceThread)
				locS = " | -l thread";
			else
				return false;
		}
		const char* locP=locS.c_str();
		if( (pVarsTmpM = strrchr(locP,'#') ) )
			 pVarsTmpM++;
		location loc(locS);
		loc.replyDstr = true;

		if(loc.linkage!=lib)
			pSocketTmpM = &reader;
		ownedProcessIdM = 0;
		loc.pipeType = 0;
//loc.print(std::cout);
		ctl::link p(loc);
		if(!p)
		{ 
			log(this,"#m could not create link %s \n (%s)", loc.c_str(), p.error() );
			error = "could not create link with "; error += loc.str(); 
			error += "\n("; error += p.error(); error += ")";
			return false; 
		}
		if(loc.linkage==lib)
		{
			threadEnvM->connectedLibsM[reader.getRPid()] = p;
			pSocketTmpM = 0;
			return false;
		}
		ownedLinksM.push_back(std::pair<int, link> (ownedProcessIdM,p) );
		if(loc.linkage==thread)
			while(pSocketTmpM)
				sys::usleep(0);
		else
		{	pSocketTmpM =0; pVarsTmpM=0; }
		reader.release();

		return true; 
	}
	void updateOwnedLinks()
	{
		std::list<std::pair<int, ctl::link> >::iterator p=ownedLinksM.begin(); 
		while(p != ownedLinksM.end())  
			if(!p->second.run())
				break;
			else
				p++;
		if(p != ownedLinksM.end())
		{	
//			fprintf(stderr, "************* waiting for pid %d ***********\n", p->first);
#ifdef CTL_Unix
			int pid = p->first;
#endif
			ownedLinksM.erase(p);
#ifdef CTL_Unix
			if(pid)
				::waitpid(pid, 0, 0);
#endif
		}
	}
	mutable PID pysIDM;
private:
	const PID &getPhysId(bool initDmn=true) const;
public:
	const PID &getRootId() const
	{ return rootEnvM->getPhysId(); }
	const PID &getThreadId() const
	{ return threadEnvM->getPhysId(); }
	bool sameThread(Environment *e) const
	{ return e->threadEnvM==threadEnvM; }
	ctl::string getLANid(const ctl::string& client, ctl::string& locS)
	{
		log("#d getLANid('%s', '%s')\n", client.c_str(),locS.c_str());
		ctl::location loc(locS);
		ctl::string tail = loc.pipeTail;
		loc.pipeTail.clear();
//		loc.pipeType = 0;
		if(!client)
		{
			if(loc.pipeType!='>')
				return getThreadId().toString();
			ctl::link p(loc);
			if(!p)
				return ctl::string();
			ownedLinksM.push_back(std::pair<int, link> (0, p) );
			std::string id;
			ctlCI::getLANid(p, ctl::string(), tail).assign(id);
			locS = tail;
			return id; 
		}
		if(loc.pipeType!='<')  
		{
			oStream os=serialize(getPhysId());
			location pidLoc(client);
			PID pid(pidLoc.portL, pidLoc.host);;
			dmnGroupM->getCommunicator()->write(os, pid, 2);
//			log("#d getLANid2('%s',  '%s')\n", tail.c_str(), locS.c_str());
			return getPhysId().toString();
		}
		ctl::link p(loc);
		if(!p)
			return ctl::string();
		ownedLinksM.push_back(std::pair<int, link> (0, p) );
		std::string id;
		ctlCI::getLANid(p, client, tail).assign(id);
//log("#d getLANid3('%s',  '%s')\n", tail.c_str(), locS.c_str());
		locS = tail;
		return id; 
	}
	void addObject(const RemoteSP &obj)
	{ /*ctl::log(#f "addObject %p\n", obj.get() );*/ objectListM.insert(obj); }
	RemoteSP getNamedObject(const std::pair<typeId, ctl::string>& key)
	{
		namedObjMapT::const_iterator objSet=namedObjMapM .find(key.second);
		if(objSet!=namedObjMapM.end() )
		{
			for(std::set<RemoteSP>::const_iterator sp=objSet->second.begin(); sp!=objSet->second.end(); sp++)
				if((*sp)->extends(key.first) )
					return *sp;
		}
		return RemoteSP();
	}
	CTL_Inline void setNamedObject(const std::pair<typeId, ctl::string> &key, const RemoteSP &sp)
	{ namedObjMapM[key.second].insert(sp); }
	void releaseObject(const RemoteSP &obj)
	{  objectListM.erase(obj); }
//		log(this,"env::releaseObject(%p[%d])\n", obj.get(), obj.getRefCount());
		
	void holdReference(const Remote *ref)
	{ 
		referenceListM.insert(ref); 
//		log("#d holdReference: this=%p ref=%p size = %d\n", this, ref, referenceListM.size() ); 
	}
	void releaseReference(const Remote *ref)
	{ 
//		log("#d releaseReference: this=%p ref=%p size = %d\n", this, ref, referenceListM.size() ); 
		CTL_Assert(referenceListM.find(ref)!=referenceListM.end() );
		referenceListM.erase(ref); 
	}
	
	void setTmpFunction(const char *pretty, const symbolT &f)
	{ tmpSymbolM.first = pretty; tmpSymbolM.second = f; }
	symbolT getTmpFunction(const char *pretty) const
	{ 
		if(tmpSymbolM.first==pretty)
			return tmpSymbolM.second;
		return symbolT();
	}
	void addFunction(const char *pretty, const symbolT &f)
	{	
		if(!tmpLinkM)
		{
			symbolTableM[pretty] = f;	
			log(this,"#c ++ %s\n", pretty);
		}
		else
			tmpLinkM->addFunction(pretty, f);		
	}
	RemoteSP getFunction(const char *fname);
	const symbolTableT &getSymTable() const
	{ return symbolTableM; }
	ctl::vector<ctl::string> getSymbols() const
	{
		ctl::vector<ctl::string> symName(symbolTableM.size() );
		for(size_t i=0; i<symbolTableM.size(); i++)
			symName[i] = symbolTableM.getKey(int(i)); 

		return symName;
	}
	ctl::vector<ctl::string> getIntrinsicSymbols() const
	{
		ctl::vector<ctl::string> symName(intrinsicSymbolTableM.size() );
		for(size_t i=0; i<intrinsicSymbolTableM.size(); i++)
			symName[i] = intrinsicSymbolTableM.getKey(int(i)); 

		return symName;
	}
	
/**/// (non) blocking read and call of RMI's
	bool inRunJobM;

	bool run(int &n, bool blocking = false, const groupImpl* =0);
	bool openLog(const char *logFile, int theme);
class global
{
public:
	static const char *getName(char *envname)
	{
		sys::threadIdT	tid = sys::getTid();
		strcpy(envname,"CTL_ENV");
		char *c=envname+7;
		for(sys::threadIdT Z=tid; Z; Z>>=4)
			*c++ = 'a' + (Z&15);
		*c = 0;
		return envname;	
	}
	static Environment *get(const char *envname)
	{
		const char *envStr=sys::getEnv(getName((char*)envname));
		return envStr ? (Environment*)objectID::str2ptr(envStr) : 0;
	}
	static Environment *get(bool &created, bool set=true)
	{ 	
		char envname[24];
		Environment *e=get(getName(envname) );
		if(!e)
		{
			e = new Environment;
			created = true;
			if(set)
			{
				char str[24];
				objectID::ptr2str(e, str);
				sys::setEnv(envname, str);
			}
		}
		return e->getChildEnv();
	}
	static void del(Environment *e, bool unset=true)
	{ 
		char envname[24];
		getName(envname);
		if(!e)
			e = get(envname);
		delete e;
		if(unset)
			sys::unsetEnv(envname);
	}
	static void del()
	{ del(0, true); }
}; // global 
}; // Environment

//CTL_Inline void setChildEnv(Environment *childEnv)
//{  env().setChildEnv(childEnv); }

CTL_Inline group getMainGroup()
{ return env().getMainGroup(); }

CTL_Inline group getMPIGroup()
{ return env().getMPIGroup(); }

CTL_Inline group getDmnGroup()
{ return env().getDmnGroup(); }

  CTL_Inline factory &pFactory()
{ return env().factoryM; }

  CTL_Inline const char *getVariable(const char *name, bool global)
{ return env().getVariable(name, global); }

  CTL_Inline ctl::vector<std::pair<const char*, ctl::string> > getAllVariables()
	{ return env().getAllVariables(); }

  CTL_Inline void addStatic(const std::pair<void*, void(*)(const void*)> p, bool toParent)
  { env().addStatic(p, toParent); }

CTL_Inline Environment *createEnv(Environment *parentEnv, channelT type)
{
	bool created = false;
	Environment *e = type == lib ? new Environment : Environment::global::get(created, true/*set*/);
	CTL_Assert(type==lib || created);
	e->setParentEnv(parentEnv);
	if(type==lib || type==thread)
	{
		e->rootEnvM = parentEnv->rootEnvM;
		if(type==lib)
		{	
			parentEnv->setChildEnv(e);
			e->threadEnvM = parentEnv->threadEnvM;
		}
	}
	return e;
}

  CTL_Inline Environment &env_(int n, const char * const *args, const char *locatorName)
{
	bool created=false;
	Environment &envS = *Environment::global::get(created, true/*set*/);
//	if(created && ini)
//	CTL_Assert(created || n==0);
	if(created)
	{
		envS.setParentStr("ctl_main");
		if(n>0)
			envS.init(0, n, args);
		else
			envS.init();
		envS.setSelector(locatorName);
		envS.setDmnGroup();
		::atexit(&Environment::global::del);
	}
	return envS;
}

CTL_Inline void Environment::initEnv(const link &p, const char *locate, void (*conn)(), ctl::group *grp =0)
{
	setSelector(locate);
	setClient(p);
	setConnector(conn);
	if(p.getIType()!=ctl::lib)
	{
		setDmnGroup(grp);
		setMainGroup(grp);
	}
	log("#x \n");
		
}
CTL_Inline void initEnv(Environment *myEnv,const link &p, const char *locate, void (*conn)(), ctl::group *grp =0)
{ myEnv->initEnv(p, locate, conn, grp); }

CTL_Inline void setLocatorBase(const ctl::lPointer<locationSelectBase> &locate)
{ env().setLocator(locate); }

CTL_Inline link getParent(const PID &logId)
{ return env().getParent(logId); }

CTL_Inline void deleteGroup(const groupImpl *grp)
{ env().deleteGroup(grp); }

/**/// get location descriptor of running link
CTL_Inline const location &myLocation()
{ return env().getMyLocation(); }

CTL_Inline void releaseObject(const RemoteSP &obj)
{ env().releaseObject(obj); }

CTL_Inline void holdReference(Environment *e, const Remote *ref)
{ e->holdReference(ref); }

CTL_Inline void releaseReference(Environment* e, const Remote *ref)
{ e->releaseReference(ref); }

CTL_Inline link getLink(const typeId &scope, const ctl::any &prop, const char *locator)
{
	ctl::Environment *e=&env();
	return e->getLink(scope, prop, locator);
}
CTL_Inline link getLink(const typeId &scope)
{ return getLink(scope, ctl::any(), 0); }

CTL_Inline void setLink(const typeId &scope, const link &p)
{ env().setLink(scope, p); }

CTL_Inline void setLocationModifier(const typeId &scope, const ctl::string &s)
{ env().setLocationModifier(scope, s); }

CTL_Inline typeId getLocationModifier(const typeId &scope)
{ return env().getLocationModifier(scope); }

CTL_Inline void addObject(Environment *e ,const RemoteSP &r)
{ e->addObject(r); }

CTL_Inline void addFunction(const char *pretty, const RemoteSP &f)
{ env().addFunction(pretty, f); }

CTL_Inline void setTmpFunction(const char *pretty, const RemoteSP &f)
{ env().setTmpFunction(pretty, f); }

CTL_Inline const ctl::vector<ctl::string> &getCommandLine(Environment *parent)
{ return parent->getCommandLine(); }

CTL_Inline link getClient()
{ return env().getClient(); }
CTL_Inline group getGroup()
{	return groupSP(getClient().getGroup()); }

CTL_Inline void setClient(const link &p)
{  env().setClient(p); }


CTL_Inline RemoteSP getFunction(const char *pretty)
{ return env().getFunction(pretty); }

CTL_Inline const char *Environment::readVariable(const char *line, const char* file, ctl::string& err, int lev) const
{
	while(isspace(*line)) line++;
	bool expo = false;
	if(strncmp(line, "include", 7)==0 && isspace(line[7]) )
	{
		line += 8;
		while(isspace(*line)) line++;
		if(lev == 0)
				err = err + "!! could not open " + line + " (recursion depth exceeded)\n" + " !! (referenced in '" + file + "')\n";
 		else
			if(readVariables(line, err, lev) )
				err = err + " !! (referenced in '" + file + "')\n";
		return 0;
	}
	else if(strncmp(line, "export", 6)==0  && isspace(line[6]) )
	{ 
		expo = true;
		line += 6;
		while(isspace(*line)) line++;
	}
	return ((file &&  file[0]=='.') ? _envSpaceM : envSpaceM).readVariable(line, expo?this:0);

}

CTL_Inline bool Environment::readVariables(const char *file, ctl::string& err, int lev, bool reportErr) const
{
	FILE *fp = ctl::fopen(file,"r");
	if(!fp)
	{
		if(reportErr)
			err = err + "!! could not open '" + file + "' (" + strerror(errno) + ")\n";
		return true;
	}
	char line[CTL_StringBufferSize];
	const char *c;
	while( (c=readLine(line,CTL_StringBufferSize,fp)) )
	{
		if(c && *c)
			while( (c=readVariable(c, file, err, lev-1)) );
	}
	ctl::fclose(fp);
	return false;
}

CTL_Inline int getFDin()
{ return env().getfdin(); }

CTL_Inline int getFDout()
{ return env().getfdout(); }

CTL_Inline int getFDerr()
{ return env().getfderr(); }

CTL_Inline int getFDlog()
{ return env().getfdlog(); }

CTL_Inline int printf(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	int len=vfprintf(env().stdoutM, fmt, ap);
	va_end(ap);
	return len;
}

#ifdef CTL_VNstdio
extern "C"
{
int vfscanf(FILE*, const char*, ...);
}

CTL_Inline int scanf(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	int len= ::vfscanf(env().stdinM, fmt, ap);
	va_end(ap);
	return len;
}
#endif
CTL_Inline char *gets(char *m, size_t max)
{	return env().ioDeviceM->readln(m, max)<0 ? 0 : m;}

CTL_Inline bool logStr(const char* str,const Environment *e)
{
#ifndef CTL_NoLog	
	int themes = 0;
	if(*str=='#')
		str = themes::getThemes(str+1, themes);
	else
		themes = 1;
	if(!e)
		e = &env();
//fprintf(stderr, "%d %d %d %d %s\n",e->fdlogM, themes, e->logThemesM & themes::silent, themes & e->logThemesM, str);
	if(e->fdlogM>=0 && !(e->logThemesM & themes::silent) && (themes & e->logThemesM))	
		(void)::write(e->fdlogM, str, ::strlen(str) );
	if(!(e->monitorThemesM & themes::silent) && (themes & e->monitorThemesM) )
		e->monitor(str, true);
#endif
	return true;
}



CTL_Inline void rethrowStr(const std::exception &e, const char* fmt,...)
{
	char msg[CTL_StringBufferSize];
	va_list ap;
	va_start(ap, fmt);

#ifndef CTL_VNstdio
	vsprintf(msg, fmt, ap);
#else
	vsnprintf(msg, CTL_StringBufferSize-1, fmt, ap);
#endif
	va_end(ap);
	ctl::string err(msg);
	err += "\n"; err += e.what();

	log("#x !! throwing '%s'\n", msg);
	CTL_Throw(exception(err));
}

  CTL_Inline void rethrowStr2bin(const std::exception &e, binaryObject &bin, const char* fmt,...)
{
	char msg[CTL_StringBufferSize];
	va_list ap;
	va_start(ap, fmt);

#ifndef CTL_VNstdio
	vsprintf(msg, fmt, ap);
#else
	vsnprintf(msg, CTL_StringBufferSize-1, fmt, ap);
#endif
	va_end(ap);
	ctl::string err(msg);
	err += "\n"; err += e.what();
	log("#x !! throwing '%s'\n", err.c_str());
	ctl::exception e2(err);	

	ctl::binaryObject b(e2);
	bin = b;
}

  CTL_Inline void throwStr(const char* fmt,...)
{
	char msg[CTL_StringBufferSize];
	va_list ap;
	va_start(ap, fmt);

#ifndef CTL_VNstdio
	vsprintf(msg, fmt, ap);
#else
	vsnprintf(msg, CTL_StringBufferSize-1, fmt, ap);
#endif
	va_end(ap);
	log("#x !! throwing '%s'\n", msg);
	CTL_Throw(exception(msg));
}

  CTL_Inline void throwStr2bin(binaryObject &bin, const char* fmt,...)
{
	char msg[CTL_StringBufferSize];
	va_list ap;
	va_start(ap, fmt);

#ifndef CTL_VNstdio
	vsprintf(msg, fmt, ap);
#else
	vsnprintf(msg, CTL_StringBufferSize-1, fmt, ap);
#endif
	va_end(ap);
	log("#x !! throwing '%s'\n", msg);
	ctl::exception e2(msg);	

	ctl::binaryObject b(e2);
	bin = b;
}

  CTL_Inline bool closeLog()
{
#ifndef CTL_NoLog	
	Environment *e = &env();
	if(e->loggerM)
	{
		ctl::fclose(e->loggerM);
		e->loggerM = 0;
		return true;
	}
#endif
	return false;
}

CTL_Inline bool openLog(const char *logFile, int theme)
{ return env().openLog(logFile, theme); }

CTL_Inline bool Environment::openLog(const char *logFile, int theme)
{
#ifndef CTL_NoLog	
	fprintf(stderrM, "openLog:logFile='%s'\n", logFile);
	if(logFile)
	{
		if(loggerM && loggerM!=stdout && loggerM!=stdoutM && loggerM!=stderr && loggerM!=stderrM)
			ctl::fclose(loggerM);
		if(strcmp(logFile,"stderr")==0 && (logFile[6]==':' || logFile[6]==0) )
			loggerM = stderrM;
		else if(strcmp(logFile,"stdout")==0)
			loggerM = stdoutM;
		else
			loggerM = fopen(logFile,"w");
		fdlogM = ::fileno(loggerM);
	}
	if(theme!=-2)
		logThemesM = theme;
	return !!loggerM;
#else
	return false;
#endif
}

CTL_Inline ctl::string getLANid()
{ return env().getThreadId().toString(); }
CTL_Inline ctl::string getLANidS(const ctl::string& client, ctl::string& loc)
{ return env().getLANid(client, loc); }

//{ return env().getRootId().c_str(); }

CTL_Inline TimeMarker &newTimeMark(const char *name)
{ return env().newTimeMark(name); }

CTL_Inline const char *getcwd()
{ return env().pathM.c_str(); }

CTL_Inline int  system(const char *cmd)
{ return sys::system(cmd, env().pathM); }

CTL_Inline FILE *fopen(const char *file, const char *mode)
{ return ::fopen( sys::absolutePath(env().pathM, file).c_str(), mode); }

CTL_Inline RemoteSP getNamedObject(const std::pair<ctl::typeId, ctl::string>& key)
{ return env().getNamedObject(key); }

CTL_Inline void setNamedObject(const std::pair<ctl::typeId, ctl::string> &key, const RemoteSP &sp)
{ env().setNamedObject(key, sp); }

CTL_Inline const char* getVersion()
{ return CTL_VersionStr; }

CTL_Inline int getPID()
{ return sys::getPid(); }

CTL_Inline bool setUsing(const ctl::typeId &ciType, const ctl::string &loc)
{
//  Could be done on demand
	ctl::link p(loc);
	if(!p)
		return false; 
	setLink(ciType, p);
	return true; 
} 

CTL_Inline PID getRootId(Environment *e)
{ return e->getRootId(); }
CTL_Inline PID getThreadId(Environment *e)
{ return e->getThreadId(); }
CTL_Inline const link &getRemoteLink(Environment *e, const PID &pid)
{ return e->getRemoteLink(pid); }
CTL_Inline  bool sameThread(Environment *a, Environment *b)
{ return a->sameThread(b); }  

CTL_Inline int ctl_main(Environment *myEnv)
{	
	int ok=0;
	char runMode = myEnv->runModeM;
	if(runMode == 'e')
	{
		vector<vector<char> > args = split(myEnv->argStringM, ' ', '[',  ']');
		int nargs= args.size() ;
		vector<char*> cargs(nargs +1);
		for(int i=0; i<nargs; i++)
			cargs[i] = args[i].begin();
		CTL_Try
		{ ok = int(ctlCI::CTL_Main(nargs, cargs.begin() ) ); }
		CTL_Catch(ctl::exception &e)
		{ ok = 1; log(myEnv,"#m !! int ctl::main caught excpetion: %s\n", e.what()); }
	}
	ctl::vector<ctl::string> sym = myEnv->getSymbols();
	if(runMode == 'A')
	{
		for(size_t i=0; i<sym.size(); i++)
				ctl::cout<<sym[i]<<"\n";
		return 1;
	}
	if(runMode == 'I')
	{
		ctl::vector<ctl::string> intrinsic=myEnv->getIntrinsicSymbols();
		for(size_t i=0; i<sym.size(); i++)
			if(intrinsic.position(sym[i])<0 && !strchr(sym[i].c_str(),'(') )
				ctl::cout<<sym[i]<<"\n";
		return 1;
	}
	if(runMode == 'H')
	{
		ctl::cout<<"location info"<<"\n";
		return 1;
	}
	return ok;
}

CTL_Inline void openLibrary(void ** vmyenv, void *vparentenv, void *proc, const char *arg, const char* (*locator)(), void(*conn)())
{
	ctl::Environment *parentEnv=(ctl::Environment*)vparentenv, *myEnv;
// threadId != 0 <==> thread
// always parentEnv != 0
	const ctl::link &p= *(ctl::link*)proc;
	myEnv = ctl::createEnv(parentEnv, p.getIType());
	const char *args[] = { arg, 0 };
	myEnv->init(parentEnv, 1, args);
	FILE* out=myEnv->stderrM;

	new ctl::groupImpl(1, args, true);
	ctl::messageInfo(out);
	ctl::initEnv(myEnv, p, locator ? (*locator)() : (const char*)0, conn);
	if(conn)
		(*conn)();
	*vmyenv = myEnv;
	int ret=ctl_main(myEnv);
	parentEnv->clearChildEnv();
	int read = 1;
	Wait W;
	if(ret == 0 && p.getIType()==ctl::thread)
		while(myEnv->run(read=1, false) )
			W(read==0);;
	return;
}

CTL_Inline int openMain(int n, char **arg, void ** vmyenv, void *vparentenv, const char* (*locator)(), void(*conn)())
{
#if !defined CTL_AllowRoot && defined CTL_Unix
	if(::geteuid()==0)
	{
		std::cerr<<" CTL components may not be started as root!\n";
		exit(1);
	}
#endif
	ctl::Environment *parentEnv = (ctl::Environment*)vparentenv, *myEnv=ctl::createEnv(parentEnv, undefined);
	myEnv->init(0, n, arg);
	if(!vparentenv) //otherwise forked
		::atexit(&Environment::global::del);
	ctl::group grp(n, arg, true, true);
	FILE* out=myEnv->stderrM;
	char line[CTL_StringBufferSize];
	if(!grp)
	{
		if(2*strlen(arg[0]) > CTL_StringBufferSize)
			return 1;
		sprintf(line, " group creation failed\n  *** %s (%s) terminated  *** \n", arg[0], arg[1]);
		ctl::messageInfo(out, line);
		return 0;
	}
	ctl::messageInfo(out);
	ctl::initEnv(myEnv, grp[-1], (*locator)(), conn, &grp );
//	ctl::initEnv(myEnv, 0, grp[-1], (*locator)(), conn );
	if(conn)
		(*conn)();
//fprintf(stderr, "LOGID=%d connector = %p\n", int(grp.logId()),conn);
	int ret=grp.logId()==0 ? ctl_main(myEnv) : 0;
	if(ret==0)
		while(grp.run() );
	
	sprintf(line,  "  *** %s (%s) terminated normally *** \n",arg[0], arg[1]);
//	grp.setError(line);
	ctl::messageInfo(out, line);
//	ctl::Environment::global::del(myEnv);
	*vmyenv = myEnv;
	return 0;
}

CTL_Inline void closeLibrary(void *venv, bool unset)
{ ctl::Environment::global::del((ctl::Environment*)venv, unset); }

} //  ctl



#endif


     
