/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_IO_H_
#define CTL_IO_H_

#include <ctl/utl/timer.h>
#ifdef CTL_MemMap
#include <sys/mman.h>
#endif
#include <ctl/system.h>
#include <ctl/inl/systypes.h>

namespace ctl
{
 struct MessHeader
 {
	enum { fixsize = 38 };
	ctl::int8 size;
	ctl::int4 tag;
	ctl::PID pid;
	ctl::int4 logId;
	bool swap;
  MessHeader(): size(-1), tag(-1), pid(-1),logId(-1),swap(false) {}
  MessHeader(const MessHeader &m): size(m.size), tag(m.tag), pid(m.pid), logId(m.logId),swap(m.swap)   {}
  MessHeader(int s, int t, const PID & l, int lid=-1) : size(s), tag(t), pid(l), logId(lid),swap(false)  {}
	oStream &write(oStream &os) const
	{ return os<<uint2(1)<<size<<tag<<pid<<logId; }	
	iStream &read(iStream &is)
	{
		uint2 swp;
		is>>swp;
		if(swp>1)
		{ 	swap = true;   is.setSwap();  }
		return is>>size>>tag>>pid>>logId; 	}	
};

namespace internal
{

  struct ioBuffer : public MessHeader
  {
    byteBuffer buffer;
		ctl::int8 no;
    ioBuffer(): no(-1) {}
 //   ioBuffer(const MessHeader &h, byteBuffer &is): MessHeader(h), buffer(is), no(-1) {}
    ioBuffer(int s, int t, const PID & l, int lid): MessHeader(s, t, l, lid) ,no(-1) {}
    ioBuffer(int s, int t, const PID & l, int lid, oStream &os): MessHeader(s, t, l, lid) ,buffer(os), no(-1) {}
		void setData(const void *dat, size_t s)
		{ buffer = oStream(s, (void*)dat); size = s; }
    void setNo(ctl::int8 n)
	 	{ no = n; }
    bool operator !() const
    { return buffer.size()==0 && !pid; }
		bool operator <(const  ioBuffer &m) const
		{
			if(tag!=m.tag)
				return tag<m.tag;
			int c=pid.cmp(m.pid);
			if(c)
				return c<0;
			return no<m.no;
		}
  };
} //  internal


class ioBase: public shared
{
public:
	virtual bool write(void *buf, int size) const=0;
	virtual size_t read(void *buf, int size) const=0;
	virtual bool write(int8 size, const vector<sys::iovec> &p) const =0;
	virtual bool operator!() const =0;
	virtual bool close() const=0;
	virtual bool test(int usec) const=0;
	virtual char *memmap(size_t) const
	{ return 0; }
	bool writeHd(const internal::ioBuffer &m) const;
	bool readHd(internal::ioBuffer &m, bool check=false) const;
};

class File: public ioBase
{
	mutable int fdM;
	bool isSocketM;
public:
	File(int fd=-1, bool isSocket =true): fdM(fd), isSocketM(isSocket) {}
	operator int() const
	{ return fdM; }
  bool operator !() const
	{ return fdM<0; }
	bool close() const;
	bool write(int8 size, const vector<sys::iovec> &p) const;
	bool write(void *buf, int size) const;	
	size_t read(void *buf, int ssize) const;
  bool test(int usec) const;
};


class socket: public ioBase
{
enum { openE, bindE, listenE, hostE };
	PID pidM, rPidM;
	mutable File fdM;
	mutable int errorM;
	sys::sockBool yesM;
	static bool getLocalAddr(sockaddr_in &sa, unsigned short port, const char *host=0);
	void setPort();
public:
#ifdef CTL_MemMap
	char *memmap(size_t size) const;
#endif
	void setRPid(const PID &rpid)
	{ rPidM = rpid; }
	PID getRPid() const
	{ return rPidM; }
	int getFD() const
	{ return int(fdM); }
	ctl::string getIP() const
	{ return pidM.host; }
	PID getPid() const
	{ return pidM; }
	socket(int fd=-1): fdM(fd), errorM(0)
	{ setPort(); }
	bool write(int8 size, const vector<sys::iovec> &p) const
	 { return fdM.write(size, p); }
	bool write(void *buf, int size) const
	{ return fdM.write(buf,size); }
	size_t read(void *buf, int size) const
	{ return fdM.read(buf,size); }
   bool test(int usec) const
	{ return fdM.test(usec); }
	unsigned int getPort() const
	{ return pidM.port; }
	bool operator !() const
	{ return fdM==-1 || errorM; }
	bool open(unsigned short port=0, bool blocking = true, const char * host =0);
	bool connect(unsigned short port, const char *host, bool fwd=false);
	bool close() const;
	PID accept(socket &acc, bool blocking = true, int usec =0) const;
	~socket()
	{ close(); }
	void release()
	{ fdM = -1; }
};

// primitiv io devices
class device
{
public:
	virtual int write(const char *m, size_t count)=0;
	virtual int read(char *m, size_t count)=0;
	int readln(char *m, size_t max)
	{
 		for(char c=0; c!='\n' && --max;)
			if(read(&c, 1)<0)
				return -1;
			else
				*m++ = c;
		return *m = 0;
	}

	virtual ~device()
	{}
};

class fileNoDevice: public device
{
protected:
	int inM, outM;
public:
	int getIn()
	{ return inM; }
	int getOut()
	{ return outM; }
	fileNoDevice(int in=0, int out=1): inM(in), outM(out)
	{}
	int write(const char *m, size_t count)
	{ return ::write(outM, m, count); }
	int read(char *m, size_t count)
	{ return ::read(inM, m, count); }
};

} // ctl

#endif
