/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_LINK_H_
#define CTL_LINK_H_


#include "communicator.h"
#include "utl/factory.h"

namespace ctl
{

struct soHandle;
struct messageForward;

//!  This class creates, holds and frees components
class link
{
friend class Environment;
friend class groupImpl;
friend struct soHandle;

	groupSP grpM;
	CommunicatorSP communicatorM;
	PID pidM;
	ioTag tagM;
	channelT linkageM;
	ctl::any propertyM;
	sys::threadIdT childTidM;
	typedef RemoteSP symbolT;
	ctl::lexicalAssoc<char, symbolT> symbolTableM;
// just be an owner of ownedLinkM
	lPointer<link> ownedLinkM;
	ctl::string errorM; 
	void init(bool owner = true);
	link(int n, const char * const *argv);
	link(const PID & pid, const PID &logId=PID(), const char *args=0);
	soHandle *getHandleRaw() const
	{ return ((soHandle*)communicatorM.get()); }
	soHandle *getHandle() const
	{
		if((linkageM==thread || linkageM==lib) && communicatorM->isValid())
			return ((soHandle*)communicatorM.get());
		return 0;
	}
	void init(const location &loc, channelT typ);
	void initStr(const char *loc, channelT typ, const char *mod=0);
	void setChildEnv(const Environment*) const;
public:
struct envSet
{
	const link *pM;
	channelT typM;
	envSet(const link *p, channelT typ=undefined): pM(p), typM((typ==undefined && p) ? pM->getType() : typ)
	{	
		if(typM==lib && pM)
			pM->setEnv(true); 
	}
	~envSet()
	{
		if(typM==lib && pM)
			pM->setEnv(); 
	}
};
	void setEnv(bool = false) const;
	ioTag getTag() const
	{ return tagM; }
	ioTag getNewTag() const
	{ return communicatorM->getTag(); }
	void setTag(ioTag tag)
	{ tagM = tag; }
	  RemoteSP getFunction(const char *pretty) const;
	void addFunction(const char *fname, const RemoteSP &f);
	const PID &getPid() const
	{ return pidM; }
	string getLocation() const
	{ return (linkageM==dmn || linkageM==tcp) ? pidM.host + " -l dmn " + string::int2string(pidM.port) : string(); }
	bool isOnLocalhost() const
	{	return location::isLocalHost(pidM.host); }
	const char* getModifier() const
	{ return linkageM == modifier ? errorM.c_str() : (const char*)0; }
	const CommunicatorSP &getCommunicator() const
	{ return communicatorM; }
	const groupImpl *getGroup() const
	{ return !communicatorM ? 0 : communicatorM->getOwner(); }
	const any &getProperty() const
	{ return propertyM; }
	template<typename T> const T *getProperty(const T* t=0) const
	{
		if(!propertyM)
			return 0;
		return propertyM.get(t);
	}
	channelT getIType() const
	{ return linkageM; }
	channelT getType() const
	{ 
		if(linkageM==thread || linkageM==lib)
			return (!childTidM || childTidM==sys::getTid() ) ? lib : thread;
	
		return linkageM; 
	}
	void setForward(messageForward* =0);
	
	const char *error() const
	{ return errorM.c_str(); } 

	ctl::vector<ctl::string> getSymbols() const;
//	link swap(const ctl::link &p);
//!  no ressource allocation
	  link(channelT  linkage=undefined);
	  link(const CommunicatorSP& commu, const PID & target =-1,ioTag type = RMI, bool owner = true);
//! Takes a property for the location selector
		explicit link(const ctl::any &property);
//! if the channelT is file loc is interpreted as an filename, otherwise as a location
	  link(const char *loc, channelT typ =undefined, const char *mod =0);
	  explicit link(const ctl::string &loc,  channelT typ =undefined, const char *mod =0);
	  explicit link(const std::string &loc,  channelT typ =undefined, const char *mod =0);
	  link(const link &l, const location &loc);
//! create a link using the given location and the specified channelT
	  link(const location &loc, channelT typ= undefined);
	  ~link();
	  bool operator==(const link &p) const
	{ return communicatorM==p.communicatorM && pidM==p.pidM; }
	  bool operator!=(const link &p) const
	{ return !((*this)==p); }
	bool closed() const;
	bool isValid() const
	{ return !!communicatorM && communicatorM->isValid(); }
	void release(bool s/*strong*/ =true)
	{ communicatorM.release(s); grpM.release(false); }
	void derelease()
	{ communicatorM.derelease(); grpM.derelease(); }
	bool callNext(bool blocking = true) const;
	bool run(int n = 0, bool blocking = false) const;
//! check wether the proces is still valid
	bool operator !() const;
	int test()
	{
		ioTag tag=ctl::DAT;
		return test(tag);
	}
	int test(ioTag &tag)
	{
		int typ = tag;
		PID src =  getPid();
		int size = getCommunicator()->test(src, typ);
		tag = typ;
		return size;
	}
	bool addJob(void(*fnc)(void*), void *recv) const;
	status writeMessage(const ctl::oStream &os, int tag=-1)  const;
	template<class T> inline const ctl::link &operator<<(const T& t) const
	{
		ioTag tag=tagM.tag<RMI ? tagM.tag : DAT;
		if(!communicatorM)
			ctl::throwStr(" *** fatal io error write by invalid link ***");		

		ctl::nWrite(communicatorM, t, getPid(), tag);
		return *this;
	}
	template<class T> inline const ctl::link &operator>>(T& t) const
	{
		ioTag tag=tagM.tag<RMI ? tagM.tag : DAT;
		ctl::nRead(communicatorM, t, getPid(), tag);
		return *this;
	}	
};


} // ctl

#endif



