/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_REMOTE_H_
#define CTL_REMOTE_H_

#include "utl/typefield.h"
#include "utl/serialize.h"
#include "utl/pointer.h"

#include "binaryType.h"
#include "system.h"
#include "utl/functiontraits.h"

namespace ctl
{

/**/// class objectID identifies remote objects
class objectID
{
	union objT { Remote *p; uint_ id; };  
	uint8 idM;
public:
	static char * ptr2str(const void *p, char *str)
	{
		objT obj; obj.p = (Remote*)p;
		str +=  2 * sizeof(Remote*);
		*str = 0;	
		uint_ Z = obj.id;	
		for(unsigned int i=0; i<2 * sizeof(Remote*); i++, Z>>=4)
			*--str = 'a' + (Z&15);
		return str;
	}
	static void *str2ptr(const  char *str)
	{
		uint_ Z=0;
		while(*str)
		{			
			Z <<= 4;
			Z += *str++ -'a';
		}
		objT obj; obj.id = Z;
		return obj.p;
	}
	CTL_Type(ctl::objectID, tupel, (idM),1);
	objectID(const Remote *p=0)
	{ objT obj; obj.p = (Remote*)p; idM = obj.id; }
	operator Remote*() const
	{ objT obj; obj.id = uint_(idM); return obj.p; }
	bool operator !() const
	{ return idM ==  0; }
};

enum { io, execpt, all=0xffff, _=-1 };
typedef enum { DAT = 1, OPER , RMI, CONTROL, ERR, now, late, local2, remote, UNDEF = 65535 } Flag;
typedef enum { success=0, retry, failed, unknown } status;
struct MID
{
	ctl::int2 idM;
	typeId strM;
	CTL_Type(ctl::MID, tupel, (idM, strM), 2);
	explicit MID(ctl::int2 id = -1, const typeId &str=typeId()): idM(id), strM(str) {}
	operator int() const
	{ return idM; }
};
//! The supported channels
// See also channel2string & string2channel!!!
typedef enum { file =0, serial=1,
							mpi=2, lam=3, pvm=4, thread=5,
							tcp=6, lib=7, dmn=8, ssh = 9, pipe=10, ext=11, cnct = 12, logid = 13,
							modifier, defaultChannel = tcp, local=17, undefined=62, keepAsIs,  invalid  } channelT;
typedef enum { named=0, typed } namedFlag;
inline oStream &operator <<(oStream &os,const ctl::channelT &x)
{ char c=x; return os.write(&c,1); }
inline iStream &operator >>(iStream &is,ctl::channelT &x)
{ char c; is.read(&c,1); x=(ctl::channelT)(c); return is;  }


//! struct location: a link location descriptor
struct location
{
	enum { separator = '$' , brasep= '[', ketsep=']', comma=';' };
	ctl::string ciName;
	ctl::string exec,	path, host, user;
	ctl::string shell, term, debug;
	channelT linkage;
	ctl::string args, localArgs;
	ctl::string log;
	ctl::string wd;
	int4 portL, portR;
	bool cypher, verbose, replyDstr;
	int4 sPort;
	int4 logId, groupSize, priority;
	ctl::string parent;
	ctl::string errorStr;
	double timeout;
	char runMode;
	ctl::string  argString;
	mutable ctl::string sshString;
	ctl::string envFile, varStr;
	ctl::string pipeTail;
	char pipeType;
	void setLinkage(channelT link)
	{ linkage = link; }	
	static ctl::string compose(int n, const char * const * arg)
	{
		ctl::string ag;
		for(int a=0; a<n && *arg[a] && *arg[a] != '#'; a++)
		{	ag += arg[a]; ag += " "; }
		return ag;
	}
	static ctl::string getBraced(char*, bool = false);
	void setArgs(const ctl::string &arg);
	void setArgs(int n, const char * const * arg)
	{	setArgs(compose(n, arg) ); 	}
	
//!   assign a terminal
	void setTerminal(const ctl::string &trm= ctl::string())
	{
		if(trm.empty())
			term = CTL_Str(CTL_Terminal);
		else if(trm[0]==' ')
		{	term =" "; return; }
		else
			term = trm + " ";
		if(term.substr(0,5) =="xterm" || term.substr(0,7) =="konsole" )
			term += " -e ";
	}
	bool hasTerminal() const
	{ return !!term && term[0] != ' ' && term!="none" && !noTerminal(); }
	bool notByTerminal() const 
	{ return linkage == dmn || linkage == lib || linkage==thread || linkage == pipe || (linkage==tcp && portR<0); }
	bool noTerminal() const
	{ return linkage == dmn || linkage == pipe || (linkage==tcp && portR<0); }
	void setUser(const ctl::string &use)
	{ user = use; }
//! assign a debugger
	void setDebug(const ctl::string &dbg = ctl::string())
	{
		if(dbg.empty())
		{	debug = CTL_Str(CTL_DBG); debug += "  "; }
		else
			debug = dbg + " ";
	}
	static bool checkSupport(channelT);

	const char *getChannelString() const
	{ return channel2string(linkage); }
   void init(const  ctl::string &);

	location() : linkage(tcp), portL(0), portR(0), cypher(false), verbose(false), replyDstr(false), sPort(0), logId(-1), groupSize(0),priority(0),timeout(0), runMode(0), pipeType(0)
	{ host = localHost(); }

//! Constructors accepting a string in ssh syntax
	location(const char *locC, channelT ch = undefined) : 
		linkage(ch), portL(0), portR(0), cypher(false),verbose(false), replyDstr(false), sPort(0), logId(-1), groupSize(0), priority(0), timeout(0), runMode(0), pipeType(0)
	{ 
		init(locC);
		setup(ch);
	}
	location(int n, const char *const* args, channelT ch = undefined): 
		linkage(ch), portL(0), portR(0), cypher(false),verbose(false), replyDstr(false), sPort(0), logId(-1), groupSize(0), priority(0), timeout(0), runMode(0), pipeType(0)
	{ 
// only remotly called
		init(compose(n,args));
		if(!portL && !portR && linkage == tcp)
			linkage = dmn;
		setup(ch);
	}
	void setup(channelT);
	void clear()
	{		exec.clear(); path.clear(); host.clear(); }
	location(const ctl::string &locStr, channelT ch = undefined): 
		linkage(ch), portL(0), portR(0), cypher(false), verbose(false), replyDstr(false), sPort(0), logId(-1), groupSize(0),priority(0), timeout(0), runMode(0), pipeType(0)
	{ 
		init(locStr);
		setup(ch);
	}
	void set();
//! constructor assigning the executable, the path, and optionally the host  (default locahost) and the channelT (default tcp)
	location(const ctl::string &e, const ctl::string &p, const ctl::string &h=ctl::string(), channelT ch= tcp):
			exec(e), path(p), host(h), linkage(ch), portL(0), portR(0), cypher(false), verbose(false), replyDstr(false), sPort(0), logId(-1), groupSize(0),priority(0), timeout(0), runMode(0), pipeType(0)
	{ set();	}
	location(const ctl::string &e, const ctl::string &p, channelT ch): exec(e), path(p), linkage(ch), portL(0), portR(0), cypher(false), verbose(false), replyDstr(false), sPort(0),priority(0), timeout(0), runMode(0), pipeType(0)
	{ set();	}

	location(ctl::uint4 port, const ctl::string &h=ctl::string(), channelT type =dmn):
				host(h), linkage(type), portL(port), portR(0), cypher(false), verbose(false), replyDstr(false), sPort(0), logId(-1), groupSize(0),priority(0), timeout(0), runMode(0), pipeType(0)
	{ char p[8]={'\0'}; ; sprintf(p,"%d",port); exec=p; set();	}
	const char *error() const
	{ return errorStr.c_str(); }
	bool operator !() const
	{ return (ciName.empty() && exec.empty() ) || !errorStr.empty(); }
	ctl::string absolutePath() const;
	ctl::string pathExec() const
	{
		if(exec=="*")
			return ctl::string(); 
		return !path ? exec : path + "/" + exec; 
	}
	ctl::string absolutePathExec() const;
	ctl::string getSSH() const;
	ctl::string str() const
	{ return getSSH(); }
	const char *c_str() const
	{ 
		if(!sshString)
			sshString = getSSH();
		return sshString.c_str();
	}
	inline friend std::ostream& operator<<(std::ostream& os, const location& loc)
	{ return os<<loc.c_str(); }
	static const char *channel2string(channelT ch)
	{
		const char *str[] = { "file","serial","mpi","lam","pvm","thread","tcp",
													"lib","dmn","ssh","pipe", "ext" ,"cnct", "logid" };
		return ch<14 ? str[ch] : "undefined";
	}
// parser
	static char *seperate(char *str, char *&left, char *&right, char c, bool back);
	bool pUser(const char*);
	bool pHost(const char*);
	bool pPath(const char*);
	bool pArgs(const char*);

	static ctl::string shortHostName(const ctl::string &h)
	{
		if(!!h && isdigit(h[0]) )
			return h;
		ctl::string hloc=(h=="localhost") ?localHost() : h;
		const char *s=hloc.c_str(), *d=strchr(s, '.');
		return d ? hloc.substr(0,d-s) : hloc;
	}
	static bool equalHost(const ctl::string &h0, const ctl::string &h1)
	{  return shortHostName(h0)==shortHostName(h1); }
	static bool isLocalHost(const ctl::string &host);
	static bool isLocalHostName(const ctl::string &host);
	static bool isLocalAreaName(const ctl::string &host);
	  static ctl::string localHost();
	static ctl::string localPath()
	{ return ctl::getcwd();	}
	static void substSlash(char *p)
	{
#ifdef CTL_WinPath
		while(p=strchr(p,'\\'))
			*p = '/';
#else
		(void)p;
#endif
	}

	void print(std::ostream &) const;	
};

	inline std::ostream& operator<<(std::ostream& os, channelT ch)
	{ return os<<location::channel2string(ch); }


struct locationSelectBase
{
	virtual ctl::string operator()(const typeId&, const ctl::any&) const=0;
	virtual ~locationSelectBase() {}
};

template<class  locator> struct locationSelector : public locationSelectBase
{
	locator impl;
	locationSelector(const locator &imp) : impl(imp) {}
	virtual ~locationSelector() {}
	ctl::string operator()(const typeId &type, const ctl::any &a) const
	{	
		ctl::result<ctl::string> loc(impl(type, a) );
		return (ctl::string)loc;
	}
};
CTL_Inline void setLocatorBase(const ctl::lPointer<locationSelectBase> &);


template<class  locator> inline void setLocator(const  locator &impl)
{ setLocatorBase(ctl::lPointer<locationSelectBase>(new locationSelector<locator>(impl) ) ); }

inline ctl::oStream &operator << (ctl::oStream &os, const ctl::status &res)
{ return os<<char(res); }

inline ctl::iStream &operator >> (ctl::iStream &is, ctl::status &res)
{ char c; is>>c; res = ctl::status(c); return is; }

} // ctl

#include "result.h"
#include "link.h"

namespace ctl
{

namespace internal
{

template<class Y> inline const Y &getValue(const Y &y, void*)
{ return y; }
template<class Y, class YY> inline const YY &getValue(const result<Y> &y, YY*)
{return ((const  resultBase<Y>&)y).getValue(true); }

template<class Y, class YY> inline const YY &getValue(const rBinResult<Y> &y, YY*)
{ y.excptM.clear(); return ((const  resultBase<Y>&)y).getValue(true); }

	struct onlyRef {};
	struct argType {};
	struct unrefConst{};
	struct constUnRef {};
	struct binSelect {};
	struct binCast {};
	struct hasBinary {};
	struct binIArg {};
	struct noRef {};
}

} // e.o. ctl

namespace ctl
{

	template<> struct typeMap<ctl::internal::onlyRef>
	{
		template<class T> struct map
		{
			enum { ref = !ctl::constTraits<T>::is };
			typedef  typename ctl::select<ref , typename ctl::binArg<T>::resultT, ctl::empty>::resultT resultT;
		};
	};
	template<> struct typeMap<ctl::internal::argType>
	{
 	 	template<class T> struct map
		{ typedef  typename ctl::optArgType<T>::resultT resultT; };
  };
	template<> struct typeMap<ctl::internal::unrefConst>
	{
		template<class T> struct map
		{ typedef  typename ctl::binArg<T>::thisT  resultT; 	};
	};


	template<> struct typeMap<ctl::internal::constUnRef>
	{
		template<class T> struct map
		{
//			enum { isConst =ctl::constTraits<T>::is || !fundamental<T>::is };
			enum { isConst =ctl::constTraits<T>::is  };
			enum { bin=sizeof(isBinary((const T*)0) )>1, fnd = bin || fundamental<T>::is }; 
			typedef typename ctl::select<isConst || !fnd, T, typename ctl::binArg<T>::resultT>::resultT  resultT; 	
		};
	};

	template<> struct typeMap<ctl::internal::binSelect>
	{
		template<class Y, class X> struct map2
		{	typedef typename ctl::binarySelect<Y,X>::resultT resultT; };
	};

	template<> struct typeMap<ctl::internal::binCast>
	{
		template<class Y, class X> struct map2
		{
			enum { XisRef = sizeof(constCheck((X*)0) )==1, isBin=sizeof(isBinary((const X*)0 ) )>1 };

			enum { YisRef = !ctl::constTraits<Y>::is && ctl::referenceTraits<Y>::is };
			typedef typename noReference<Y>::resultT Y_;
			typedef ctl::binaryCaster<X, Y,Y_,int2type<YisRef> > binCast;
			typedef typename ctl::select<!isBin, X&,
				typename ctl::select<XisRef,typename binCast::temporary,
											typename binCast::constTemporary>::resultT>::resultT resultT;
		};
	};
	template<> struct typeMap<ctl::internal::hasBinary>
	{
		template<class T> struct map
		{
			enum { isBin = ctl::isBinaryRep<T>::value };
			typedef typename ctl::select<isBin,typename ctl::binArg<T>::resultT, ctl::empty>::resultT resultT; };
	};
	template<> struct typeMap<ctl::internal::binIArg>
	{
		template<class T> struct map
		{ typedef typename ctl::binArg<T>::resultT resultT;  };
	};
	template<> struct typeMap<ctl::internal::noRef>
	{
		template<class T> struct map
		{ typedef typename ctl::noReference<T>::resultT resultT;  };
	};

};

namespace ctl
{


struct defaultCall
{
	enum { isEmpty = true };
	static void connect() {}
	virtual ~defaultCall() {}
//	virtual ctl::string print() const
//	{ return "unknown function"; }
	template<class Y, class F, class Arg> Y call(const F &f, const Arg &x, Y*, ctl::cPointer<ctl::excptHandle>& excpt, const char *pretty)
	{
//		logger lg(1, "<= ");
//		lg << pretty;		
		CTL_Try
		{ return f(x); }
		CTL_Catch(const ctl::exception &e)
		{ excpt = new ctl::excptHandle; rethrowStr2bin(e, excpt->binM, " *** in:  %s ***\n", pretty); 	}
		CTL_Catch(const ctl::groupException &e)
		{ CTL_Throw(e); 	}
		CTL_CatchAll
		{ excpt = new ctl::excptHandle; throwStr2bin(excpt->binM, " *** unknown exception in:  %s ***\n", pretty); 	}
		return Y();
	}
};

template<class RFICall> struct rFunction;


struct cloner
{
	static void clone(const void *, falseT)
	{}
	static void clone(binaryType *, falseT)
	{}
	static void clone(const binaryType *b, falseT)
	{ b->clone(); }
	template<class CI> static void clone(const CI *ci, trueT)
	{
		CTL_Try
		{ ci->CTL_getCIPThrows(false); }
		CTL_Catch(const ctl::exception &e)
		{ ctl::rethrowStr(e," *** can't write empty object %s ***",  CI::CTL_typeNameS().c_str() ); }
	}
	template<class X> void operator()(X &x) const
	{
		enum { isCI=sizeof(CTL_isCI((X*)0,(ctl::empty*)0) )>1 };
		clone(&x, int2type<isCI>());
	}

};

template<class oY, typename ciArgs, class Caller>  class RMICall
{
	typedef RMICall<oY, ciArgs, Caller> thisT;
	template<class T> static status returnStatus(const T&)
	{ return success; }
	static status  returnStatus(status res)
	{ return res; }
	static void clearBuffer(void*, ctl::iStream &is)
	{ is.clear(); }
	static void clearBuffer(status*, ctl::iStream &is)
	{;
		if(is.copyBuffer())
			is.clear();
	}
	typedef typename ctl::fieldMap<ciArgs,internal::binIArg>::resultT args;
	typedef typename ctl::fieldMap<ciArgs,internal::unrefConst>::resultT urcArgs;
	typedef typename ctl::fieldMap<ciArgs,internal::hasBinary>::resultT binArgs;

public:
	typedef ciArgs inArgs;
	typedef typename ctl::fieldMap<ciArgs,internal::onlyRef>::resultT returnArgs;
	typedef typename ctl::fieldMap<ciArgs,ctl::internal::constUnRef>::resultT curArgs;
	typedef args origArgs;
	typedef args optArgs;

	typedef typename ctl::returnType<oY>::Y origY;
	typedef typename resultType<origY>::resultT resY;
	typedef typename resY::valueT  retY;
	typedef typename ctl::noVoid<oY>::resultT  Y;

	typedef Caller origCall;
	enum { 	retEmpty = ctl::same<ctl::empty, Y>::value || ctl::same<status, Y>::value,
					argEmpty = returnArgs::isEmpty,
					isEmpty = retEmpty && argEmpty && Caller::isEmpty };

private:
	template<typename Func>
		static inline status call_(Func f, ctl::iStream &is,  ctl::trueT)
	{
		urcArgs arg;
		ctl::int4 tag;
		is>>arg>>tag>>ctl::iStream::eofTest();
		clearBuffer((retY*)0, is);
		retY y(internal::getValue(f(arg), (retY*)0) );
		return returnStatus(y);
	}
	template<typename Func>
		static inline status call_(Func f, ctl::iStream &is, ctl::falseT)
	{	
		urcArgs arg;
		typename resY::RP rp;

		ctl::int4 tag;
		is>>arg>>rp>>tag>>ctl::iStream::eofTest();
		clearBuffer((retY*)0, is);
		typedef typename Func::resultT fres;
		resY r=f(arg);
		binaryObject bin;
		if(!!r.excptM)
		{
			bin = r.excptM->binM;
			r.excptM->binM.clear();
		}
		retY y = r.getValue(true/*logOff*/); 	
		if(returnStatus(y)!=retry)
			rp->send(bin, y , returnArgs(arg), tag);
		return returnStatus(y); 
	}
	static inline resY send_(const char *pretty, const MID &mid, const optArgs &arg, ctl::trueT, const ctl::link &p, objectID trgt)
	{
		char buf[CTL_defaultBufferSize];
		ctl::uint4 rtag = RMI;	

		if(!!p)
			p.writeMessage(ctl::serializeB(trgt, mid, arg, rtag, buf, CTL_defaultBufferSize,p.getCommunicator()) );
		resY res;
		res.setPretty(pretty);
		return res;
	}
	static inline resY send_(const char *pretty, const MID &mid, const optArgs &arg, ctl::falseT, const ctl::link &p, objectID trgt)
	{
		ctl::uint4 tag = p.getNewTag();
		returnArgs rargs(arg);

		ctl::link p2(p);
		p2.setTag(tag);
		resY rRes(p2,rargs, pretty); 		

		typename resY::RP rp=rRes.getRP();
		char buf[CTL_defaultBufferSize];		
  		
		status ret = failed;
		if(!!p)
		{
			CTL_Try
			{  ret = p.writeMessage(ctl::serializeB(trgt,mid,arg,rp,tag, buf,CTL_defaultBufferSize,p.getCommunicator()) );	}
			CTL_Catch(const ctl::exception &e)
			{ return resY(retY(), new excptHandle(e), pretty); }
			CTL_CatchAll
			{  	CTL_Throw;	 }
		}
		return ret == failed ? resY(retY(),  new excptHandle(ctl::exception(" *** send failed ***") ), pretty ) : rRes;
	}
public:
	enum { hasBinRep = !binArgs::isEmpty };

	template<class Functor>
		static inline status call(ctl::trueT, Functor f, ctl::iStream &is)
	{
		typedef typename Functor::args fArgs;
		typedef typename ctl::fieldMap<fArgs, ctl::internal::noRef>::resultT mArgs;
		typedef typename ctl::fieldMap2<ciArgs, mArgs, ctl::internal::binSelect>::resultT nArgs;
	
		return RMICall<origY,nArgs,Caller>::call(ctl::falseT(),f,is);
	}
	template<typename Functor>
		static inline status call(ctl::falseT, Functor f, ctl::iStream &is)
	{	return call_(f, is, ctl::int2type<isEmpty>()); }

	template<class C> static resY send(const C &c, const MID &mid, const optArgs &arg, ioTag tag)
	{
		objectID trgt=0;  	PID pid=-1;
		ctl::link p = c.checkRP(mid, trgt, pid);
		p.setTag(tag);
		
		if(!p)
			return resY();	
		return send_((const char*)0 ,mid, arg, ctl::int2type<isEmpty>(), p, trgt);
	}

	static resY  send(const ctl::link &p, const char *pretty, const optArgs &arg)
	{	return send_(pretty, MID(0, pretty), arg, ctl::int2type<isEmpty>(), p,  objectID()); 	}

	template<class Functor> struct tempType
	{
		typedef typename Functor::args outArgs;
		typedef typename ctl::fieldMap2<outArgs,inArgs, internal::binCast>::resultT resultT;
	};

	template<class Functor>
		static retY call(ctl::trueT, Functor f, const optArgs &y)
	{
		typedef const typename tempType<Functor>::resultT args;
		return retY(f(args(y))); 	
	}

	template<class fid, class arg>
		static resY functionCall(const arg &x, const char *pretty, fid*, const typeId &citype, ctl::link &p)
	{	
		if(p.getType()==ctl::invalid)
			ctl::throwStr(" *** invalid link in %s ***", pretty );
		typedef  retY vY;
// try to get a process
		rFunction<thisT> *fobj = 0;

// look only for link by call of use or by environment
		if(!p)
			p = ctl::getLink(citype, ctl::any(), "");
		if(!p && !(fobj=(ctl::rFunction<thisT>*)(ctl::getFunction(pretty).get()) ) )
		{
			p = ctl::getLink(citype, p.getProperty(), sys::getLocatorName());
			if(!p)
			{
				ctl::string msg = ctl::string("no link for '") + pretty + "' found";
				ctl::binaryObject a=ctl::exception(msg);
				return resY(vY(), new excptHandle(a), pretty);
			}
		}
	// local call
		if(fobj)
			return (*fobj)((optArgs&)x);

		channelT typ = !p ? serial : p.getType();
		if(typ != lib && typ != thread && typ != serial)
		{
			CTL_Try
			{ return send(p, pretty, (optArgs&)x ); }
			CTL_Catch(const ctl::exception &err)
			{
				ctl::binaryObject a(err);
				return resY(vY(), new excptHandle(a), pretty);
			}
		}
		fobj = (ctl::rFunction<thisT>*)(p.getFunction(pretty).get());
		if(!fobj)
		{
			ctl::exception err(ctl::string(" *** function ") + pretty + " not connected ***");
			ctl::log("#x !^ exception '%s' thrown\n", err.what());
			ctl::binaryObject a(err);
			return resY(vY(), new excptHandle(a), pretty);
		}
		if(typ==thread)
		{
			resY y = threadCall<const ctl::rFunction<thisT>&>(*fobj, x, p, pretty );
			y.setLink(p);
			return y;

		}
		link::envSet tmp(&p);
		resY y=(*fobj)((optArgs&)x);
		y.setPretty(pretty);
		internal::setLinkApplier appl(p);
		appl(y.yM);
		((arg&)x).apply(appl);				

		return y;		
	}
template<class F> struct threadCaller
{
	typedef internal::resThreadReceiver<origY, F, curArgs, optArgs > receiverT;
	static void callByThread(void *v)
	{
		receiverT *recv=(receiverT*)v;

		resY r;
		r = recv->funcM(recv->argsM);
		if(isEmpty)
		{ recv->receivedM = true;	delete recv; }
		else
		{
			if(!r.excptM)
				recv->tmpyM = r.getValue(true);
			else
			{
				recv->tmpbinM = r.excptM->binM;
				r.excptM->binM.clear();
			}
			recv->receivedM = true;
		}
 	}
	static resY call(F f, const ctl::link &p, const optArgs &x, const char *&pretty)
	{
		curArgs xCopy(x);
// clone constant binaries
		xCopy.apply(cloner());
		receiverT *recv=new receiverT(x, xCopy, f, p);
		if(isEmpty	)
			recv->linkM.release();
		resY res(isEmpty ? (receiverT*)0 : recv, (bool)argEmpty);
		res.setPretty(pretty);
		p.addJob(&callByThread, recv);
		return res;
	}
};

template<class F> static resY threadCall(F f, const optArgs &x, const ctl::link &p, const char *pretty)
{	return threadCaller<F>::call(f, p, x, pretty); }

};

} // ctl


template<class T> inline std::ostream &operator<<(std::ostream &os, const ctl::result<T> &res)
{ 	return os<<(const T&)res; }

template<class T> inline std::ostream &operator<<(std::ostream &os, const ctl::rBinResult<T> &res)
{
	ctl::Assert<false> CTL_ERROR__abstract_result_not_printable;
	return os<<(const T&)res;
}

#endif

