/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_ANY_H_
#define CTL_ANY_H_

#include <map>
#include <new>
#include "typename.h"
#include "serialize.h"
#include "pointer.h"
#include "vector.h"


namespace ctl
{

class iStream;
class oStream;
struct binaryObject;
class any;

inline typeId type_name(const binaryObject*)
{ return typeId(); }

inline bool factory_read(any &a, const  typeId &type, iStream &is);
inline bool factory_write(const any &a, const typeId &type, oStream &os);
template<class T> inline void factory_clone(any &a, const typeId &type, const T *);

template<class T> struct genRepT;
template<class T> struct genRepIs;
template<class T> struct genRepHas;


template<> struct fundamental<binaryObject> { enum { is = true }; };

class genericRep: public shared
{
	typeId nameM;
	const std::type_info *idM;
protected:
public:
	typedef cPointer<genericRep> SP;
	virtual const void *value() const =0;
private:
	template<class T> T* cast(trueT, const T*)
	{
		typeId vStr=getTypeName(), tStr = typeName<T>().string();
		return (ctl::equal(typeid(T), *getTypeId())  || (!vStr.empty() && !tStr.empty() && strcmp(vStr.c_str(), tStr.c_str() )==0)) ? (T*)value() : 0; 
	}
	template<class T> T* cast(falseT, const T*)
	{
		if (ctl::equal(typeid(T), *getTypeId()) )
			return (T*)value();
#ifndef CTL_NRTTI
		return dynamic_cast<T*>(this);
#else
		return 0;
#endif
	}
	template<class T> static SP create_(const T&t, trueT)
	{ return SP(new genRepHas<T>(t) ); }
	template<class T> static SP create_(const T&t, falseT)
	{ return SP(new genRepIs<T>(t) ); 	}
public:
	genericRep(const typeId &name, const std::type_info *id): nameM(name), idM(id)
	{}
	virtual SP CTL_clone() const=0;
	const std::type_info *getTypeId() const
	{ return idM; }
	typeId getTypeName() const
	{ return nameM; }
	virtual ~genericRep() {}	
	template<class T> static SP CTL_create(const T&t)
	{ return create_(t,  int2type<fundamental<T>::is>() ); }
	template<class T>  T *get(const T *)
#ifdef CTL_MSC	
	{ return cast(int2type<fundamental<T>::is>(), (const T*)0); }
#else	
	{ return cast<T>(int2type<fundamental<T>::is>(), (const T*)0); }
#endif
	virtual bool equal(const genericRep*) const=0;
	virtual oStream& write(oStream& os) const=0;
};


	template<class T> inline T *any_getP(const any&,T*);
//! class any stores any type and allows safe casts
class any
{
public:
	typedef cPointer<genericRep> repT;
private:
	template<class T> friend inline T *any_getP(const any&,T*);
	mutable repT repM;
	void unique()
	{
		if(repM.isShared())
			 repM=repM->CTL_clone();
	}
	template<class T> static void compileCheck(T*)
	{
		enum { ok = property<hasTypeName, T>::value || property<hasBinaryCheck, T>::value };
		ctl::Assert<ok> CTL_ERROR__Type_wo_serialisation_to_any_assigned;
		(void)CTL_ERROR__Type_wo_serialisation_to_any_assigned;
	}
public:
	any() {}
	template<class T> any (const T &t)
	{
		compileCheck((T*)0);
		typeId name=ctl::typeName<T>(t);
		if(!name.empty() && !fundamental<T>::is)
			ctl::factory_clone(*this, name , &t);
		else
			repM = genericRep::CTL_create(t);
	}
	any(const any &obj): repM(obj.repM) {}
	template<class T> any  &operator= (const T &t)
	{
		compileCheck((T*)0);
		if(!!repM && !repM.isShared() &&  ctl::equal(typeid(T), *getTypeId() ) )
			*(repM->get((T*)0)) = t;
		else
			repM = genericRep::CTL_create(t);
		return *this;
	}
	any &operator = (const any &obj)
	{ repM = obj.repM;  return *this; }
	oStream& write(oStream& os) const
	{ 
		if(!repM)
			return os;
		return repM->write(os); 
	}
	template<class T> T *get(const T * =0)
	{  
		if(!repM)
			return 0;
		unique();
		return ctl::any_getP(*this,(T*)0);	
	}
	template<class T>  const T *get(const T * =0) const
	{ 
		if(!repM)
			return 0;
		return ctl::any_getP(*this,(T*)0);	
	}	
#ifndef CTL_MSC
   template<class T> operator T*()
	{ return get((const T*)0); }
   template<class T> operator const T*() const
	{ return get((const T*)0); }
#endif
	void clear()
	{ repM.clear(); }
	bool operator !() const
	{ return !repM; }
	bool operator ==(const any& a) const
	{ 
		if(!repM && !a)
			return true;
		if(!repM || !a || !ctl::equal(*a.getTypeId(),  *getTypeId() ) )
			return false;
		return repM->equal(a.repM.get() );
	}
	typeId getTypeName() const
	{ return !repM ? typeId() : repM->getTypeName(); }
	const std::type_info *getTypeId() const
	{ return !repM ? 0 :  repM->getTypeId(); }
};

namespace internal
{
	template<typename S>  inline bool equal(const S*, const S*, falseT)
	{ return false; }
	template<typename S>  inline bool equal(const S* a, const S* b, trueT)
	{	return serialize(*a).compare(serialize(*b)) == 0; }
//	{	return serialize(*a)==serialize(*b); }
}

template<typename T> struct genRepT: public genericRep
{
//	template<typename S>  inline friend  bool operator==(const S&, const S&);
		enum { serializable = property<hasTypeName, T>::value || property<hasBinaryCheck, T>::value };
	genRepT():genericRep(ctl::typeName<T>().string(),&typeid(T))
	{}
	oStream& write(oStream& os, trueT) const
	{ return os<<*(const T*)value(); }
/*
	oStream& write(oStream& os, falseT) const
	{ 
		throwStr(" *** ctl::oStream&<<(ctl::oStream&, %s) not defined ***", typeid(T).name()); 
		return os; 
	}
*/
	oStream& write(oStream& os) const
	{ return write(os, int2type<serializable>() ); }
//	{ return write(os, typename property<hasTypeName, T>::resultT() ); }
	bool equal(const genericRep* t) const
	{ 
		return internal::equal((const T*)value(), (const T*)t->value(), int2type<serializable>());	}
};

template<class T> struct genRepIs :public genRepT<T> ,  public T
{
	typedef any::repT SP;
	genRepIs(const T &t): T(t) {}
	genRepIs(const genRepIs<T> &t): genRepT<T>(), T((const T&)t) {}
	~genRepIs() throw() {}
	SP CTL_clone() const	
	{ return SP(new genRepIs<T>(*this) ); }	
	const void *value() const
	{ return static_cast<const T*>(this); }
};

template<class T> struct genRepHas : public genRepT<T>
{
	typedef any::repT SP;
	T tM;
	genRepHas(const T &t): tM(t)  {}
	SP CTL_clone() const	
	{ return SP(new genRepHas<T>(tM) ); }	
	const void *value() const
	{ return &tM; }
};

template<class T> inline T *pointer_cast(any &obj, const T* =0)
{ return obj.get((const T*)0); }
template<class T> inline const T *pointer_cast(const any &obj, const T* =0)
{ return obj.get((const T*)0); }


struct binaryObject
{
operator ctl::property<ctl::hasTypeName>() const;

static ctl::typeId CTL_typeNameS()
{ return "ctl::binaryObject"; }
ctl::typeId CTL_typeName() const
{ return CTL_typeNameS(); }

	typeId type, msg;
	mutable ctl::vector<char> buffer;
// definition in serialze.inl
	oStream &write(oStream &os) const;
	iStream &read(iStream &is);
	binaryObject() {}
	binaryObject(const typeId &typ, uint4 size, const char *buf): type(typ), buffer(size,buf) {}
	template<class T>  binaryObject(const T& t)
	{
		type = ctl::typeName<T>();
		msg = type;
		oStream os = serialize(t);
		buffer = ctl::vector<char>(os.size(), os.buffer() );		
	}
	binaryObject(const ctl::exception& e): msg(e.what())
	{
		type = ctl::typeName<ctl::exception>();
		oStream os = serialize(e);
		buffer = ctl::vector<char>(os.size(), os.buffer() );		
	}
	void clear()
	{ type .clear(); buffer.clear(); }
	void throwIt();
	bool operator !() const
	{ return type.empty(); }
	template<class T> any cast(T *&t, trueT) const
	{
		any a;
		iStream is(buffer);
		typeId name_T=ctl::typeName<T>();
		if(!name_T.empty() && strcmp(type.c_str(), name_T.c_str() )==0)
		{
			t = pointer_cast<T>(a = T());
			is>>*t;
		}
		else
		{
			ctl::factory_read(a, type.c_str(), is);
			t = pointer_cast<T>(a);
		}
		return a;
	}
	any cast(void *, falseT) const
	{ return any(); }
};

template<class T> inline T *any_getP(const any &a,T*)
{
	T *t=a.repM->get((T*)0);
	if(!t && ctl::equal(*a.getTypeId(), typeid(binaryObject) ) )
	{
		const  binaryObject *obj = a.repM->get((const binaryObject*)0);
		typename ctl::property<ctl::hasTypeName, T>::resultT hasName;
		any aa = obj->cast(t, hasName);
		if(!!aa)
			a.repM = aa.repM;
	}
	return t;
}


struct excptHandle : public shared
{
	binaryObject binM;
	excptHandle(const binaryObject &bin=binaryObject()): binM(bin)
	{}
	void throwIt()
	{ binM.throwIt(); }
	ctl::string msg() const
	{ return binM.msg; }
	ctl::typeId type() const
	{ return binM.type; }
	~excptHandle()
	{ binM.throwIt(); }
};

} // ctl

/*
dangerous: used if no operator exist
inline std::ostream &operator<<(std::ostream &os, const ctl::any &a)
{ return os<<"\""<<(!a.getTypeName().empty() ? a.getTypeName() : "nil")<<"\"";}
*/

#endif
