/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_FACTORY_H_
#define CTL_FACTORY_H_

#include "types.h"
#include "lexicalassoc.h"
#include "serialize.h"
#include <typeinfo>


namespace ctl
{

class any;

struct polymorphIOBase
{
	virtual ~polymorphIOBase() {}
	virtual oStream &CTL_write(const void *, oStream &) const=0;
	virtual iStream &CTL_read(void *&, iStream &, bool) const=0;
	virtual iStream &CTL_read(any &, iStream &) const=0;
	virtual oStream &CTL_write(const any &, oStream &) const=0;
	virtual void CTL_clone(any &, const void *) const=0;
};


template<class T, class Base> struct polymorphIO: public polymorphIOBase, public Base
{
	~polymorphIO() throw() {}
	oStream &CTL_write(const void *obj, oStream &os) const
	{ return os<<*(const T*)obj; }
	iStream &CTL_read(void *&obj, iStream &is, bool thrw) const
	{
		if (thrw)
		{
			obj = 0;
			T tobj;	
			is>>tobj;
			CTL_Throw((const T&)tobj);
		}
		const T *&t=(const T*&)obj;
		is.create(t);
		is>>*(T*)t;
		return is;
	}
	iStream &CTL_read(any &a, iStream &is) const;
	oStream &CTL_write(const any &, oStream &) const;
	void CTL_clone(any &a,	const void *t) const;
};


inline bool factory_read(any &a, const typeId &type, iStream &is);
inline bool factory_write(const any &a, const typeId &type, oStream &os);
template<class T> inline void factory_clone(any &a, const typeId &type, const T *);

class factory;
inline factory &factoryS();

class factory
{
	struct typeIdLess
	{
		bool operator()(const std::type_info *t0, const std::type_info *t1) const
		{ return t0 && t1 ? compare(*t0, *t1) < 0 : t0<t1; } 
	};
friend bool factory_read(any &a, const typeId &type, iStream &is);
friend bool factory_write(const any &a, const typeId &type, oStream &os);
friend factory &factoryS();
template<class T> friend void factory_clone(any &a, const typeId &type, const T *);

	lexicalAssoc<char, const polymorphIOBase*> ioM;
#ifndef CTL_NRTTI
	typedef std::map<const std::type_info*, typeId, typeIdLess> typeidMapT;
	typeidMapT namesM;
#endif

	template<class T> static T* createObj(const T* =0)
	{	return new T;  }
public:
	virtual ~factory()
	{
		for(unsigned int i=0; i<ioM.size(); i++)
		{
			ioM.get(i)->~polymorphIOBase();
			delete ioM.get(i);
		}
	}
	typeId find(const std::type_info *id) const
	{
		typeidMapT::const_iterator fnd = namesM.find(id);
		if(fnd!=namesM.end())
			return fnd->second;
		return typeId();
	}
	const polymorphIOBase *const *find(const std::type_info *id, const typeId &type) const
	{
		typeId typeN, typeI= find(id);
		if(!typeI.empty())
			typeN = typeI;
		else
			typeN = type;
		return	typeN.empty() ? 0 : ioM.find(typeN.c_str());
	}
	template<class T> static bool add(T* , const typeId &type )
	{		
		factory &fac = factoryS();
		enum { fd=fundamental<T>::is };
		typedef typename ctl::select<fd, empty, T>::resultT base;
		if(type.empty())
			return false;
		if(fac.find(&typeid(T) ).empty() )
			fac.namesM[&typeid(T)] = type;
		const char *str=type.c_str();
		if(!fac.ioM.find(str))
		{
			ctl::log("#c ++ %s\n", str);
			fac.ioM[str] = createObj((polymorphIO<T, base>*) 0);
		}
		return true;
	}

	template<class T> static bool write(const T *t, typeId &typName, oStream &os)
	{
		if(t && !typName.empty())
		{
			const polymorphIOBase *const *io=factoryS().find(&typeid(*t), typName.c_str());
			if(!io)
				os<<*t;
			else
				(*io)->CTL_write(t,os);
			return true;
		}
		return false;
	}

	template<class T> static void read_(const T *&t, const polymorphIOBase *, falseT, iStream &)
	{ t = 0; }
	template<class T> static void read_(const T *&t, const polymorphIOBase *io, trueT,	iStream &is)
	{
#ifndef CTL_NRTTI
		if(dynamic_cast<const T*>(io))
		{
			io->CTL_read((void*&)t,is, false);
			return;
		}
#else
		(void)io;
#endif
			t = 0;
	}
	void throwIt(const typeId &type, iStream &is) const;
	template<class T> static void read(const T *&t, const typeId &type, iStream &is)
	{
		if(!type.empty())
		{
			typeId nameOfT=ctl::typeName<T>(*t);
// same typeId's
			if(!nameOfT.empty() && strcmp(type.c_str(), nameOfT.c_str() )==0)
			{ 
				is.create(t);
				is>>*(T*)t;
				return; 
			}
			const polymorphIOBase * const * io=factoryS().ioM.find(type.c_str());
			if(!io)
			{
				t = 0;
				return;
			}		
			read_(t, *io, int2type<ctl::property<ctl::hasTypeName,T>::intrinsic>(), is);

		}
	}
};

  CTL_Inline factory &pFactory();

inline factory &factoryS()
{
	factory &fctrS =pFactory();
	return fctrS;
}

inline typeId type_name(const void *)
{ return "?"; }
template<int n> inline ctl::typeId type_name(const int2type<n>* =0)
{ return string::int2string(n); }

} // ctl


CTL_SetTypeName(ctl::exception);
CTL_SetTypeName(bool);
CTL_SetTypeName(char);
CTL_SetTypeName2(uchar, unsigned char);
CTL_SetTypeName(int2);
CTL_SetTypeName(uint2);
CTL_SetTypeName(int4);
CTL_SetTypeName(uint4);
CTL_SetTypeName(int8);
CTL_SetTypeName(uint8);
CTL_SetTypeName(real4);
CTL_SetTypeName(real8);
#include "any.h"
CTL_SetTypeName(ctl::any);

namespace ctl
{

template<class T> inline typeId connectType2(T* t, const typeId &type)
{
	bool succ=factory::add(t, type);
	return succ ? type : typeId();
}

template<class T> inline typeId connectType(T* t)
{ return connectType2(t,ctl::typeName<T>() ); }

template<class T, class Base> inline iStream &polymorphIO<T,Base>::CTL_read(any &a, iStream &is) const
{
		a = T();
		T *t=pointer_cast<T>(a);
		return is>>*t;
}

template<class T, class Base> inline oStream &polymorphIO<T,Base>::CTL_write(const any &a, oStream &os) const
{
	 		const T *t=pointer_cast<T>(a);
		return os<<*t;
}

template<class T, class Base> inline void polymorphIO<T,Base>::CTL_clone(any &a,	const void *t) const
{ a = *(const T*)t; }

inline bool factory_write(const any &a, const typeId &type, oStream &os)
{
	if(!type.empty())
	{
		const polymorphIOBase * const * io=factoryS().find(a.getTypeId(), type.c_str());
		if(io)
			(*io)->CTL_write(a,os);
		return !!io;
	}
	return false;	
}

inline bool factory_read(any &a, const typeId &type, iStream &is)
{
	a.clear();
	if(!type.empty())
	{
		const polymorphIOBase * const * io=factoryS().ioM.find(type.c_str());
		if(!!io)
			(*io)->CTL_read(a,is);
		return !!io;
	}
	return false;
}	


template<class T> inline void factory_clone(any &a, const typeId &type, const T *t)
{
	a.clear();
	enum { fd=fundamental<T>::is };
	typedef typename ctl::select<fd, type2type<T>, T>::resultT base;

	if(!type.empty() && type[0] !='?')
	{
		const polymorphIOBase *IO=0;
		const polymorphIOBase * const *io=factoryS().find(&typeid(*t), type.c_str());

		if(!io)
		{
			IO = new polymorphIO<T, base>(); //newIO(int2type<fd>(), (T*)0);
			io = &IO;
			factoryS().ioM[type.c_str()] = IO;
		}
		(*io)->CTL_clone(a,t);
	}
}


} //  ctl


#endif


