/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_LEXICAL_ASSOC_H_
#define CTL_LEXICAL_ASSOC_H_

#include "vector.h"


namespace ctl
{
class oStream;
class iStream;

namespace internal
{

template<class charT>   struct bounds;

template<> struct bounds<char>
{ enum {low=0, high=127, end='\0' };  };

template<> struct bounds<wchar_t>
{ enum {low=0, high=32767, end=L'\0' }; };

template<class charT>  inline bool lessThan(const charT *a,const charT *b)
{
	--a; --b;
	while(*++a==*++b)
		if(*a== (charT)bounds<charT>::end)
			return false;
	return *a<*b;
}

template<class charT>
inline unsigned int length(const charT *a)
{
	if(!a)
		return 0;
	const charT *start=a;
	while(*a!=(charT)bounds<charT>::end) a++;
	return a - start;
}

}   // internal

/**/// accessor for results of methods or functions

template<class charT,class Y>   class lexicalAssoc
{
	vector<charT> strM;
	unsigned int strSizeM;
	vector<int> offM;
	unsigned int offSizeM;
	vector<Y> yM;
	bool nullM;

// a bsearch implemetation
	unsigned int lowerBound(const charT *s) const
	{
		int a0=0,a1=offSizeM;
		const int *off=offM.p();
		const charT *str = strM.p();
		
// invariant of loop is str[a0] <= s < str[a1]
		while(a1-a0>1)
		{
			int m=a0+((a1-a0)>>1);
			if(internal::lessThan(s,str+off[m]))
				a1 = m;
			else
				a0 = m;
		}
		return a0;
	}
	int index(const charT *s) const
	{
		int pos=lowerBound(s);
		return internal::lessThan(strM.p()+offM[pos],s)  ? -1-pos : pos;
	}
	void init()
	{
		if(!!offM)
			return;
		strM.resize(32);
		offM.resize(8);
		yM.resize(8);
		strM[0]=(charT)internal::bounds<charT>::end;
		strM[1]=(charT)internal::bounds<charT>::high;
		strM[2]=(charT)internal::bounds<charT>::end;
		offM[0] = 0;  	offM[1] = 1;  	offM[2] = 3;
	}
public:
	size_t size() const
	{ return offSizeM - ( nullM ? 1 : 2); }
	bool empty() const
	{ return size()==0; }
	lexicalAssoc(): strSizeM(3), offSizeM(2), nullM(false)
	{}
	int insert(const Y &yy,const charT *s, int pos=0)
	{
		init();
		if(!s || *s==(charT)internal::bounds<charT>::end)
		{
			yM[0] = yy;
			nullM = true;
			return 0;
		}
		if(!pos)
			pos = lowerBound(s);
		else
			pos = -(pos+1);
		int *off=offM.p();
		charT *str = strM.p();

		if(!internal::lessThan(str+off[pos],s))
		{
			yM[pos] = yy;
			return  pos;
		}
		int index=pos+1;
		int len=internal::length(s)+1;
		strSizeM += len;
		offSizeM++;
		if(strSizeM>strM.size())
		strM.resize(strSizeM + (strSizeM>>2));
		if(offSizeM>=offM.size())
		{
			offM.resize(offSizeM + (offSizeM>>2));
			yM.resize(offSizeM + (offSizeM>>2));
		}
		off=offM.p();
		str = strM.p();

		int i;
		for(i=strSizeM-len-1; i>=off[index]; i--)
			str[i+len] = str[i];
		for(i=0; i<len; i++)
			str[off[index]+i] = s[i];
       CTL_Assert(str[strSizeM-2]==(charT)internal::bounds<charT>::high);
		Y *y=yM.p();
		for(i=offSizeM-1; i>=index; i--)
		{
			off[i+1] = off[i]+len;
			y[i+1] = y[i];
		}
      CTL_Assert(str[offM[offSizeM-1]]==(charT)internal::bounds<charT>::high);
		yM[index] = yy;
		return index;
	}
	void erase(const charT *s)
	{
		init();
		if(!s || *s==(charT)internal::bounds<charT>::end)
		{
			yM[0] = Y();
			nullM = false;
			return;
		}   	
		int pos=index(s);
		if(pos<0)
			return;
		
		int *off=offM.p();
		charT *str = strM.p();
		
		int len=internal::length(s)+1;
		strSizeM -= len;
		
		unsigned int i;
		for(i=off[pos]; i<strSizeM; i++)
			str[i] = str[i+len];
		offSizeM--;
		for(i=pos; i<offSizeM; i++)
			off[i] = off[i+1]-len;

		Y *y = yM.p();
		for(i=pos; i<offSizeM; i++)
			y[i] = y[i+1];
	}
	Y &get(unsigned int i)
	{
		CTL_Assert(i<size());
		return yM[i+(nullM ? 0 : 1)];
	}
	const Y &get(unsigned int i) const
	{
		CTL_Assert(i<size());
		return yM[i+(nullM ? 0 : 1)];
	}
	const charT *getKey(unsigned int i) const
	{
		CTL_Assert(i<size());
		return strM.p() + offM[i+(nullM ? 0 : 1)];
	}
	Y &operator[](const charT *s)
	{
		init();
		if(!s || *s==(charT)internal::bounds<charT>::end)
			return yM[0];
		int pos = index(s);
		if(pos<0)
			pos = insert(Y(),s,pos);
		return yM[pos];
	}
	const Y *find(const charT *s) const
	{
		if(!offM)
			return 0;
		if(!s || *s==(charT)internal::bounds<charT>::end)
			return yM.p();
		int pos = index(s);
		if(pos<0)
			return 0;
		return yM.p()+pos;
	}
	void clear()
	{ strM.clear(); 	offM.clear();  yM.clear(); }
	ctl::oStream &write(ctl::oStream &os) const
	{ return os<<strM<<strSizeM<<offM<<offSizeM<<yM<<nullM; }
	ctl::iStream &read(ctl::iStream &is)
	{ return is>>strM>>strSizeM>>offM>>offSizeM>>yM>>nullM; }
	std::ostream &print(std::ostream &os) const
	{
		unsigned int i = nullM ? 0 : 1;
		for(; i<offSizeM-1; i++)
			os<<"'"<<strM.p()+offM[i]<<"'"<<" -> "<<yM[i]<<"\n";
		return os;
	}
};

template<class K, typename T> inline std::ostream &operator<<(std::ostream &os, const ctl::lexicalAssoc<K,T> &v)
{	return v.print(os); };


} // ctl

#endif
