/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_MATRIX_H_
#define CTL_MATRIX_H_

#include "ctl.h"

namespace ctl
{

template<typename T> class matrix : public vector<T>
{
	typedef vector<T> baseT;
public:
	using vector<T>::iterator;
	using vector<T>::const_iterator;

	using vector<T>::size;
	using vector<T>::begin;
	using vector<T>::end;
	typedef typename vector<T>::sizeT sizeT;
private:
//	using vector<T>::resize;
	uint8 dim1M,dim2M;
public:
#ifdef CTL_Template
	CTL_Template(ctl::matrix,(T), 1, isArray, (T,_,_), 3 );
	iStream &read(iStream &is)
	{ 
		is>>dim1M>>dim2M; 
		baseT::resize(dim1M*dim2M);
		return is.read( begin(), dim1M * dim2M);
	} 
	oStream &write(oStream &os) const
	{ os<<dim1M<<dim2M; return os.write(begin(), size() ); } 
#endif
	matrix(const matrix<T>& m): vector<T>(m), dim1M(m.dim1M), dim2M(m.dim2M)
	{}
	template<typename S>
		matrix(const matrix<S>& m): vector<T>((const vector<S>&)m), dim1M(m.dim1()), dim2M(m.dim2())
	{}
	matrix(sizeT dim1, sizeT dim2,const T* t=0): vector<T>(dim1*dim2,t),dim1M(dim1), dim2M(dim2)
	{}
	matrix(sizeT dim1, sizeT dim2,const T &t): vector<T>(dim1*dim2,t),dim1M(dim1), dim2M(dim2)
	{}
	matrix(const vector<T> &v): vector<T>(v), dim1M(v.size()), dim2M(1)
	{}
	matrix():vector<T>(0),dim1M(0), dim2M(0)	
	{}
	vector<T> col(sizeT i) const
	{
		CTL_Assert(i<dim2M);
		vector<T> v(dim1());
		T *pv=v.begin(),*pe = pv+dim1();
		const double *m=begin()+i;
		while(pv<pe)
		{
			*pv++ = *m;
			m += dim2M;
		}
		return v;
	}
	vector<T> row(sizeT i) const
	{
		CTL_Assert(i<dim1());
		vector<T> v(dim2M);
		T *pv=v.begin(),*pe = pv+dim2M;
		const double *m=begin()+i*dim2M;
		while(pv<pe)
			*pv++ = *m++;
		return v;
	}
	void setCol(size_t i,const vector<T> &v)
	{
		CTL_Assert(i<dim2M);
		CTL_Assert(v.size()==dim1M);
		const T *pv=v.begin(),*pe = pv+dim1M;
		T *m=begin()+i;
		while(pv<pe)
		{
			*m = *pv++;
			m += dim2M;
		}
	}
	vector<T> getCol(size_t i) const
	{ return col(i); }
	void setRow(size_t i,const vector<T> &v)
	{
		CTL_Assert(i<dim1M);
		CTL_Assert(v.size()==dim2M);
		const T *pv=v.p(),  * const pe = pv+dim2M;
		T *m=begin()+i*dim2M;
		while(pv<pe)
			*m++ = *pv++;
	}
	vector<T> getRow(size_t i) const
	{ return row(i); }

	void resize(sizeT dim1, sizeT dim2)
	{
		matrix cp(*this);
		size_t d1=dim1<dim1M ? dim1:dim1M;
		size_t d2=dim2<dim2M ? dim2:dim2M;
		dim1M = dim1; dim2M = dim2;
		baseT::resize(dim1*dim2);
		for(size_t i=0; i<d1; i++)
			for(size_t j=0; j<d2; j++)
				(*this)(i, j) = cp(i, j);
	}	
	void clear() 
	{
		((vector<T>*)this)->clear();
		dim1M = dim2M = 0;
	}
	sizeT dim1() const
	{ return dim1M; }
	sizeT dim2() const
	{ return dim2M; }
	const T& operator()(sizeT i,sizeT j) const
	{
		CTL_Assert(i<dim1() && j<dim2M);
		return begin()[dim2M*i+j];
	}
	T& operator()(sizeT i,sizeT j)
	{
		CTL_Assert(i<dim1() && j<dim2M);
		return begin()[dim2M*i+j];
	}

	vector<T> operator()(const vector<T>& v) const
	{ return *this * v; }
	matrix<T> operator()(const matrix<T>& m) const
	{ return *this * m; }
	matrix<T> &operator  +=(const matrix<T> &X)
	{ ((baseT*)this)->operator +=(X); return *this; }
	matrix<T> &operator  -=(const matrix<T> &X)
	{ ((baseT*)this)->operator -=(X); return *this; }
	matrix<T> &operator  *=(const T &f)
	{ ((baseT*)this)->operator *=(f); return *this; }
	matrix<T> &operator  /=(const T &f)
	{ ((baseT*)this)->operator /=(f); return *this; }
};


template<class T> inline T sum(const vector<T> &v)
{
	T s=T(), *a=v.p(), *o=a + v.size();
	while(a<o)
		s += *a++;
	return s;
};

template<class T> inline vector<T> operator+(const vector<T> &v0,  const T &t)
{
	 vector<T> s(v0);
	 for(unsigned int i=0; i<s.size(); i++)
		 s[i] += t;
	 return s;
};

template<class T> inline vector<T> operator-(const vector<T> &v0,  const T &t)
{
	 vector<T> s(v0);
	 for(unsigned int i=0; i<s.size(); i++)
		 s[i] -= t;
	 return s;
};

template<class T> inline matrix<T> operator-(const matrix<T> &v)
{
	matrix<T> s(v);
	T *a=s.begin(), *o=a + s.size();
	while(a<o)
		*a = -*a, a++;
	return s;
}
template<class T> inline vector<T> operator*(const vector<T> &v,  const vector<T> &f)
{
	CTL_Assert(v.size()==f.size());
	vector<T> s(v);
	T *a=s.p(), *o=a + v.size();
	const T *vp=v.p(), *fp=f.p();
	while(a<o)
		*a++ = *vp++ * *fp++;
	 return s;
};

template<class T> inline vector<T> operator/(const vector<T> &v,  const vector<T> &f)
{
	CTL_Assert(v.size()==f.size());
	vector<T> s(v);
	T *a=s.p(), *o=a + v.size();
	const T *vp=v.p(), *fp=f.p();
	while(a<o)
		*a++ = *vp++ / *fp++;
	 return s;
};

namespace internal
{
	CTL_Registrate(11, std::complex<float>);
	CTL_Registrate(12, std::complex<double>);

	template<typename S, typename T> struct resultType<ctl::vector<S>, ctl::vector<T>, '*'>
	{
		typedef typename resultType<S, T, '*'>::resultT U;
		typedef ctl::vector<U>  resultT;
	};

	template<typename S, typename T> struct resultType<ctl::matrix<S>, ctl::vector<T>, '*'>
	{
		typedef typename resultType<S, T, '*'>::resultT U;
		typedef ctl::vector<U>  resultT;
	};
	template<typename S, typename T> struct resultType<ctl::matrix<S>, ctl::matrix<T>, '*'>
	{
		typedef typename resultType<S, T, '*'>::resultT U;
		typedef ctl::matrix<U>  resultT;
	};


	template<typename S, typename T> struct resultType<ctl::vector<S>, ctl::vector<T>, '+'>
	{
		typedef typename resultType<S, T, '+'>::resultT U;
		typedef ctl::vector<U>  resultT;
	};

	template<typename S, typename T> struct resultType<ctl::matrix<S>, ctl::matrix<T>, '+'>
	{
		typedef typename resultType<S, T, '+'>::resultT U;
		typedef ctl::matrix<U>  resultT;
	};

}

template<typename S, typename T> inline typename internal::resultType<S, T, '*'>::resultT scal(const S &a, const T &b)
{ return a*b; }
template<typename S, typename T> inline typename internal::resultType<S, T, '*'>::resultT scal(const vector<S> &v0,  const vector<T> &v1)
{
	CTL_Assert(v0.size()==v1.size());
	const S *a=v0.p(), *o=a + v0.size();
	const T *b=v1.p();
	typedef typename internal::resultType<S, T, '*'>::resultT U;
	U s=U();
	while(a<o)
		s += *a++ * *b++;
	return s;
}

template<typename S, typename T> inline typename internal::resultType<S, T, '*'>::resultT scal(const matrix<S> &v0,  const matrix<T> &v1)
{ return scal((const vector<T>&)v0, (const vector<T>&)v1); }

template<class T> inline T normSquare(const vector<T> &v0)
{
	const T *a=v0.p(), *o=a + v0.size();
	T s=T();
	while(a<o)
		s += *a * *a, a++;
	return s;
}

	template<class T> inline T norm(const vector<T> &d)
	{ return sqrt(scal(d,d) ); }

	template<class T> inline matrix<T> trans(const matrix<T> &A)
	{
CTL_Profile("trans");
		const int d1=A.dim1(), d2=A.dim2();
		const T *a= A.p(), *oa=a+d1*d2;
		matrix<T> C(d2, d1);
		T *c=C.p();
		T *cc;
		while(a<oa)
		{
			const T *const oa2=a+d2;
			cc = c++;
			while(a<oa2)
			{	*cc = *a++; cc += d1;	}
		}
		return C;
	}

	template<class T> inline void appendCol(matrix<T> &A, const vector<T> &row)
	{
CTL_Profile("append col");
		const int d1=A.dim1() ? A.dim1() : row.size(), d2=A.dim2();
		CTL_Assert(size_t(d1)==row.size() );
		A.resize(size_t(d1), size_t(d2+1) );
		T *a=A.begin();
		const T *r=row.begin();
		for(int i=d1-1; i>=1; i--)
		{
			for(int j=d2-1; j>=0; j--)
				a[i*d2 + i + j] = a[i*d2 + j];
			a[i*d2+i+d2] = r[i];
		}
		a[d2] = r[0];
	}

	template<class T> inline void appendRow(matrix<T> &A, const vector<T> &row)
	{
CTL_Profile("append row");
		const int d1=A.dim1(), d2=A.dim2() ? A.dim2() : row.size();
		CTL_Assert(size_t(d2)==row.size() );
		A.resize(size_t(d1+1), size_t(d2) );
		T *a=A.begin() + d1*d2;
		const T *r=row.begin();
		for(int i=0; i<d2; i++)
			*a++ = *r++;
	}

	template<class T> inline void deAppendCol(matrix<T> &A, vector<T> &row)
	{
CTL_Profile("deappend col");
		const int d1=A.dim1(), d2=A.dim2()-1;
		row.resize(d1);
		CTL_Assert(size_t(d1)==row.size() );
		T *a=A.p();
		row[0] = a[d2];
		for(int i=0; i<d1; i++)
		{
			row[i] = a[i*d2+i+d2];
			for(int j=0; j<d2; j++)
				a[i*d2 + j] = a[i*d2 + i+ j];
		}
		A.resize(d1, d2);
	}

	template<class T> inline T trace(const matrix<T> &A)
	{
		CTL_Assert(A.dim1()==A.dim2() );
		T s = 0;
		for(size_t i=0; i<A.dim1(); i++)
			s += A(i,i);
		return s;
	}
	template<class T> inline matrix<T> operator*(const matrix<T> &A, const T &f)
	{
		matrix<T> C(A);
		T *c=C.begin(), *oc=C.begin()+C.size();
		while(c<oc)
			*c++ *= f;
		return C;
	}
	template<class T> inline matrix<T> operator/(const matrix<T> &A, const T &f)
	{ return A * (1.0/f); }

	template<class T> inline matrix<T> operator+(const matrix<T> &A, const matrix<T> &B)
	{
		size_t s = A.size();
		CTL_Assert((A.dim1()==B.dim1()) && (A.dim2()==B.dim2()) );
		const T *b=B.p(), *ob=b+s;
		matrix<T> C(A);
		T *c = C.p();
		while(b<ob)
			*c++ += *b++;
		return C;
	}
	template<class T> inline matrix<T> operator-(const matrix<T> &A, const matrix<T> &B)
	{
		size_t s = A.size();
		CTL_Assert((A.dim1()==B.dim1()) && (A.dim2()==B.dim2()) );
		CTL_Assert(s==B.size());
		const T *b=B.p(), *ob=b+s;
		matrix<T> C(A);
		T *c = C.p();
		while(b<ob)
			*c++ -= *b++;
		return C;
	}
	template<class T> inline vector<T> operator*(const matrix<T> &A, const vector<T> &V)
	{
CTL_Profile("M * V");
		const size_t d1 = A.dim1(), d2=A.dim2();
		CTL_Assert(V.size()==d2);
		const T *a=A.p(), *v=V.p();
		vector<T> W(d1);
		T* w= W.p();
		for(size_t i=0; i<d1; i++)
			for(size_t j=0; j<d2; j++)
				w[i] += a[d2*i+j] * v[j];
		return W;
	}

	template<class T> inline matrix<T> operator*(const matrix<T> &A, const matrix<T> &B)
	{
//CTL_ProfileC("matrix * matrix 1",A.size()==1 || B.size()==1);
		const size_t d1 = A.dim1(), d2=A.dim2(), d3= B.dim2();
		CTL_Assert(B.dim1()==d2);
		const T *a=A.p(), *b=B.p();
		matrix<T> C(d1, d3);
		T* c= C.p();

CTL_Profile("matrix * matrix");
		const T *const oa=a+d1*d2,*const ob=b+d2*d3;

		for(const T* aa=a; aa<oa; aa+=d2)
		{
			for(const T *bb=b-d3; bb<b; bb++)
			{
				const T *aaa=aa, *bbb=bb;
				T t=0;
				while((bbb+=d3)<ob)
					t += *aaa++ * *bbb;
				*c++ = t;
			}
		}
		return C;
	}

	template<class T> inline ctl::matrix<T> diag(const ctl::vector<T> &v)
	{
		matrix<T> D(v.size(), v.size() );
		T* d=D.begin();
		const T* a=v.begin(), *o=v.end();
		while(a<o)
		{
			*d = *a++;
			d += v.size()+1;
		}
		return D;
	}

	template<class T> inline ctl::vector<T> diag(const ctl::matrix<T> &M)
	{
		CTL_Assert(M.dim1()==M.dim2() );
		vector<T> d(M.dim1() );
		const T* m=M.begin();
		T* a=d.begin(), *o=d.end();
		while(a<o)
		{
			*a++ = *m;
			m += d.size()+1;
		}
		return d;
	}

	template<class T> inline matrix<T> diagInv(const vector<T> &v)
	{
		matrix<T> D(v.size(), v.size() );
		T* d=D.begin();
		const T* a=v.begin(), *o=v.end();
		while(a<o)
		{
			*d = (*a==0) ? 0 : *a;
			a++;
			d += v.size()+1;
		}
		return D;
	}

	template<class T> inline ctl::matrix<T> dyad(const ctl::vector<T> &a, const ctl::vector<T> &b)
	{
	CTL_Profile("dyad");
	ctl::matrix<T> M(a.size(), b.size() );
	T *m=M.p();
	const T *pa=a.begin(), *pb=b.begin();
	for(size_t i=0; i<a.size(); i++)
		for(size_t j=0; j<b.size(); j++)
			*m++ = pa[i] * pb[j];
	return M;
}

template<class T> inline ctl::matrix<T> dyad(const ctl::matrix<T> &A, const ctl::matrix<T> &B)
{
	ctl::matrix<T> M(A.dim1() * B.dim1(), A.dim2() * B.dim2());
	double *m=M.p();
	for(size_t i=0; i<A.dim1(); i++)
		for(size_t j=0; j<B.dim1(); j++)
			for(size_t k=0; k<A.dim2(); k++)
				for(size_t l=0; l<B.dim2(); l++)
					*m++ = A(i,k) * B(j, l);
	return M;
}

template<class T> inline std::ostream &operator<<(std::ostream &str,const ctl::matrix<T> &A)
{
typedef typename ctl::matrix<T>::sizeT sizeT; 
	if(A.size()==0)
		return str;
	if(A.size()==1)
	{
		str<<A[0];
		return str;
    }
	for(sizeT i=0; i<A.dim1(); i++)
	{
		str<<"| "<<A(i,0);
		for(sizeT j=1; j<A.dim2(); j++)
			str<<", "<<A(i,j);
		str<<" |\n";
	}
	return str;
}

}



#endif

