/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_POINTER_H_
#define CTL_POINTER_H_

#include <assert.h>
#include "types.h"
#include <typeinfo>

namespace ctl
{
class oStream;
class iStream;

	class shared
	{
		mutable int refCountM; 
		void (*delM)(void*);
		static void deleteS(void *p)
		{ delete[] (char*)p; }
	public:
		shared *getSharedP() const
		{ return (shared*)this; }
		void *operator new(size_t s)
		{ return new char[s]; }
		void *operator new(size_t, void *p)
		{ return p; }
		void operator delete(void *p)
		{
			shared *thisP = static_cast<shared*>(p);
			if(thisP && thisP->delM)
				(*thisP->delM)(p);
		}
		void release()
		{ delM = 0; }
		void operator delete(void *, void*)
		{}
		int getRefCount() const
		{ return  refCountM; }
		shared(): refCountM(0), delM(deleteS)
		{}
		shared(const shared&): refCountM(0), delM(deleteS)
		{}
		shared& operator=(const shared&)
		{ return *this; }

		virtual ~shared()
		{}		
		void ref() const
		{ refCountM++; }
		void deref() const
		{ refCountM--; }

		void unref() const
		{
			if(refCountM<=0)
				ctl::log("#w !! multiple destruction of %p\n", this);
			else if(--refCountM==0 && delM)
				delete (shared*)this;
		}
	};

// -> group.h
	template<class T>  inline shared *getSharedP(const T*);
// intrinsic pointer
	template <class T>  class cPointer
	{
	typedef shared pointedT;
		void moveRefTo(const T *obj, bool hold)
		{
			if(obj && hold)
				getSharedP(obj)->ref();
			if(soM && holdM)
				getSharedP(soM)->unref();
		}
		void ref() const
		{
			if(soM && holdM)
				getSharedP(soM)->ref();
		}
		void unref() const
		{
			if(soM && holdM)
				getSharedP(soM)->unref();
		}
	public:
		int getRefCount() const
		{ return soM ? getSharedP(soM)->getRefCount() : 0; }
		cPointer	(): holdM(true),soM(0)
		{}
		cPointer(const T *obj, bool hold =true):holdM(hold),soM((T*)obj)
		{
//			CTL_Assert(static_cast<pointedT*>(soM)==(pointedT*)(soM));
			ref();
		}
		template<class U>
			cPointer	(const cPointer<U>& obj):holdM(obj.holdM),soM(obj.get())
//			cPointer	(const cPointer<U>& obj):soM((U*)obj.get())
		{ ref(); }		
		cPointer(const cPointer<T>& obj):holdM(obj.holdM), soM(obj.soM)
		{ ref(); }		
		~cPointer	()
		{ unref(); }	
		void clear()
		{
			unref();
			soM = 0;			
		}
// usage is dangerous
		void destruct()
		{
			shared *so=getSharedP(soM);
			soM = 0;
			delete so;
		}
		void release(bool strong=true)
		{ 
			if(strong)
				soM = 0;
			else
				unref();
			holdM = false;
		}
		void derelease()
		{
			if(!holdM)
			{
				holdM = true;
				ref();
			}
		}
		template<class U>
			cPointer<T>   &operator=	(const cPointer<U> &obj)
		{
//			((pointedT*)obj.get());
			moveRefTo(obj.get(), obj.holdM);
			soM = (U*)obj.soM;
			holdM = obj.holdM;
			return *this;
		}
		cPointer<T>   &operator=	(const cPointer<T> &obj)
		{
//			moveRefTo((pointedT*)obj.soM);
			moveRefTo(obj.soM, obj.holdM);
			holdM = obj.holdM;
			soM = (T*)obj.soM;
			return *this;
		}
		const cPointer<T>   &operator=	(const T* obj)
		{
//			moveRefTo((pointedT*)obj);
			moveRefTo(obj, true);
			soM = (T*)obj;
			return *this;
		}
		T* operator->()
		{
			if(!soM)
			{
				log("#w !! dereferencing zero cPointer<%s>\n", typeid(T).name() );
				CTL_Assert(!!soM);
			}
			return soM;
		}
		const T* operator->()	const
		{
			if(!soM)
			{
				log("#w !! dereferencing zero const cPointer<%s>\n", typeid(T).name() );
				CTL_Assert(!!soM);
			}
			return soM;
		}
		const	T& operator*() const
		{ return *soM;}
		T& operator*()
		{ return *soM;}
		bool	operator==	(const cPointer<T>& obj)	const
		{	return obj.soM == soM;}
		bool	operator!=	(const cPointer<T>& obj)	const
		{	return obj.soM != soM;}
		bool	operator<	(const cPointer<T>& obj)	const
		{	return obj.soM < soM;}
		const T*	get()	const
		{ return soM;}
		bool operator!() const
		{ return soM==0;}
		bool isShared() const
		{ return soM ? soM->getSharedP()->getRefCount()>1 : false; }
   	oStream &write(oStream &os) const
		{  return os<<soM; }
		iStream &read(iStream &is)
		{	T *p;  is>>p; (*this)=p; return is; }
		bool holdM;
	protected:
		T*	soM;
	};

namespace internal
{
class reflist
{
protected:
	bool holdM;
	mutable const reflist *prevM,*succM;
	reflist(bool hold = true): holdM(hold)	
	{ prevM = succM = this; }
	reflist(const reflist &p): holdM(p.holdM)
	{	ref(p);	}
	void ref(const reflist &p) const
	{
//		CTL_Assert(p.succM);
		if(!p.succM)
		{ prevM = succM = 0; 	return; }
		prevM=&p; succM=p.succM;
		succM->prevM = prevM->succM = this;
	}
	void link(const reflist &p) const
	{
		if(!succM)
			return;
		const reflist *tmp = p.succM;
		succM->prevM = &p;
		p.succM = succM;
		succM = tmp;
		tmp->prevM = this;
	}
	bool lastHolder() const
	{
		for(const reflist *tmp = succM; tmp != this; tmp=tmp->succM)
			if(tmp->holdM)
				return false;
		return holdM;
	}
	bool unref()
	{	
//		CTL_Assert(succM);
//		CTL_Assert( (holdM && !!succM) || (!holdM && !succM) );
		if(!succM)
			return false;
		if(succM!=this)
		{
			bool del = lastHolder();
			prevM->succM = succM;
			succM->prevM = prevM;
			prevM = succM = this;
			return del;
		}
		CTL_Assert(prevM==this);
		return holdM;
	}
	const reflist *next() const
	{ return succM; }
	bool isShared() const
	{ return succM!=this; }
public:
	int getRefCount() const 
	{
		int refc=int(holdM);
		for(const reflist *r=next(); !!r && r!=this; refc += int(r->holdM) ,r=r->next()) ;
		return refc;
	}
	bool isHolding() const
	{ return holdM; }

};

template<class T> struct alloc
{
	static void destroy(const void *t)
	{ delete (const T*)t; }
};

	template<> struct alloc<void> 
	{ static void destroy(const void *) {}};
	struct voidPointer
	{
		const void *p;
		void (*destroy)(const void *);
		empty operator *() const;
		template<class S> operator const S*() const
		{ return (const S*)p; }
		operator const void*() const
		{ return p; }
		bool operator!() const
		{ return !p; }
		bool operator==(const voidPointer &v) const
		{ return p==v.p; }
		bool operator!=(const voidPointer &v) const
		{ return p!=v.p; }
		bool operator==(const void *v) const
		{ return p==v; }
		template<class S> voidPointer(const S *s): p(s),destroy(&internal::alloc<S>::destroy)
		{}
		voidPointer(): p(0), destroy(0) {}
	};
}  // e.o. internal

// non intrinsic pointer
template <class T>  class lPointer : public internal::reflist
{
	typedef lPointer<T> thisT; 
	typedef internal::reflist baseT;
	enum { isVoid = ctl::same<T,void>::value };
	typedef typename ctl::select<isVoid, internal::voidPointer,const T*>::resultT pointerT;
	typedef typename ctl::select<isVoid, empty, T>::resultT pointedType;
	using  internal::reflist::holdM;
	pointerT soM;
	static void del(internal::voidPointer &t)
	{
		if(t.destroy)
			(*t.destroy)(t.p);
	}
	static void del(const T* t)
	{ delete (T*)t; }
	void destroy()
	{  if(soM) del(soM); }
	void replace(const T* obj)
	{
		const T *t=soM;

		for(lPointer<T> *i=(thisT*)next(); !!i && i!=this; i=(thisT*)i->next())
			i->soM = obj;
		soM = (T*)obj;
		if(t)
			del(t);
	}
	template<class S> void  assign(const lPointer<S> &p)
	{
		if(((const void*)soM)==p.get() && holdM == p.isHolding() )
			return;
		bool d=unref();
		holdM = p.isHolding();
		if(d)
			destroy();
		ref(p);
		soM = (const T*)(const void*)(p.get());
	}
public:
	pointerT	get ()	const
	{ return soM;}

	lPointer(): soM((T*)0) {}
	template<class S>
		lPointer(const lPointer<S> &p):baseT(p),soM(p.get())
	{}
	lPointer	(const thisT &p):baseT(p),soM(p.soM)
	{}
#if defined (CTL_MSC) &&  _MSC_VER<1274
	lPointer	(const lPointer<void> &p):baseT(p),soM((T*)p.get().p)
	{}
#endif
	template<class S> 	
		explicit lPointer(const S *p, bool hold=true): baseT(hold),soM(p) {}
	~lPointer()
	{	
		if(unref() && !!soM)
			destroy();
	}
	template<class S> 	thisT   &operator=(const lPointer<S> &p)
	{
		assign(p);
		return *this;
	}
	template<class S> 	const thisT &set(const lPointer<S> &p)
	{
		assign(p);
		return *this;
	}
	thisT   &operator=(const thisT &p)
	{
		assign(p);	
		return *this;
	}
	template<class S> const thisT &operator<<=(const lPointer<S> &p)
	{
		replace(p.get());
 		link(p);
 		return *this;
	}

	void release(bool s/*strong*/ =true)
	{
		if(s)
		{
			unref();
			prevM = succM =0;
		}
		holdM = false;		
	}
	void derelease()
	{ holdM = true; }
	void destruct()
	{ replace(0); }
	void clear()
	{
		if(unref())
			destroy();
		soM = 0;
	}
	const pointedType &operator*()	const
	{ return *soM;}

	const T*	operator->()	const
		{
			if(!soM)
			{
				log("#w !! dereferencing zero const lPointer<%s>\n", typeid(T).name() );
				CTL_Assert(!!soM);
			}
			return soM;
		}
	T* operator->()
		{
			if(!soM)
			{
				log("#w !! dereferencing zero lPointer<%s>\n", typeid(T).name() );
				CTL_Assert(!!soM);
			}
			return (T*)soM;
		}
	bool	operator==(const thisT& obj)	const
	{	return obj.soM == soM;}
	bool	operator!=(const thisT& obj)	const
	{	return obj.soM != soM;}
	bool	operator<	(const thisT& obj)	const
	{	return obj.soM < soM;}

	bool operator !() const
	{ return !soM;}
	bool isShared() const                               
	{ return soM && baseT::isShared(); }	
	
// implemented in serialze.inl
	oStream &write(oStream &os) const;
	iStream &read(iStream &is);
};


}  // ctl
#endif
