/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_SERIALIZE_H_
#define CTL_SERIALIZE_H_

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <vector>
#include <string>
#include <queue>

#include "timer.h"
#include "types.h"
#include "vector.h"
#include "pointer.h"
#include "typename.h"

namespace ctl
{

//	typedef std::exception exception;

class commuBase
{
	bool mustSwapM, hasWriteVM;
public:
	commuBase(bool mustSwap, bool hasWriteV):  mustSwapM(mustSwap), hasWriteVM(hasWriteV) {}
	bool mustSwap() const
	{return mustSwapM; }
	bool mustCopy() const
	{return !hasWriteVM; }
};

template<int n> struct exceptionN: public std::exception
{
	ctl::string whatM;
	exceptionN(const ctl::string &s = ctl::string()): whatM(s) {}
	virtual ~exceptionN() throw() {}
	const char* what () const throw()  { return whatM.c_str (); }
//	virtual const char* what () const { return whatM.c_str (); }
};
template<int n>	inline std::ostream &operator<<(std::ostream &os, const exceptionN<n> &e)
{ return os<<e.whatM; }

typedef exceptionN<0> exception;
typedef exceptionN<1> groupException;
namespace internal
{
	template<class T> inline void copyBytes(char *dest, const char *x, uint8 n , bool swap, T*)
	{
//		CTL_ProfileC("copyBytes", n>500);
		if(dest)
		{
			if(swap && sizeof(T)>1 )
			{
				const char *end = dest+n*sizeof(T) ;
				x += sizeof(T);
				for(;dest<end; x+=2*sizeof(T))
					for(unsigned int j=0; j<sizeof(T); j++)
			 			*dest++ = * --x;
			}
			else
				memcpy(dest, x, n*sizeof(T) );
		}
	}
}  //  internal

template<class T> class stack
{
	int cM;
	vector<T> stM;
public:
	stack() : cM(0) {}
	void reset()
	{ cM = 0; }
	int size() const
	{ return cM; }
	void push(const T &t)
	{ 
		cM++; 
		stM.push_back(t); 
	}
	T pop()
	{ 
		CTL_Assert(cM>0);
		T t= stM[--cM]; 
		stM.resize(cM);
		return t; 
	}
};

template<class T> class stackMemory
{
	stack<int> stackM;
	vector<T> heapM;
	int cM;
public:
	stackMemory(): cM(0)
	{}
	size_t size() const
	{ return heapM.size(); }
	void reset()
	{
		stackM.reset();
		cM = 0;
	}
	void push(const T &t)
	{
		stackM.push(int(heapM.size()));
		heapM.push_back(t);
	};
	T &next()
	{
		stackM.push(cM);
		return heapM[cM++];
	};
	T &pop()
	{
		int hc = stackM.pop();
		return heapM[hc];
	}
};

struct PID
{
	struct  IPvX
	{
		ctl::int4 port;
		ctl::uint2 ip[8];
		IPvX() {}
		IPvX(const PID &pid);
		operator PID() const;
	};
	ctl::int4 port;
	ctl::string host;
	ctl::string toString() const
	{
//		char buf[12];
//		sprintf(buf,"%d", port);  
		return host+":"+ port;
	} 
/*
	PID(const ctl::string &s)
	{
		int at=s.find(':');
		const char *p=s.c_str();
		if(at>=0)
		{
			host=s.substr(0,at);
			p += at+1;
		}
		sscanf(p,"%d",&port);
	}
*/
	int cmp(const PID &p) const;
	bool operator==(const PID &p) const
	{ return cmp(p)==0;  }
	bool operator!=(const PID &p) const
	{ return cmp(p)!=0; }
	bool operator==(ctl::int4 p) const
	{ return port == p; }
	operator uint4() const { return port; }
	bool isLogical() const
	{ return host.empty(); }
	bool operator !() const
	{ return port<0; }
	bool operator <(const PID &p) const
	{ return cmp(p)<0; }
	PID(ctl::int4 t, const ctl::string& h=ctl::string()): port(t), host(h) {}
	PID(): port(-1) {}
};


struct dataObject
{
	uint8 n;
	uint4 size;
	const void *data;
	dataObject(uint8 nn=0, uint4 s=0, const void *d=0): n(nn), size(s), data(d)
	{}
};
 // class byteBuffer
class byteBuffer
{
protected:
	const vector<char> bufferM;
	char *posM, *startM, *endM;
	vector<dataObject> dataM;
	uint4 dataCM;
	uint8 extSizeM;
	bool  swapM;
public:
	
	lPointer<commuBase> commuM;
   const  vector<dataObject> &getData() const
	{ return  dataM; }
 	byteBuffer(uint8 size=0,  char *buf =0): posM(0),startM(0),endM(0),dataCM(0),extSizeM(0),swapM(false)
 	{ setBuffer(size, buf);  }
	byteBuffer(const vector<char> &buf): bufferM(buf),posM((char*)bufferM.begin()),startM(startM),endM(startM+bufferM.size()),dataCM(0),extSizeM(0),swapM(false)
	{}
	byteBuffer &operator=(const byteBuffer &b)
	{
		setBuffer(b.bufferM);
		dataM = b.dataM;
		extSizeM = b.extSizeM;
		swapM = b.swapM;
		return *this;
	}
	void clear()
	{
  	((vector<char>&)bufferM).clear();
		startM = posM = endM = 0;
		extSizeM = 0;
	}
  void setBuffer(const vector<char> &buf)
	{
		((vector<char>&)bufferM) = buf;
	  	posM = startM = (char*)bufferM.begin(); ;
	  	endM = startM + bufferM.size();
	}
  void setBuffer(uint8 size, char *buf)
  {
  	if(size==0)
  		return;
  	if(buf==0)
  	{
  		((vector<char>&)bufferM).resizeRaw(size_t(size));
  		buf = (char*)bufferM.begin();
  	}
  	posM = startM = buf ;
  	endM = startM + size;
  }
	static bool mustSwap()
	{	int test=1; return  *((char*)&test)!=1; }
	void seek(unsigned int pos)
	{ posM = startM+pos; }
	unsigned int tell() const
	{ return posM-startM; }
	const char *buffer() const
	{ return startM; }
	const char *currentBuffer() const
	{ return posM; }
	char *buffer()  { return startM; }
 	uint8 currentSize() const
 	{ return posM-startM; }
 	bool empty() const
 	{ return posM == endM; }
	uint8 size() const
	{ return endM-startM + extSizeM; }
	void skip(uint8 d, unsigned char c=255)
	{
		if(c!=255)
			::memset(posM, c, d);
		posM  += d;
	}
	int compare(const byteBuffer& b) const
	{ return size()==b.size() ? strncmp(startM, b.startM, size() ) : int(size()-b.size()); }
	bool operator==(const byteBuffer& b) const
	{ return compare(b)==0; }
};

struct blockBegin
{
	int8 size;
	blockBegin(): size(0) {}
};
struct blockEnd
{
	int8 size;
	blockEnd(): size(0) {}
};

//! class oStream
class oStream : public byteBuffer
{
	struct objectBlock: public shared
	{
		std::map<const void*,int> logAddrM;
		stackMemory<int8> stackM;
	};
	mutable cPointer<objectBlock> objBM;
	const enum { Count=0, Write=1 } modusM;	
	void addDate(uint8 n, uint4 s,const  void *d)
	{
		CTL_Assert(dataM.size()>dataCM);
		dataM[2*dataCM].n = posM-(char*)dataM[2*dataCM].data;
		dataM[2*dataCM+1] = dataObject(n,s,d);
		dataM[2*dataCM+2] = dataObject(endM-posM,1, posM);
		dataCM++;
		extSizeM += n * s;
	}
public:
	uint4 getNData() const
	{ return dataCM; }
//	enum { headSize = MessHeader::fixsize };
 	oStream(uint8 bsize, void *buf=0, const lPointer<commuBase> &commu =lPointer<commuBase>(), uint4 nData =1): objBM(0), modusM(Write)
 	{
		CTL_Assert(bsize>0 && nData>0);
		commuM = commu;
		setBuffer(bsize, (char*)buf);
		dataM.resize(2*nData-1);							
		dataM[0] = dataObject(size(),1,startM);
	}
 	oStream(const lPointer<commuBase> &commu):byteBuffer(0), objBM(0), modusM(Count)
	{  dataCM=1;	commuM = commu;	}
	bool counting() const
	{ return modusM==Count; }
 	bool writing() const
	{ return modusM==Write; }
	template<class T>
	     oStream &write(const T *t, uint8 n=1, bool cpy = false)
	{
			bool copy = cpy || n*sizeof(T) < CTL_MaxInline || !commuM || commuM->mustCopy();
			if(writing())
			{
				if(copy)
					internal::copyBytes(posM, (const char*)t, n, swapM, (T*)0);
				else
					addDate(n, sizeof(T), t);
			}
			else
				if(!copy)
					dataCM++;
			if(copy)
				posM += n*sizeof(T);
			return *this;
	}
	bool writeFile(FILE *str)
	{
 	 	if(!str)
			return false;
 	 	uint8 w=fwrite(buffer(),1,int4(currentSize()),str);
 	 	return  w==currentSize();	 	
	}
	bool writeFile(const char *fname)
	{
 	 	FILE *str= fopen(fname,"w");
 	 	bool ok= writeFile(str);
 	 	if(str)
			fclose(str);
 	 	return ok;
	}
	int getLogAddr(const void *adr) const
	{
		if(!objBM)
 			objBM = new  objectBlock;

		std::map<const void*,int>::const_iterator adrI=objBM->logAddrM.find(adr);

		if(adrI!=objBM->logAddrM.end())
			return adrI->second;
		return -1;
	}
	void appendLogAddr(const void *adr)
	{
		size_t log = objBM->logAddrM.size()+1;
		objBM->logAddrM[adr] = int(log);
	}
	void setBlocks(const stackMemory<int8> &stack)
	{
		if(!stack.size())
			return;
		objBM = new  objectBlock;
		objBM->stackM = stack; objBM->stackM.reset();
	}
	void getBlocks(stackMemory<int8> &block) const
	{
		if(!!objBM)
			block = objBM->stackM;
	}
	bool write(const void *t,  const typeId &type);
	oStream &write(blockBegin)
  {
		if(!objBM)
			objBM = new  objectBlock;			
 		if(counting())
 			objBM->stackM.push(tell());
 		else
 			write(&objBM->stackM.next(),1);
 		return *this;   		
  }
	oStream &write(blockEnd)
  {
 		int8 &hc=objBM->stackM.pop();
 		if(counting())
 		{
 			hc = tell() - hc;
 			write(&hc,1);
 		}
	return *this;   		
  }
	bool operator==(const oStream& os) const
	{ return compare(os)==0; }
};

//! class iStream
class iStream : public byteBuffer
{
	typedef  std::pair<int, const char*> blockT;
	struct objectBlock: public shared
	{
		std::map<const void*,lPointer<void> > physAddrSPM;
		std::map<int,const void*>  physAddrM;
		stack<blockT> blockM;
	};
	mutable cPointer<objectBlock> objBM;
	const char *limitM;
	int8 currentBlockM;
public:
	PID sourceM, hostM;
	iStream(): byteBuffer(0),objBM(0), limitM(endM), currentBlockM(0) {}
	iStream(uint8 size, const void *buf, PID src=PID(), PID host=PID()):byteBuffer(size,(char*)buf), objBM(0),limitM(endM),	currentBlockM(size),sourceM(src), hostM(host)
	{}
	iStream(const vector<char> &buf): byteBuffer(buf.size(), (char*)buf.begin()), objBM(0), limitM(endM),  currentBlockM(buf.size())
	{}
//	iStream(const oStream &os): byteBuffer(os.size(), (char*)os.buffer()), objBM(0), limitM(endM),  currentBlockM(os.size())
	iStream(const oStream &os): byteBuffer(os), objBM(0), limitM(endM),  currentBlockM(os.size())
	{ seek(0); dataCM = 1;}
	void set(uint8 size, char *buf, PID src=PID(), PID host=PID())
	{
		setBuffer(size, buf);
		if(!!src)
			sourceM = src;
		if(!!host)
			hostM = host;
		limitM = endM;
	  	currentBlockM = size;
	}
	void set(const byteBuffer &buf)
	{
		*(byteBuffer*)this = buf;
		limitM = endM;
	  currentBlockM = size();
	}
	template<class T> void create(const T *&t)
	{
		enum { abstract = conversion<abstractT<T>, T>::exist };
		if(!t)
			createObject(t, (void*)0, int2type<abstract>() );
		appendPhysAddr(t);
	}
	void setSwap()
	{ swapM = true; }
	bool getSwap() const
	{ return swapM; }
	void error(const char* mess, uint8 size)
	{
		char *p=posM;
		posM = (char*)limitM;
//		assert(false);
		ctl::throwStr(" *** %s: bufferSize = %d totalSize = %d size = %d block = %d ***",
						mess, int(endM-p), int(endM-startM), int(size),  int(currentBlockM) );
	}
	
	template<class T>
	     iStream &read(T *t, uint8 n=1)
	{
 		if(posM+n*sizeof(T)>limitM)
		{
			if(dataM.size()>dataCM && dataM[dataCM].size == sizeof(T))
			{
				setBuffer(dataM[dataCM].n*dataM[dataCM].size,(char*)dataM[dataCM].data);
				dataCM++;
				limitM = endM;
			}
			else
				error("unexpected end of stream", n*sizeof(T));
		}
//swap ?
		internal::copyBytes((char*)t, posM, n, swapM,(T*)0);
		posM += n*sizeof(T);
		return *this;
	}
	static bool strTermS(const char *pos,const char *end, size_t &size)
	{
		const char *c=pos;
		while(c<end && *c) c++;
		size = c-pos;
		return c==end;
	}
	bool strTerm(size_t &size) const
	{ return strTermS(posM, endM, size); }
	bool strTerm(size_t &size, size_t &n) const
	{
		const char *c=posM;
		size_t s=0;
		size = 0;
		while(c<endM && *c)
		{
			bool end=strTermS(c,endM,s);
			size += s;
			n++;
			if(end)
				return end;
			c += s+1;
		}
		return c==endM;
	}
	bool wstrTerm(size_t &size) const
	{
		const char *c=posM;
		while(c+1<endM && c[0] && c[1]) c+=sizeof(wchar);
		size = (c-posM)/sizeof(wchar);
		return c==endM;
	}
 	bool readFile(FILE *str)
 	{
		if(!str)
			return false;
		fseek(str,0,SEEK_END);
		size_t size = ftell(str);
		rewind(str);
 	 	setBuffer(size,0);
 	 	limitM = endM;
 	 	currentBlockM = size;
 	 	size_t w=fread(buffer(),1,size,str);
 	 	return  w==size;	 	
	}
	bool readFile(const char *fname)
	{
		FILE *str= ::fopen(fname,"r");
		bool ok= readFile(str);
		if(str)
			::fclose(str);
		return ok;
	}
	lPointer<void> getPhysAddrSP(const void *adr) const
	{
		if(!objBM)
			objBM  = new objectBlock;
		std::map<const void *,lPointer<void> >::const_iterator phys=objBM->physAddrSPM.find(adr);
		return phys!=objBM->physAddrSPM.end() ? phys->second : lPointer<void>();
	}
	const void *getPhysAddr(int log) const
	{
		if(!objBM)
			objBM  = new objectBlock;
		std::map<int,const void* >::const_iterator phys=objBM->physAddrM.find(log);
		return phys!=objBM->physAddrM.end() ? phys->second : 0;
	}
	void appendPhysAddrSP(const void  *adr,const lPointer<void> &phys)
	{
		objBM->physAddrSPM[adr] = phys;
	}
	void appendPhysAddr(const void* phys)
	{
		size_t log = objBM->physAddrM.size()+1;
		objBM->physAddrM[int(log)] = phys;
	}
 	uint8 currentBlock() const
 	{ return currentBlockM; }
    bool read(void *&t, typeId &type);
	iStream &read(blockBegin)
	{
		if(!objBM)
			objBM  = new objectBlock;
		objBM->blockM.push(blockT(int4(currentBlockM), limitM));

		read(&currentBlockM);
		limitM = posM + currentBlockM;	
		CTL_Assert(limitM <= endM);
		return *this;   		
	}
	iStream &read(blockEnd)
	{
		posM = (char*)limitM;
		blockT block=objBM->blockM.pop();
		currentBlockM = block.first;
		limitM = block.second;
		CTL_Assert(limitM <= endM);
		return *this;   		
	}
    bool copyBuffer()
	{
		if(!bufferM)
		{
			((vector<char>&)bufferM) = vector<char>(endM-startM, startM);
	  		posM = startM = (char*)bufferM.p();;
	  		limitM = endM = startM + bufferM.size();
			return true;
		}
		posM = startM;
		return false;
	}
    struct eofTest
    {};
};


inline ctl::oStream &operator<<(ctl::oStream &os, ctl::blockBegin b)
{ return os.write(b); }
inline ctl::oStream &operator<<(ctl::oStream &os, ctl::blockEnd b)
{ return os.write(b); }
inline ctl::iStream &operator>>(ctl::iStream &is, ctl::blockBegin b)
{ return is.read(b); }
inline ctl::iStream &operator>>(ctl::iStream &is, ctl::blockEnd b)
{ return is.read(b); }
inline ctl::iStream &operator>>(ctl::iStream &is, ctl::iStream::eofTest)
{
	if(!is.empty())
	{
		is.error("unexpected data in stream", 0);
	}
	return is;
}

inline ctl::oStream &operator<<(ctl::oStream &os, const ctl::exception &ex);
inline ctl::iStream &operator>>(ctl::iStream &is, ctl::exception &ex);

template<class T> inline uint8 serialSize(const T &t, stackMemory<int8> &blocks, uint4 &nData, const lPointer<commuBase> &commu)
{
	oStream counter(commu); 	counter<<t;
	counter.getBlocks(blocks);
	nData =  counter.getNData();
	return counter.currentSize();
}

template<class T0, class T1> inline uint8 serialSize(const T0 &t0, const T1 &t1, stackMemory<int8> &blocks, uint4 &nData,const lPointer<commuBase> &commu)
{
	oStream counter(commu); 	counter<<t0<<t1;
	counter.getBlocks(blocks);
	nData =  counter.getNData();
	return counter.currentSize();
}

template<class T0, class T1, class T2> inline uint8 serialSize(const T0 &t0, const T1 &t1, const T2 &t2, stackMemory<int8> &blocks, uint4 &nData,const lPointer<commuBase> &commu)
{
	oStream counter(commu); 	counter<<t0<<t1<<t2;
	counter.getBlocks(blocks);
	nData =  counter.getNData();
	return counter.currentSize();
}

template<class T0, class T1, class T2, class T3> inline uint8 serialSize(const T0 &t0, const T1 &t1, const T2 &t2, const T3 &t3, stackMemory<int8> &blocks, uint4 &nData,const lPointer<commuBase> &commu)
{
	oStream counter(commu); 	counter<<t0<<t1<<t2<<t3;
	counter.getBlocks(blocks);
	nData =  counter.getNData();
	return counter.currentSize();
}

template<class T0, class T1, class T2, class T3, class T4>
	inline uint8 serialSize(const T0 &t0, const T1 &t1, const T2 &t2, const T3 &t3, const T4 &t4, stackMemory<int8> &blocks, uint4 &nData,const lPointer<commuBase> &commu)
{
	oStream counter(commu); 	counter<<t0<<t1<<t2<<t3<<t4;
	counter.getBlocks(blocks);
	nData =  counter.getNData();
	return counter.currentSize();
}

template<class T0, class T1, class T2, class T3, class T4, class T5>
	inline uint8 serialSize(const T0 &t0, const T1 &t1, const T2 &t2, const T3 &t3, const T4 &t4, const T5 &t5,stackMemory<int8> &blocks, uint4 &nData,const lPointer<commuBase> &commu)
{
	oStream counter(commu); 	counter<<t0<<t1<<t2<<t3<<t4<<t5;
	counter.getBlocks(blocks);
	nData =  counter.getNData();
	return counter.currentSize();
}

template<class T0, class T1, class T2, class T3, class T4, class T5, class T6>
	inline uint8 serialSize(const T0 &t0, const T1 &t1, const T2 &t2, const T3 &t3, const T4 &t4, const T5 &t5, const T6 &t6,stackMemory<int8> &blocks, uint4 &nData,const lPointer<commuBase> &commu)
{
	oStream counter(commu); 	counter<<t0<<t1<<t2<<t3<<t4<<t5<<t6;
	counter.getBlocks(blocks);
	nData =  counter.getNData();
	return counter.currentSize();
}

template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7>
	inline uint8 serialSize(const T0 &t0, const T1 &t1, const T2 &t2, const T3 &t3, const T4 &t4, const T5 &t5, const T6 &t6, const T7 &t7,stackMemory<int8> &blocks, uint4 &nData,const lPointer<commuBase> &commu)
{
	oStream counter(commu); 	counter<<t0<<t1<<t2<<t3<<t4<<t5<<t6<<t7;
	counter.getBlocks(blocks);
	nData =  counter.getNData();
	return counter.currentSize();
}

template <class T> inline oStream serialize(const T &t)
{ return serializeB(t,0,0,lPointer<commuBase>()); }
template <class T0, class T1> inline oStream serialize(const T0 &t0, const T1 &t1)
{ return serializeB(t0, t1,0,0,lPointer<commuBase>()); }
template <class T0, class T1, class T2> inline oStream serialize(const T0 &t0, const T1 &t1, const T2 &t2)
{ return serializeB(t0, t1, t2,0,0,lPointer<commuBase>()); }
template <class T0, class T1, class T2,  class T3> 	inline oStream serialize(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3)
{ return serializeB(t0, t1, t2, t3, 0,0,lPointer<commuBase>()); }
template <class T0, class T1, class T2,  class T3,  class T4> 	
	inline oStream serialize(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3, const T4 &t4)
{ return serializeB(t0, t1, t2, t3, t4, 0,0,lPointer<commuBase>()); }
template <class T0, class T1, class T2, class T3, class T4, class T5> 	
	inline oStream serialize(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3, const T4 &t4, const T5 &t5)
{ return serializeB(t0, t1, t2, t3, t4, t5, 0,0,lPointer<commuBase>()); }
template <class T0, class T1, class T2, class T3, class T4, class T5, class T6> 	
	inline oStream serialize(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3, const T4 &t4, const T5 &t5, const T6 &t6)
{ return serializeB(t0, t1, t2, t3, t4, t5, t6, 0,0,lPointer<commuBase>()); }
template <class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7> 	

	inline oStream serialize(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3, const T4 &t4, const T5 &t5, const T6 &t6, const T7 &t7)
{ return serializeB(t0, t1, t2, t3, t4, t5, t6, t7, 0,0,lPointer<commuBase>()); }

template <class T> inline oStream serializeB(const T &t, void *buffer, uint8 bsize, const lPointer<commuBase> &commu=lPointer<commuBase>())
{
	stackMemory<int8> blocks;
	uint4 nData =0;
	uint8 size = serialSize(t, blocks, nData, commu);
	CTL_Assert(nData>0);
	oStream os(size,size>bsize ? 0 : buffer, commu, nData);
	os.setBlocks(blocks);
	return os<<t;
}

template <class T0, class T1> inline oStream serializeB(const T0 &t0, const T1 &t1, void *buffer, uint8 bsize, const lPointer<commuBase> &commu)
{
	stackMemory<int8> blocks;
	uint4 nData =0;
	uint8 size = serialSize(t0,t1,blocks, nData, commu);
	oStream os(size, size>bsize ? 0 : buffer, commu, nData);
	os.setBlocks(blocks);
	return os<<t0<<t1;
}

template <class T0, class T1, class T2> inline oStream serializeB(const T0 &t0, const T1 &t1, const T2 &t2, void *buffer, uint8 bsize, const lPointer<commuBase> &commu)
{
	stackMemory<int8> blocks;
	uint4 nData =0;
	uint8 size = serialSize(t0,t1,t2,blocks, nData, commu);
	oStream os(size, size>bsize ? 0 : buffer, commu, nData);
	os.setBlocks(blocks);
	return os<<t0<<t1<<t2;
}

template <class T0, class T1, class T2,  class T3>
	inline oStream serializeB(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3, void *buffer, uint8 bsize, const lPointer<commuBase> &commu)
{
	stackMemory<int8> blocks;
	uint4 nData =0;
	uint8 size = serialSize(t0,t1,t2,t3,blocks, nData, commu);
	oStream os(size, size>bsize ? 0 : buffer, commu, nData);
	os.setBlocks(blocks);
	return os<<t0<<t1<<t2<<t3;
}

template <class T0, class T1, class T2,  class T3, class T4>
	inline oStream serializeB(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3 , const T4 &t4, void *buffer, uint8 bsize, const lPointer<commuBase> &commu)
{
	stackMemory<int8> blocks;
	uint4 nData =0;
	uint8 size = serialSize(t0,t1,t2,t3,t4,blocks, nData, commu);
	oStream os(size, size>bsize ? 0 : buffer, commu, nData);
	os.setBlocks(blocks);
	return os<<t0<<t1<<t2<<t3<<t4;
}

template <class T0, class T1, class T2,  class T3, class T4, class T5>
	inline oStream serializeB(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3 , const T4 &t4, const T5 &t5, void *buffer, uint8 bsize, const lPointer<commuBase> &commu)
{
	stackMemory<int8> blocks;
	uint4 nData =0;
	uint8 size = serialSize(t0,t1,t2,t3,t4,t5,blocks, nData, commu);
	oStream os(size, size>bsize ? 0 : buffer, commu, nData);
	os.setBlocks(blocks);
	return os<<t0<<t1<<t2<<t3<<t4<<t5;
}

template <class T0, class T1, class T2,  class T3, class T4, class T5, class T6>
	inline oStream serializeB(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3 , const T4 &t4, const T5 &t5, const T6 &t6, void *buffer, uint8 bsize, const lPointer<commuBase> &commu)
{
	stackMemory<int8> blocks;
	uint4 nData =0;
	uint8 size = serialSize(t0,t1,t2,t3,t4,t5,t6,blocks, nData, commu);
	oStream os(size, size>bsize ? 0 : buffer, commu, nData);
	os.setBlocks(blocks);
	return os<<t0<<t1<<t2<<t3<<t4<<t5<<t6;
}

template <class T0, class T1, class T2,  class T3, class T4, class T5, class T6, class T7>
	inline oStream serializeB(const T0 &t0, const T1 &t1, const T2 &t2 , const T3 &t3 , const T4 &t4, const T5 &t5, const T6 &t6, const T7 &t7, void *buffer, uint8 bsize,const lPointer<commuBase> &commu)
{
	stackMemory<int8> blocks;
	uint4 nData =0;
	uint8 size = serialSize(t0,t1,t2,t3,t4,t5,t6,t7,blocks, nData, commu);
	oStream os(size, size>bsize ? 0 : buffer, commu, nData);
	os.setBlocks(blocks);
	return os<<t0<<t1<<t2<<t3<<t4<<t5<<t6<<t7;
}
    
template <class T> inline T &deserialize(iStream &is, T &t)
{ 	is>>t;    	return t; }

}  // ctl

#endif

