/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_TIMER_H_
#define CTL_TIMER_H_

#include "types.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>


#ifdef CTL_Prof
#	ifdef CTL_Win32
#		include <time.h>
#	else
//#	include <ctime>
#		include <sys/time.h>
#	endif
#endif

namespace ctl
{

#ifdef CTL_Prof

#ifdef CTL_MSC
	///////////////////////////////////////////////////////////////////////////
	//! some assembler code to get a timer
	/*!
	*/////////////////////////////////////////////////////////////////////////
inline  uint8 getTick(void)
{
	// Count processor clocks
	int8 t;
	__asm {
		pushad
			cpuid                           ; serial command
			_emit	0x0F                    ; it is...
			_emit	0x31                   ; ... rdtsc
			lea		ecx, t
			mov		dword ptr [ecx], eax    ; # of processor clocks
			mov		dword ptr [ecx+4], edx  ; if __int64 is used
			popad
	}
	return t;
}	

#elif defined(__i386__) & !defined(__APPLE__)  & !defined(__PGI) 
inline uint8 getTick()
{
        unsigned long long int x;
        __asm__ volatile (" .byte 0x0f, 0x31"   :   "=A"   (x));
        return x;
}
#elif defined (__x86_64__) & !defined(__APPLE__)  & !defined(__PGI)
inline uint8 getTick()
{
  unsigned hi, lo;
  __asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
  return ( (uint8)lo)|( ((uint8)hi)<<32 );
}

/*

#elif defined(__powerpc__)

typedef unsigned long long int unsigned long long;

static __inline__ unsigned long long rdtsc(void)
{
  unsigned long long int result=0;
  unsigned long int upper, lower,tmp;
  __asm__ volatile(
                "0:                  \n"
                "\tmftbu   %0           \n"
                "\tmftb    %1           \n"
                "\tmftbu   %2           \n"
                "\tcmpw    %2,%0        \n"
                "\tbne     0b         \n"
                : "=r"(upper),"=r"(lower),"=r"(tmp)
                );
  result = upper;
  result = result<<32;
  result = result|lower;

  return(result);
}

#endif
*/
/*
inline uint8 getTick (void)
{
  int8 t;
  asm volatile (".byte 0x0f, 0x31"
       : "=A" (t)
       :
       :  "%ebx", "%ecx" );
  return t;
}
*/
#elif defined(CTL_Sparc)
extern "C"
{
int __clock_gettime(clockid_t, struct timespec*);
inline uint8 getTick(void)
{
        struct timespec tp;
        __clock_gettime(CLOCK_REALTIME, &tp);
        return int8(tp.tv_sec * int8(1000000000) + int8(tp.tv_nsec));
}
}
#elif defined(CTL_Alpha) || defined(CTL_HPUX) || defined(CTL_SGI)
inline int8 getTick(void)
{
	struct timespec tp;
	clock_gettime(CLOCK_REALTIME, &tp);
	return int8(tp.tv_sec * int8(1000000000) + int8(tp.tv_nsec));
}
#else
inline ctl::int8 getTick(void)
{ return 0; }

#endif

#else
inline ctl::uint8 getTick(void)
{ return 0; }
#endif
// CTL_Prof

class Timer
{
	double secPerTickM;
	friend Timer &timer();
	uint8 t0M;
	uint8 tM;
	Timer():secPerTickM(0), t0M(getTick())
	{
#ifdef CTL_Tick
		secPerTickM = 1.0/CTL_Tick;
#endif
		tM=getTick();
	}
public:
	double secPerTick() const
	{ return secPerTickM; }
	double dt()
	{
		uint8 dt = getTick()-tM;
		tM += dt;
		return dt*secPerTickM;
	}
	double t()
	{ return (tM=getTick())-t0M*secPerTickM; }
};
	
inline Timer &timer()
{
	static Timer timerS;
	return timerS;
}

class TmpMark;

class TimeMarker
{
	friend class TmpMark;
	char nameM[81];
	int8 tM;
	int nM;
public:
	TimeMarker(const char *name): tM(0),nM(0) 
	{ strncpy(nameM, name, 80); nameM[80] = 0; }
	~TimeMarker()
	{
		if(!nM)
			return;
		double f=tM*timer().secPerTick();
#ifdef CTL_NoLog
#	define CTL_TimeLog fprintf(stderr,
#else 
#	define 	CTL_TimeLog log(
#endif
#ifdef CTL_Tick
		if(CTL_Tick<2)
			CTL_TimeLog " ** in %s ** spent time: %8.3g ticks due to %7d call(s) (%e tick/call)\n",nameM,f,nM,f/nM);
		else
#endif
			CTL_TimeLog " ** in %s ** spent time: %8.3g sec due to %7d call(s) (%e sec/call)\n",nameM,f,nM,f/nM);
#undef CTL_TimeLog
	}
};
	
class TmpMark
{
	TmpMark &operator=(const TmpMark&);
public:
	TimeMarker &M;
	int8 tM;
	TmpMark(TimeMarker &m, bool c = true):M(m), tM(c ? getTick() : 0)
	{}
	~TmpMark()
	{
		if(tM!=0)
		{
			M.nM++;
			M.tM+=getTick()-tM;
		}
	}
};
CTL_Inline TimeMarker &newTimeMark(const char*);
} // ctl

#ifdef CTL_Prof
#	ifdef CTL_MSC
#		define CTL_LINE _
#	else
#		define CTL_LINE __LINE__
#	endif
#	define CTL_CAT_L2(a,b) a##b
#	define CTL_CAT_L1(a,b) CTL_CAT_L2(a,b)
#	define CTL_Lined(X)  CTL_CAT_L1(X, CTL_LINE)
#	define CTL_Profile(s) static ctl::TimeMarker& CTL_Lined(CTL_MarkerS)=ctl::newTimeMark(s);\
		const ctl::TmpMark CTL_Lined(CTL_MarkerL)(CTL_Lined(CTL_MarkerS) );
#	define CTL_ProfileC(s, expr) static  ctl::TimeMarker& CTL_Lined(CTL_MarkerS)=ctl::newTimeMark(s);\
		const ctl::TmpMark CTL_Lined(CTL_MarkerL) (CTL_Lined(CTL_MarkerS), expr);

//#	ifdef CTL_Debug
//#	endif
#ifdef CTL_PrettyFunction
#		define CTL_Here ctl::log("#d ?? %s at %s:%d\n", CTL_PrettyFunction, __FILE__, __LINE__);
#		define CTL_HereC(str) ctl::log("#d %s: %s at %s:%d\n", #str, CTL_PrettyFunction, __FILE__, __LINE__);
#else
#		define CTL_Here ctl::log("#d ?? %s:line %d tick %u pid = %u\n",__FILE__,__LINE__, int(ctl::getTick()), ctl::sys::getTid());
#		define CTL_HereC(str) ctl::log("#d %s  %s:line %d tick %u pid = %u\n",#str,__FILE__,__LINE__, int(ctl::getTick()), ctl::sys::getTid());
#endif
#else
// CTL_Prof

#	define CTL_Profile(s)
#	define CTL_ProfileC(s, expr)
//#	ifdef CTL_Debug
#		define CTL_Here ctl::log("#d ?? %s:line %d\n",__FILE__,__LINE__);
#		define CTL_HereC(str) ctl::log("#d %s  %s:line %d\n",#str,__FILE__,__LINE__);
//#	endif

#endif
// CTL_Prof

//#	ifndef CTL_Debug
//#		define CTL_Here
//#		define CTL_HereC(str)
//#	endif

#endif
