/******************************************************************************

This file is part of the Component Template Library (CTL) 

Copyright (C) 2006  Rainer Niekamp
email:              r.niekamp@tu-bs.de

The CTL is free software; you can redistribute it and/or modify it under the 
terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2.1 of the License, or (at your option) any later 
version. We only ask that proper credit be given to the author.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License 
for more details.

You should have received a copy of the GNU Lesser General Public License along 
with this library; if not, write to the Free Software Foundation, 
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

******************************************************************************/

#ifndef CTL_VECTOR_ALG_H_
#define CTL_VECTOR_ALG_H_

#include "vector.h"

namespace ctl
{

namespace internal
{
	template<class iter0, class iter1> inline void add(iter0 a0, iter0 o0, iter1 a1)
	{
		while(a0<o0)
			*a0++ += *a1++;
	}
	template<class iter0, class iter1> inline void add(iter0 a0, iter0 o0, iter1 a1, iter1 o1)
	{
		while(a0<o0 && a1<o1)
			*a0++ += *a1++;
	}
	template<class iter0, class iter1, class T> inline void add2(iter0 a0, iter0 o0, iter1 a1, const T& t)
	{
		while(a0<o0)
			*a0++ += *a1++ * t;
	}
	template<class iter0, class iter1, class T> inline void add2(iter0 a0, iter0 o0, iter1 a1, iter1 o1, const T& t)
	{
		while(a0<o0 && a1<o1)
			*a0++ += *a1++ * t;
	}
	template<class iter0, class iter1> inline void sub(iter0 a0, iter0 o0, iter1 a1)
	{
		while(a0<o0)
			*a0++ -= *a1++;
	}
	template<class iter0, class iter1> inline void sub(iter0 a0, iter0 o0, iter1 a1, iter1 o1 )
	{
		while(a0<o0 && a1<o1)
			*a0++ -= *a1++;
	}
	template<class iter, class T> inline void multR(iter a, iter o, const T& t)
	{
		while(a<o)
			*a++ *= t;
	}
	template<class iter, class T> inline void multL(const T& t, iter a, iter o)
	{
		while(a<o)
		{	*a = t* *a; a++; }
	}
	template<class iter, class T> inline void div(iter a, iter o, const T& t)
	{
		while(a<o)
			*a++ /= t;
	}
	template<class iter, class T> inline void neg(iter a, iter o)
	{
		while(a<0)
		{	*a = -*a;  a++; }
	}
	template<class T, class iter0, class iter1> inline void multS(iter0 a0, iter0 o0, iter1 a1, T& s)
	{
		while(a0<o0)
			s += *a0++ * *a1++;
	}
	template<class T, class iter0, class iter1> inline void multS(iter0 a0, iter0 o0, iter1 a1, iter1 o1, T& s)
	{
		while(a0<o0 && a1<o1)
			s += *a0++ * *a1++;
	}
}

template<class T> inline vector<T>& vector<T>::operator +=(const vector<T>& s)
{
	CTL_Assert(sizeM==s.sizeM);
	internal::add(begin(), end(), s.begin());
	return *this;
}

template<class T> inline sequence<T>& sequence<T>::operator +=(const sequence<T>& s)
{
	internal::add(begin(), end(), s.begin(), s.end());
	return *this;
}

template<class T> inline vector<T>& vector<T>::operator -=(const vector<T>& s)
{
	CTL_Assert(sizeM==s.sizeM);
	internal::sub(begin(), end(), s.begin());
	return *this;
}

template<class T> inline sequence<T>& sequence<T>::operator -=(const sequence<T>& s)
{
	internal::sub(begin(), end(), s.begin(), s.end());
	return *this;
}

template<class T> inline vector<T>& vector<T>::operator *=(const T& f)
{
	internal::multR(begin(), end(), f);
	return *this;
}

template<class T> inline sequence<T>& sequence<T>::operator *=(const T& f)
{
	internal::multR(begin(), end(), f);
	return *this;
}

template<class T> inline vector<T>& vector<T>::operator /=(const T& f)
{
	internal::div(begin(), end(), f);
	return *this;
}

template<class T> inline sequence<T>& sequence<T>::operator /=(const T& f)
{
	internal::div(begin(), end(), f);
	return *this;
}

template<class T> inline vector<T> operator-(const vector<T>& s)
{
	vector<T> v(s);
	internal::neg(v.begin(), v.end() );
	return v;
}

template<class T> inline sequence<T> operator-(const sequence<T>& s)
{
	sequence<T> v(s);
	internal::neg(v.begin(), v.end() );
	return v;
}
	
template<class T> inline vector<T> operator+(const vector<T>& v0,  const vector<T>& v1)
{
	 vector<T> s(v0);
	 return s += v1;
}

template<class T> inline sequence<T> operator+(const sequence<T>& v0,  const sequence<T>& v1)
{
	sequence<T> s(v0);
	return s += v1;
}

template<class T> inline vector<T> operator-(const vector<T>& v0,  const vector<T>& v1)
{
	vector<T> s(v0);
	return s -= v1;
}

template<class T> inline sequence<T> operator-(const sequence<T>& v0,  const sequence<T>& v1)
{
	sequence<T> s(v0);
	return s -= v1;
}

template<class T> inline vector<T> operator*(const vector<T>& v,  const T& f)
{
	 vector<T> s(v);
	 return s *= f;
}

template<class T> inline sequence<T> operator*(const sequence<T>& v,  const T& f)
{
	sequence<T> s(v);
	return s *= f;
}

template<class T> inline vector<T> operator*(const T& f, const vector<T>& v)
{
	vector<T> s(v);
	internal::multL(f, s.begin(), s.end());
	return s;
}

template<class T> inline sequence<T> operator*(const T& f, const sequence<T>& v)
{
	sequence<T> s(v);
	internal::multL(f, s.begin(), s.end());
	return s;
}

template<class T> inline vector<T> operator/(const vector<T>& v,  const T& f)
{
	 vector<T> s(v);
	 return s /= f;
}

template<class T> inline sequence<T> operator/(const sequence<T>& v,  const T& f)
{
	sequence<T> s(v);
	return s /= f;
}

template<class T> inline  void vector<T>::add(const vector<T>& v, const T& f)
{
	CTL_Assert(size()==v.size());
	internal::add2(begin(), end(), v.begin(), f);
}

template<class T> inline  void sequence<T>::add(const sequence<T>& v, const T& f)
{
	internal::add2(begin(), end(), v.begin(), v.end(), f);
}

template<class S, class T> inline typename internal::resultType<S, T, '*'>::resultT scalMult(const S& v0,  const T& v1)
{ return v0 * v1;}

template<class S, class T> inline typename internal::resultType<S, T, '*'>::resultT scalMult(const vector<S>& v0,  const vector<T>& v1)
{
	CTL_Assert(v0.size()==v1.size());
	typename internal::resultType<S, T, '*'>::resultT s=T();
	internal::multS(v0.begin(), v0.end(), v1.begin(), s);
	return s;
}

template<class S, class T> inline typename internal::resultType<S, T, '*'>::resultT scalMult(const sequence<S>& v0,  const sequence<T>& v1)
{
	typename internal::resultType<S, T, '*'>::resultT s=T();
	internal::multS(v0.begin(), v0.end(), v1.begin(), v1.end(), s);
	return s;
}

template<class F, class T> inline vector<T> apply(F f,  const vector<T>& x)
{
	vector<T> y(x.size());
	const T *a=x.begin(), *o=a + x.size();
	T *py = y.begin();
	while(a<o)
		*py++ = f(*a++);
	return y;
}

namespace internal
{
template<class T>  inline const T *closingKet(const T *a, const  T *o, const T& bra, const T& ket)
{
	int cl=1;
	for(const T *i=a; i<o; i++)
	{
		if(*i==ket && --cl==0)
			return i;
		if(*i==bra)
			cl++;
	}
	return 0;
}

template<class T>
	inline ctl::vector<T> terminatedVector(const T*a, const T* o, 	bool terminate, const T& term)
{
	int e = terminate ? 1 :0;
	ctl::vector<T> v(o-a+e,a);
	if(terminate)
		v[o-a] = term;
	return v;
}
	
}  // internal


template<class T>
	inline ctl::vector<ctl::vector<T> > split(const ctl::vector<T>& v, const T& sep, const T& bra, const T& ket,
																bool terminate=true, const T& term=T())
{
	CTL_Assert(sep!=bra);
	ctl::vector<ctl::vector<T> > splt;
	if(!v)
		return splt;
	const T* a=v.begin(), *o=a+v.size();
	const T *lastI=a;
	for(const T *i=a; i<o; i++)
	{
		if(*i==bra)
		{
			if(i>lastI)
				splt.push_back(internal::terminatedVector(lastI,i,terminate,term));
			lastI=i+1;
			if(!(i = internal::closingKet(i+1,o,bra,ket)))
				break;
			splt.push_back(internal::terminatedVector(lastI,i,terminate,term));
			if(++i==o)
				return splt;
			lastI = i;
			while(*lastI==sep)
				lastI++;
		}
		if(*i==sep || *i==ket)
		{
			if(i>=lastI)
				splt.push_back(internal::terminatedVector(lastI,i,terminate,term));
			lastI = i+1;
		}
	}
	splt.push_back(internal::terminatedVector(lastI,o,terminate,term));
	return splt;
}
	

} // ctl
#endif
